@@ -313,6 +313,7 @@ struct rspamd_config {
 	gboolean enable_experimental;                   /**< Enable experimental plugins						*/
 	gboolean disable_pcre_jit;                      /**< Disable pcre JIT									*/
 	gboolean disable_lua_squeeze;                   /**< Disable lua rules squeezing						*/
+	gboolean own_lua_state;                         /**< True if we have created lua_state internally		*/
 
 	gsize max_diff;                                 /**< maximum diff size for text parts					*/
 	gsize max_cores_size;                           /**< maximum size occupied by rspamd core files			*/
@@ -351,7 +352,7 @@ struct rspamd_config {
 	ucl_object_t *rcl_obj;                          /**< rcl object											*/
 	ucl_object_t *config_comments;                  /**< comments saved from the config						*/
 	ucl_object_t *doc_strings;                      /**< documentation strings for config options			*/
-	GHashTable * c_modules;                         /**< hash of c modules indexed by module name			*/
+	GPtrArray *c_modules;                           /**< list of C modules			*/
 	GHashTable * composite_symbols;                 /**< hash of composite symbols indexed by its name		*/
 	GList *classifiers;                             /**< list of all classifiers defined                    */
 	GList *statfiles;                               /**< list of all statfiles in config file order         */
@@ -367,7 +368,8 @@ struct rspamd_config {
 
 	GList *maps;                                    /**< maps active										*/
 	gdouble map_timeout;                            /**< maps watch timeout									*/
-	gdouble map_file_watch_multiplier;              /**< multiplier for watch timeout when maps are files */
+	gdouble map_file_watch_multiplier;              /**< multiplier for watch timeout when maps are files	*/
+	gchar *maps_cache_dir;                          /**< where to save HTTP cached data						*/
 
 	gdouble monitored_interval;                     /**< interval between monitored checks					*/
 	gboolean disable_monitored;                     /**< disable monitoring completely						*/
@@ -378,6 +380,7 @@ struct rspamd_config {
 	gchar * checksum;                               /**< real checksum of config file						*/
 	gchar * dump_checksum;                          /**< dump checksum of config file						*/
 	gpointer lua_state;                             /**< pointer to lua state								*/
+	gpointer lua_thread_pool;                       /**< pointer to lua thread (coroutine) pool				*/
 
 	gchar * rrd_file;                               /**< rrd file to store statistics						*/
 	gchar * history_file;                           /**< file to save rolling history						*/
@@ -445,11 +448,16 @@ struct rspamd_config {
 gboolean rspamd_parse_bind_line (struct rspamd_config *cfg,
 	struct rspamd_worker_conf *cf, const gchar *str);
 
+
+enum rspamd_config_init_flags {
+	RSPAMD_CONFIG_INIT_DEFAULT = 0,
+	RSPAMD_CONFIG_INIT_SKIP_LUA = (1 << 0)
+};
 /**
  * Init default values
  * @param cfg config file
  */
-struct rspamd_config *rspamd_config_new (void);
+struct rspamd_config *rspamd_config_new (enum rspamd_config_init_flags flags);
 
 /**
  * Free memory used by config structure
@@ -481,10 +489,15 @@ enum rspamd_post_load_options {
 	RSPAMD_CONFIG_INIT_LIBS = 1 << 1,
 	RSPAMD_CONFIG_INIT_SYMCACHE = 1 << 2,
 	RSPAMD_CONFIG_INIT_VALIDATE = 1 << 3,
-	RSPAMD_CONFIG_INIT_NO_TLD = 1 << 4
+	RSPAMD_CONFIG_INIT_NO_TLD = 1 << 4,
+	RSPAMD_CONFIG_INIT_PRELOAD_MAPS = 1 << 5,
 };
 
-#define RSPAMD_CONFIG_LOAD_ALL (RSPAMD_CONFIG_INIT_URL|RSPAMD_CONFIG_INIT_LIBS|RSPAMD_CONFIG_INIT_SYMCACHE|RSPAMD_CONFIG_INIT_VALIDATE)
+#define RSPAMD_CONFIG_LOAD_ALL (RSPAMD_CONFIG_INIT_URL| \
+		RSPAMD_CONFIG_INIT_LIBS| \
+		RSPAMD_CONFIG_INIT_SYMCACHE| \
+		RSPAMD_CONFIG_INIT_VALIDATE| \
+		RSPAMD_CONFIG_INIT_PRELOAD_MAPS)
 
 /**
  * Do post load actions for config
