@@ -20,6 +20,7 @@
 #include "uthash_strcase.h"
 #include "filter.h"
 #include "lua/lua_common.h"
+#include "lua/lua_thread_pool.h"
 #include "map.h"
 #include "map_helpers.h"
 #include "map_private.h"
@@ -56,6 +57,7 @@ static gchar * rspamd_ucl_read_cb (gchar * chunk,
 	struct map_cb_data *data,
 	gboolean final);
 static void rspamd_ucl_fin_cb (struct map_cb_data *data);
+static void rspamd_ucl_dtor_cb (struct map_cb_data *data);
 
 guint rspamd_config_log_id = (guint)-1;
 RSPAMD_CONSTRUCTOR(rspamd_config_log_init)
@@ -114,7 +116,7 @@ rspamd_parse_bind_line (struct rspamd_config *cfg,
 }
 
 struct rspamd_config *
-rspamd_config_new (void)
+rspamd_config_new (enum rspamd_config_init_flags flags)
 {
 	struct rspamd_config *cfg;
 
@@ -133,7 +135,6 @@ rspamd_config_new (void)
 	cfg->max_diff = 20480;
 
 	rspamd_config_init_metric (cfg);
-	cfg->c_modules = g_hash_table_new (rspamd_str_hash, rspamd_str_equal);
 	cfg->composite_symbols =
 		g_hash_table_new (rspamd_str_hash, rspamd_str_equal);
 	cfg->classifiers_symbols = g_hash_table_new (rspamd_str_hash,
@@ -172,7 +173,12 @@ rspamd_config_new (void)
 	cfg->min_word_len = DEFAULT_MIN_WORD;
 	cfg->max_word_len = DEFAULT_MAX_WORD;
 
-	cfg->lua_state = rspamd_lua_init ();
+	if (!(flags & RSPAMD_CONFIG_INIT_SKIP_LUA)) {
+		cfg->lua_state = rspamd_lua_init ();
+		cfg->own_lua_state = TRUE;
+		cfg->lua_thread_pool = lua_thread_pool_new (cfg->lua_state);
+	}
+
 	cfg->cache = rspamd_symbols_cache_new (cfg);
 	cfg->ups_ctx = rspamd_upstreams_library_init ();
 	cfg->re_cache = rspamd_re_cache_new ();
@@ -193,6 +199,8 @@ rspamd_config_new (void)
 #endif
 	cfg->default_max_shots = DEFAULT_MAX_SHOTS;
 	cfg->max_sessions_cache = DEFAULT_MAX_SESSIONS;
+	cfg->maps_cache_dir = rspamd_mempool_strdup (cfg->cfg_pool, RSPAMD_DBDIR);
+	cfg->c_modules = g_ptr_array_new ();
 
 	REF_INIT_RETAIN (cfg, rspamd_config_free);
 
@@ -205,8 +213,6 @@ rspamd_config_free (struct rspamd_config *cfg)
 	struct rspamd_config_post_load_script *sc, *sctmp;
 	struct rspamd_worker_log_pipe *lp, *ltmp;
 
-	rspamd_map_remove_all (cfg);
-
 	DL_FOREACH_SAFE (cfg->finish_callbacks, sc, sctmp) {
 		luaL_unref (cfg->lua_state, LUA_REGISTRYINDEX, sc->cbref);
 		g_free (sc);
@@ -217,23 +223,16 @@ rspamd_config_free (struct rspamd_config *cfg)
 		g_free (sc);
 	}
 
-	if (cfg->monitored_ctx) {
-		rspamd_monitored_ctx_destroy (cfg->monitored_ctx);
-	}
+	rspamd_map_remove_all (cfg);
+	rspamd_mempool_destructors_enforce (cfg->cfg_pool);
 
 	g_list_free (cfg->classifiers);
 	g_list_free (cfg->workers);
 	rspamd_symbols_cache_destroy (cfg->cache);
-#ifdef WITH_HIREDIS
-	if (cfg->redis_pool) {
-		rspamd_redis_pool_destroy (cfg->redis_pool);
-	}
-#endif
 	ucl_object_unref (cfg->rcl_obj);
 	ucl_object_unref (cfg->config_comments);
 	ucl_object_unref (cfg->doc_strings);
 	ucl_object_unref (cfg->neighbours);
-	g_hash_table_unref (cfg->c_modules);
 	g_hash_table_remove_all (cfg->composite_symbols);
 	g_hash_table_unref (cfg->composite_symbols);
 	g_hash_table_remove_all (cfg->cfg_params);
@@ -244,14 +243,29 @@ rspamd_config_free (struct rspamd_config *cfg)
 	g_hash_table_unref (cfg->wrk_parsers);
 	g_hash_table_unref (cfg->trusted_keys);
 
+	rspamd_re_cache_unref (cfg->re_cache);
+	rspamd_upstreams_library_unref (cfg->ups_ctx);
+	g_ptr_array_free (cfg->c_modules, TRUE);
+
+	if (cfg->lua_state && cfg->own_lua_state) {
+		lua_thread_pool_free (cfg->lua_thread_pool);
+		lua_close (cfg->lua_state);
+	}
+
+#ifdef WITH_HIREDIS
+	if (cfg->redis_pool) {
+		rspamd_redis_pool_destroy (cfg->redis_pool);
+	}
+#endif
+
+	rspamd_mempool_delete (cfg->cfg_pool);
+	if (cfg->monitored_ctx) {
+		rspamd_monitored_ctx_destroy (cfg->monitored_ctx);
+	}
 	if (cfg->checksum) {
 		g_free (cfg->checksum);
 	}
 
-	rspamd_re_cache_unref (cfg->re_cache);
-	rspamd_upstreams_library_unref (cfg->ups_ctx);
-	rspamd_mempool_delete (cfg->cfg_pool);
-	lua_close (cfg->lua_state);
 	REF_RELEASE (cfg->libs_ctx);
 
 	DL_FOREACH_SAFE (cfg->log_pipes, lp, ltmp) {
@@ -645,6 +659,11 @@ rspamd_config_parse_log_format (struct rspamd_config *cfg)
 	return TRUE;
 }
 
+static void
+rspamd_urls_config_dtor (gpointer _unused)
+{
+	rspamd_url_deinit ();
+}
 
 /*
  * Perform post load actions
@@ -743,6 +762,9 @@ rspamd_config_post_load (struct rspamd_config *cfg,
 		else {
 			rspamd_url_init (cfg->tld_file);
 		}
+
+		rspamd_mempool_add_destructor (cfg->cfg_pool, rspamd_urls_config_dtor,
+				NULL);
 	}
 
 	init_dynamic_config (cfg);
@@ -834,7 +856,11 @@ rspamd_config_post_load (struct rspamd_config *cfg,
 			ret = FALSE;
 		}
 
-		return rspamd_symbols_cache_validate (cfg->cache, cfg, FALSE) && ret;
+		ret = rspamd_symbols_cache_validate (cfg->cache, cfg, FALSE) && ret;
+	}
+
+	if (opts & RSPAMD_CONFIG_INIT_PRELOAD_MAPS) {
+		rspamd_map_preload (cfg);
 	}
 
 	return ret;
@@ -1076,7 +1102,8 @@ rspamd_include_map_handler (const guchar *data, gsize len,
 			   "ucl include",
 			   rspamd_ucl_read_cb,
 			   rspamd_ucl_fin_cb,
-			   (void **)pcbdata);
+			   rspamd_ucl_dtor_cb,
+			   (void **)pcbdata) != NULL;
 }
 
 /*
@@ -1105,12 +1132,15 @@ rspamd_include_map_handler (const guchar *data, gsize len,
 #define RSPAMD_VERSION_MINOR_MACRO "VERSION_MINOR"
 #define RSPAMD_VERSION_PATCH_MACRO "VERSION_PATCH"
 #define RSPAMD_BRANCH_VERSION_MACRO "BRANCH_VERSION"
+#define RSPAMD_HOSTNAME_MACRO "HOSTNAME"
 
 void
 rspamd_ucl_add_conf_variables (struct ucl_parser *parser, GHashTable *vars)
 {
 	GHashTableIter it;
 	gpointer k, v;
+	gchar *hostbuf;
+	gsize hostlen;
 
 	ucl_parser_register_variable (parser,
 			RSPAMD_CONFDIR_MACRO,
@@ -1152,6 +1182,23 @@ rspamd_ucl_add_conf_variables (struct ucl_parser *parser, GHashTable *vars)
 			"no");
 #endif
 
+	hostlen = sysconf (_SC_HOST_NAME_MAX);
+
+	if (hostlen <= 0) {
+		hostlen = 256;
+	}
+	else {
+		hostlen ++;
+	}
+
+	hostbuf = g_alloca (hostlen);
+	memset (hostbuf, 0, hostlen);
+	gethostname (hostbuf, hostlen - 1);
+
+	/* UCL copies variables, so it is safe to pass an ephemeral buffer here */
+	ucl_parser_register_variable (parser, RSPAMD_HOSTNAME_MACRO,
+			hostbuf);
+
 	if (vars != NULL) {
 		g_hash_table_iter_init (&it, vars);
 
@@ -1337,6 +1384,19 @@ rspamd_ucl_fin_cb (struct map_cb_data *data)
 	}
 }
 
+static void
+rspamd_ucl_dtor_cb (struct map_cb_data *data)
+{
+	struct rspamd_ucl_map_cbdata *cbdata = data->cur_data;
+
+	if (cbdata != NULL) {
+		if (cbdata->buf != NULL) {
+			g_string_free (cbdata->buf, TRUE);
+		}
+		g_free (cbdata);
+	}
+}
+
 gboolean
 rspamd_check_module (struct rspamd_config *cfg, module_t *mod)
 {
@@ -1400,22 +1460,19 @@ rspamd_init_filters (struct rspamd_config *cfg, bool reconfig)
 {
 	GList *cur;
 	module_t *mod, **pmod;
-	struct module_ctx *mod_ctx;
+	guint i = 0;
+	struct module_ctx *mod_ctx, *cur_ctx;
 
 	/* Init all compiled modules */
-	if (!reconfig) {
-		for (pmod = cfg->compiled_modules; pmod != NULL && *pmod != NULL; pmod ++) {
-			mod = *pmod;
 
-			if (rspamd_check_module (cfg, mod)) {
-				mod_ctx = g_malloc0 (sizeof (struct module_ctx));
-
-				if (mod->module_init_func (cfg, &mod_ctx) == 0) {
-					g_hash_table_insert (cfg->c_modules,
-							(gpointer) mod->name,
-							mod_ctx);
-					mod_ctx->mod = mod;
-				}
+	for (pmod = cfg->compiled_modules; pmod != NULL && *pmod != NULL; pmod ++) {
+		mod = *pmod;
+		if (rspamd_check_module (cfg, mod)) {
+			if (mod->module_init_func (cfg, &mod_ctx) == 0) {
+				g_assert (mod_ctx != NULL);
+				g_ptr_array_add (cfg->c_modules, mod_ctx);
+				mod_ctx->mod = mod;
+				mod->ctx_offset = i ++;
 			}
 		}
 	}
@@ -1425,7 +1482,14 @@ rspamd_init_filters (struct rspamd_config *cfg, bool reconfig)
 
 	while (cur) {
 		/* Perform modules configuring */
-		mod_ctx = g_hash_table_lookup (cfg->c_modules, cur->data);
+		mod_ctx = NULL;
+		PTR_ARRAY_FOREACH (cfg->c_modules, i, cur_ctx) {
+			if (g_ascii_strcasecmp (cur_ctx->mod->name,
+					(const gchar *)cur->data) == 0) {
+				mod_ctx = cur_ctx;
+				break;
+			}
+		}
 
 		if (mod_ctx) {
 			mod = mod_ctx->mod;
@@ -1674,9 +1738,14 @@ rspamd_config_is_module_enabled (struct rspamd_config *cfg,
 	GList *cur;
 	struct rspamd_symbols_group *gr;
 	lua_State *L = cfg->lua_state;
+	struct module_ctx *cur_ctx;
+	guint i;
 
-	if (g_hash_table_lookup (cfg->c_modules, module_name)) {
-		is_c = TRUE;
+	PTR_ARRAY_FOREACH (cfg->c_modules, i, cur_ctx) {
+		if (g_ascii_strcasecmp (cur_ctx->mod->name, module_name) == 0) {
+			is_c = TRUE;
+			break;
+		}
 	}
 
 	if (g_hash_table_lookup (cfg->explicit_modules, module_name) != NULL) {
@@ -1851,6 +1920,9 @@ rspamd_config_radix_from_ucl (struct rspamd_config *cfg,
 	const ucl_object_t *cur, *cur_elt;
 	const gchar *str;
 
+	/* Cleanup */
+	*target = NULL;
+
 	LL_FOREACH (obj, cur_elt) {
 		type = ucl_object_type (cur_elt);
 
@@ -1861,13 +1933,18 @@ rspamd_config_radix_from_ucl (struct rspamd_config *cfg,
 
 			if (rspamd_map_is_map (str)) {
 				if (rspamd_map_add_from_ucl (cfg, cur_elt,
-						description, rspamd_radix_read, rspamd_radix_fin,
+						description,
+						rspamd_radix_read,
+						rspamd_radix_fin,
+						rspamd_radix_dtor,
 						(void **)target) == NULL) {
 					g_set_error (err, g_quark_from_static_string ("rspamd-config"),
 							EINVAL, "bad map definition %s for %s", str,
 							ucl_object_key (obj));
 					return FALSE;
 				}
+
+				return TRUE;
 			}
 			else {
 				/* Just a list */
@@ -1881,12 +1958,17 @@ rspamd_config_radix_from_ucl (struct rspamd_config *cfg,
 		case UCL_OBJECT:
 			/* Should be a map description */
 			if (rspamd_map_add_from_ucl (cfg, cur_elt,
-					description, rspamd_radix_read, rspamd_radix_fin,
+					description,
+					rspamd_radix_read,
+					rspamd_radix_fin,
+					rspamd_radix_dtor,
 					(void **)target) == NULL) {
 				g_set_error (err, g_quark_from_static_string ("rspamd-config"),
 						EINVAL, "bad map object for %s", ucl_object_key (obj));
 				return FALSE;
 			}
+
+			return TRUE;
 			break;
 		case UCL_ARRAY:
 			/* List of IP addresses */
@@ -1913,6 +1995,11 @@ rspamd_config_radix_from_ucl (struct rspamd_config *cfg,
 		}
 	}
 
+	/* Destroy on cfg cleanup */
+	rspamd_mempool_add_destructor (cfg->cfg_pool,
+			(rspamd_mempool_destruct_t)rspamd_map_helper_destroy_radix,
+			*target);
+
 	return TRUE;
 }
 
