@@ -21,9 +21,10 @@
 
 #define RSPAMD_SESSION_FLAG_WATCHING (1 << 0)
 #define RSPAMD_SESSION_FLAG_DESTROYING (1 << 1)
+#define RSPAMD_SESSION_FLAG_CLEANUP (1 << 2)
 
 #define RSPAMD_SESSION_IS_WATCHING(s) ((s)->flags & RSPAMD_SESSION_FLAG_WATCHING)
-#define RSPAMD_SESSION_IS_DESTROYING(s) ((s)->flags & RSPAMD_SESSION_FLAG_DESTROYING)
+#define RSPAMD_SESSION_CAN_ADD_EVENT(s) (!((s)->flags & (RSPAMD_SESSION_FLAG_DESTROYING|RSPAMD_SESSION_FLAG_CLEANUP)))
 
 #define msg_err_session(...) rspamd_default_log_function(G_LOG_LEVEL_CRITICAL, \
         "events", session->pool->tag.uid, \
@@ -44,6 +45,9 @@
 
 INIT_LOG_MODULE(events)
 
+/* Average symbols count to optimize hash allocation */
+static struct rspamd_counter_data events_count;
+
 struct rspamd_watch_stack {
 	event_watcher_t cb;
 	gpointer ud;
@@ -63,11 +67,22 @@ struct rspamd_async_event {
 	struct rspamd_async_watcher *w;
 };
 
+static guint rspamd_event_hash (gconstpointer a);
+static gboolean rspamd_event_equal (gconstpointer a, gconstpointer b);
+
+/* Define **SET** of events */
+KHASH_INIT (rspamd_events_hash,
+		struct rspamd_async_event *,
+		char,
+		false,
+		rspamd_event_hash,
+		rspamd_event_equal);
+
 struct rspamd_async_session {
 	session_finalizer_t fin;
 	event_finalizer_t restore;
 	event_finalizer_t cleanup;
-	GHashTable *events;
+	khash_t(rspamd_events_hash) *events;
 	void *user_data;
 	rspamd_mempool_t *pool;
 	struct rspamd_async_watcher *cur_watcher;
@@ -90,56 +105,80 @@ static guint
 rspamd_event_hash (gconstpointer a)
 {
 	const struct rspamd_async_event *ev = a;
-	rspamd_cryptobox_fast_hash_state_t st;
-	union {
+	union _pointer_fp_thunk {
 		event_finalizer_t f;
 		gpointer p;
-	} u;
+	};
+	struct ev_storage {
+		union _pointer_fp_thunk p;
+		gpointer ud;
+	} st;
 
-	u.f = ev->fin;
+	st.p.f = ev->fin;
+	st.ud = ev->user_data;
 
-	rspamd_cryptobox_fast_hash_init (&st, rspamd_hash_seed ());
-	rspamd_cryptobox_fast_hash_update (&st, &ev->user_data, sizeof (gpointer));
-	rspamd_cryptobox_fast_hash_update (&st, &u, sizeof (u));
-
-	return rspamd_cryptobox_fast_hash_final (&st);
+	return rspamd_cryptobox_fast_hash (&st, sizeof (st), rspamd_hash_seed ());
 }
 
+static void
+rspamd_session_dtor (gpointer d)
+{
+	struct rspamd_async_session *s = (struct rspamd_async_session *)d;
+
+	/* Events are usually empty at this point */
+	rspamd_set_counter_ema (&events_count, s->events->n_buckets, 0.5);
+	kh_destroy (rspamd_events_hash, s->events);
+}
 
 struct rspamd_async_session *
-rspamd_session_create (rspamd_mempool_t * pool, session_finalizer_t fin,
-	event_finalizer_t restore, event_finalizer_t cleanup, void *user_data)
+rspamd_session_create (rspamd_mempool_t * pool,
+					   session_finalizer_t fin,
+					   event_finalizer_t restore,
+					   event_finalizer_t cleanup,
+					   void *user_data)
 {
-	struct rspamd_async_session *new;
+	struct rspamd_async_session *s;
 
-	new = rspamd_mempool_alloc0 (pool, sizeof (struct rspamd_async_session));
-	new->pool = pool;
-	new->fin = fin;
-	new->restore = restore;
-	new->cleanup = cleanup;
-	new->user_data = user_data;
-	new->events = g_hash_table_new (rspamd_event_hash, rspamd_event_equal);
+	s = rspamd_mempool_alloc0 (pool, sizeof (struct rspamd_async_session));
+	s->pool = pool;
+	s->fin = fin;
+	s->restore = restore;
+	s->cleanup = cleanup;
+	s->user_data = user_data;
+	s->events = kh_init (rspamd_events_hash);
 
-	rspamd_mempool_add_destructor (pool,
-		(rspamd_mempool_destruct_t) g_hash_table_destroy,
-		new->events);
+	if (events_count.mean > 4) {
+		kh_resize (rspamd_events_hash, s->events, events_count.mean);
+	}
+	else {
+		kh_resize (rspamd_events_hash, s->events, 4);
+	}
 
-	return new;
+	rspamd_mempool_add_destructor (pool, rspamd_session_dtor, s);
+
+	return s;
 }
 
 struct rspamd_async_event *
 rspamd_session_add_event (struct rspamd_async_session *session,
-	event_finalizer_t fin,
-	void *user_data,
-	GQuark subsystem)
+						  struct rspamd_async_watcher *w,
+						  event_finalizer_t fin,
+						  gpointer user_data,
+						  GQuark subsystem)
 {
 	struct rspamd_async_event *new_event;
+	gint ret;
 
 	if (session == NULL) {
 		msg_err ("session is NULL");
-		abort ();
+		g_assert_not_reached ();
+	}
+
+	if (!RSPAMD_SESSION_CAN_ADD_EVENT (session)) {
+		msg_debug_session ("skip adding event subsystem: %s: "
+					 "session is destroying/cleaning",
+				g_quark_to_string (subsystem));
 
-		/* Not reached */
 		return NULL;
 	}
 
@@ -149,26 +188,40 @@ rspamd_session_add_event (struct rspamd_async_session *session,
 	new_event->user_data = user_data;
 	new_event->subsystem = subsystem;
 
-	if (RSPAMD_SESSION_IS_WATCHING (session)) {
-		new_event->w = session->cur_watcher;
-		new_event->w->remain ++;
-		msg_debug_session ("added event: %p, pending %d events, "
-				"subsystem: %s, watcher: %d",
-				user_data,
-				g_hash_table_size (session->events),
-				g_quark_to_string (subsystem),
-				new_event->w->id);
+	if (w == NULL) {
+		if (RSPAMD_SESSION_IS_WATCHING (session)) {
+			new_event->w = session->cur_watcher;
+			new_event->w->remain++;
+			msg_debug_session ("added event: %p, pending %d events, "
+							   "subsystem: %s, watcher: %d (%d)",
+					user_data,
+					kh_size (session->events),
+					g_quark_to_string (subsystem),
+					new_event->w->id,
+					new_event->w->remain);
+		} else {
+			new_event->w = NULL;
+			msg_debug_session ("added event: %p, pending %d events, "
+							   "subsystem: %s, no watcher!",
+					user_data,
+					kh_size (session->events),
+					g_quark_to_string (subsystem));
+		}
 	}
 	else {
-		new_event->w = NULL;
+		new_event->w = w;
+		new_event->w->remain++;
 		msg_debug_session ("added event: %p, pending %d events, "
-				"subsystem: %s, no watcher!",
+						   "subsystem: %s, explicit watcher: %d (%d)",
 				user_data,
-				g_hash_table_size (session->events),
-				g_quark_to_string (subsystem));
+				kh_size (session->events),
+				g_quark_to_string (subsystem),
+				new_event->w->id,
+				new_event->w->remain);
 	}
 
-	g_hash_table_insert (session->events, new_event, new_event);
+	kh_put (rspamd_events_hash, session->events, new_event, &ret);
+	g_assert (ret > 0);
 
 	return new_event;
 }
@@ -192,17 +245,39 @@ rspamd_session_remove_event (struct rspamd_async_session *session,
 	void *ud)
 {
 	struct rspamd_async_event search_ev, *found_ev;
+	khiter_t k;
 
 	if (session == NULL) {
 		msg_err ("session is NULL");
 		return;
 	}
 
+	if (!RSPAMD_SESSION_CAN_ADD_EVENT (session)) {
+		/* Session is already cleaned up, ignore this */
+		return;
+	}
+
 	/* Search for event */
 	search_ev.fin = fin;
 	search_ev.user_data = ud;
-	found_ev = g_hash_table_lookup (session->events, &search_ev);
-	g_assert (found_ev != NULL);
+	k = kh_get (rspamd_events_hash, session->events, &search_ev);
+	if (k == kh_end (session->events)) {
+		gchar t;
+
+		msg_err_session ("cannot find event: %p(%p)", fin, ud);
+		kh_foreach (session->events, found_ev, t, {
+			msg_err_session ("existing event %s: %p(%p)",
+					g_quark_to_string (found_ev->subsystem),
+					found_ev->fin, found_ev->user_data);
+		});
+
+		(void)t;
+
+		g_assert_not_reached ();
+	}
+
+	found_ev = kh_key (session->events, k);
+	kh_del (rspamd_events_hash, session->events, k);
 
 	/* Remove event */
 	fin (ud);
@@ -212,8 +287,8 @@ rspamd_session_remove_event (struct rspamd_async_session *session,
 		msg_debug_session ("removed event: %p, subsystem: %s, "
 				"pending %d events, watcher: %d (%d pending)", ud,
 				g_quark_to_string (found_ev->subsystem),
-				g_hash_table_size (session->events),
-				found_ev->w->id, found_ev->w->remain);
+				kh_size (session->events),
+				found_ev->w->id, found_ev->w->remain - 1);
 
 		if (found_ev->w->remain > 0) {
 			if (--found_ev->w->remain == 0) {
@@ -225,34 +300,12 @@ rspamd_session_remove_event (struct rspamd_async_session *session,
 		msg_debug_session ("removed event: %p, subsystem: %s, "
 				"pending %d events, no watcher!", ud,
 				g_quark_to_string (found_ev->subsystem),
-				g_hash_table_size (session->events));
+				kh_size (session->events));
 	}
 
-	g_hash_table_remove (session->events, found_ev);
-
 	rspamd_session_pending (session);
 }
 
-static gboolean
-rspamd_session_destroy_callback (gpointer k, gpointer v, gpointer d)
-{
-	struct rspamd_async_event *ev = v;
-	struct rspamd_async_session *session = d;
-
-	/* Call event's finalizer */
-	msg_debug_session ("removed event on destroy: %p, subsystem: %s",
-			ev->user_data,
-			g_quark_to_string (ev->subsystem));
-
-	if (ev->fin != NULL) {
-		ev->fin (ev->user_data);
-	}
-
-	/* We ignore watchers on session destroying */
-
-	return TRUE;
-}
-
 gboolean
 rspamd_session_destroy (struct rspamd_async_session *session)
 {
@@ -261,7 +314,7 @@ rspamd_session_destroy (struct rspamd_async_session *session)
 		return FALSE;
 	}
 
-	if (!(session->flags & RSPAMD_SESSION_FLAG_DESTROYING)) {
+	if (!rspamd_session_blocked (session)) {
 		session->flags |= RSPAMD_SESSION_FLAG_DESTROYING;
 		rspamd_session_cleanup (session);
 
@@ -276,14 +329,29 @@ rspamd_session_destroy (struct rspamd_async_session *session)
 void
 rspamd_session_cleanup (struct rspamd_async_session *session)
 {
+	struct rspamd_async_event *ev;
+
 	if (session == NULL) {
 		msg_err ("session is NULL");
 		return;
 	}
 
-	g_hash_table_foreach_remove (session->events,
-			rspamd_session_destroy_callback,
-			session);
+	session->flags |= RSPAMD_SESSION_FLAG_CLEANUP;
+
+	kh_foreach_key (session->events, ev, {
+		/* Call event's finalizer */
+		msg_debug_session ("removed event on destroy: %p, subsystem: %s",
+				ev->user_data,
+				g_quark_to_string (ev->subsystem));
+
+		if (ev->fin != NULL) {
+			ev->fin (ev->user_data);
+		}
+	});
+
+	kh_clear (rspamd_events_hash, session->events);
+
+	session->flags &= ~RSPAMD_SESSION_FLAG_CLEANUP;
 }
 
 gboolean
@@ -291,7 +359,7 @@ rspamd_session_pending (struct rspamd_async_session *session)
 {
 	gboolean ret = TRUE;
 
-	if (g_hash_table_size (session->events) == 0) {
+	if (kh_size (session->events) == 0) {
 		if (session->fin != NULL) {
 			msg_debug_session ("call fin handler, as no events are pending");
 
@@ -366,7 +434,7 @@ rspamd_session_events_pending (struct rspamd_async_session *session)
 
 	g_assert (session != NULL);
 
-	npending = g_hash_table_size (session->events);
+	npending = kh_size (session->events);
 	msg_debug_session ("pending %d events", npending);
 
 	if (RSPAMD_SESSION_IS_WATCHING (session)) {
@@ -456,6 +524,35 @@ rspamd_session_get_watcher (struct rspamd_async_session *session)
 	}
 }
 
+struct rspamd_async_watcher*
+rspamd_session_replace_watcher (struct rspamd_async_session *s,
+		struct rspamd_async_watcher *w)
+{
+	struct rspamd_async_watcher *res = NULL;
+
+	g_assert (s != NULL);
+
+	if (s->cur_watcher) {
+		res = s->cur_watcher;
+
+		if (!w) {
+			/* We remove watching, so clear watching flag as well */
+			s->flags &= ~RSPAMD_SESSION_FLAG_WATCHING;
+
+		}
+
+		s->cur_watcher = w;
+	}
+	else {
+		if (w) {
+			s->flags |= RSPAMD_SESSION_FLAG_WATCHING;
+		}
+
+		s->cur_watcher = w;
+	}
+
+	return res;
+}
 
 rspamd_mempool_t *
 rspamd_session_mempool (struct rspamd_async_session *session)
@@ -463,4 +560,12 @@ rspamd_session_mempool (struct rspamd_async_session *session)
 	g_assert (session != NULL);
 
 	return session->pool;
+}
+
+gboolean
+rspamd_session_blocked (struct rspamd_async_session *session)
+{
+	g_assert (session != NULL);
+
+	return !RSPAMD_SESSION_CAN_ADD_EVENT (session);
 }
\ No newline at end of file
