@@ -20,6 +20,7 @@
 #include "html.h"
 #include "html_tags.h"
 #include "html_colors.h"
+#include "html_entities.h"
 #include "url.h"
 #include <unicode/uversion.h>
 #include <unicode/ucnv.h>
@@ -172,397 +173,87 @@ static struct html_tag_def tag_defs[] = {
 	TAG_DEF(Tag_WBR, "wbr", (CM_INLINE | CM_EMPTY)),
 };
 
-struct _entity;
-typedef struct _entity entity;
+KHASH_MAP_INIT_INT (entity_by_number, const char *);
+KHASH_MAP_INIT_STR (entity_by_name, const char *);
+KHASH_MAP_INIT_STR (tag_by_name, struct html_tag_def);
+KHASH_MAP_INIT_INT (tag_by_id, struct html_tag_def);
+KHASH_INIT (color_by_name, const rspamd_ftok_t *, struct html_color, true,
+		rspamd_ftok_icase_hash, rspamd_ftok_icase_equal);
 
-struct _entity {
-	gchar *name;
-	uint code;
-	gchar *replacement;
-};
-
-
-static entity entities_defs[] = {
-	/*
-	** Markup pre-defined character entities
-	*/
-	{"quot", 34, "\""},
-	{"amp", 38, "&"},
-	{"apos", 39, "'"},
-	{"lt", 60, "<"},
-	{"gt", 62, ">"},
-
-	/*
-	** Latin-1 character entities
-	*/
-	{"nbsp", 160, " "},
-	{"iexcl", 161, "!"},
-	{"cent", 162, "cent"},
-	{"pound", 163, "pound"},
-	{"curren", 164, "current"},
-	{"yen", 165, "yen"},
-	{"brvbar", 166, NULL},
-	{"sect", 167, NULL},
-	{"uml", 168, "uml"},
-	{"copy", 169, "c"},
-	{"ordf", 170, NULL},
-	{"laquo", 171, "\""},
-	{"not", 172, "!"},
-	{"shy", 173, NULL},
-	{"reg", 174, "r"},
-	{"macr", 175, NULL},
-	{"deg", 176, "deg"},
-	{"plusmn", 177, "+-"},
-	{"sup2", 178, "2"},
-	{"sup3", 179, "3"},
-	{"acute", 180, NULL},
-	{"micro", 181, NULL},
-	{"para", 182, NULL},
-	{"middot", 183, "."},
-	{"cedil", 184, NULL},
-	{"sup1", 185, "1"},
-	{"ordm", 186, NULL},
-	{"raquo", 187, "\""},
-	{"frac14", 188, "1/4"},
-	{"frac12", 189, "1/2"},
-	{"frac34", 190, "3/4"},
-	{"iquest", 191, "i"},
-	{"Agrave", 192, "a"},
-	{"Aacute", 193, "a"},
-	{"Acirc", 194, "a"},
-	{"Atilde", 195, "a"},
-	{"Auml", 196, "a"},
-	{"Aring", 197, "a"},
-	{"AElig", 198, "a"},
-	{"Ccedil", 199, "c"},
-	{"Egrave", 200, "e"},
-	{"Eacute", 201, "e"},
-	{"Ecirc", 202, "e"},
-	{"Euml", 203, "e"},
-	{"Igrave", 204, "i"},
-	{"Iacute", 205, "i"},
-	{"Icirc", 206, "i"},
-	{"Iuml", 207, "i"},
-	{"ETH", 208, "e"},
-	{"Ntilde", 209, "n"},
-	{"Ograve", 210, "o"},
-	{"Oacute", 211, "o"},
-	{"Ocirc", 212, "o"},
-	{"Otilde", 213, "o"},
-	{"Ouml", 214, "o"},
-	{"times", 215, "t"},
-	{"Oslash", 216, "o"},
-	{"Ugrave", 217, "u"},
-	{"Uacute", 218, "u"},
-	{"Ucirc", 219, "u"},
-	{"Uuml", 220, "u"},
-	{"Yacute", 221, "y"},
-	{"THORN", 222, "t"},
-	{"szlig", 223, "s"},
-	{"agrave", 224, "a"},
-	{"aacute", 225, "a"},
-	{"acirc", 226, "a"},
-	{"atilde", 227, "a"},
-	{"auml", 228, "a"},
-	{"aring", 229, "a"},
-	{"aelig", 230, "a"},
-	{"ccedil", 231, "c"},
-	{"egrave", 232, "e"},
-	{"eacute", 233, "e"},
-	{"ecirc", 234, "e"},
-	{"euml", 235, "e"},
-	{"igrave", 236, "e"},
-	{"iacute", 237, "e"},
-	{"icirc", 238, "e"},
-	{"iuml", 239, "e"},
-	{"eth", 240, "e"},
-	{"ntilde", 241, "n"},
-	{"ograve", 242, "o"},
-	{"oacute", 243, "o"},
-	{"ocirc", 244, "o"},
-	{"otilde", 245, "o"},
-	{"ouml", 246, "o"},
-	{"divide", 247, "/"},
-	{"oslash", 248, "/"},
-	{"ugrave", 249, "u"},
-	{"uacute", 250, "u"},
-	{"ucirc", 251, "u"},
-	{"uuml", 252, "u"},
-	{"yacute", 253, "y"},
-	{"thorn", 254, "t"},
-	{"yuml", 255, "y"},
-
-	/*
-	** Extended Entities defined in HTML 4: Symbols
-	*/
-	{"fnof", 402, "f"},
-	{"Alpha", 913, "alpha"},
-	{"Beta", 914, "beta"},
-	{"Gamma", 915, "gamma"},
-	{"Delta", 916, "delta"},
-	{"Epsilon", 917, "epsilon"},
-	{"Zeta", 918, "zeta"},
-	{"Eta", 919, "eta"},
-	{"Theta", 920, "theta"},
-	{"Iota", 921, "iota"},
-	{"Kappa", 922, "kappa"},
-	{"Lambda", 923, "lambda"},
-	{"Mu", 924, "mu"},
-	{"Nu", 925, "nu"},
-	{"Xi", 926, "xi"},
-	{"Omicron", 927, "omicron"},
-	{"Pi", 928, "pi"},
-	{"Rho", 929, "rho"},
-	{"Sigma", 931, "sigma"},
-	{"Tau", 932, "tau"},
-	{"Upsilon", 933, "upsilon"},
-	{"Phi", 934, "phi"},
-	{"Chi", 935, "chi"},
-	{"Psi", 936, "psi"},
-	{"Omega", 937, "omega"},
-	{"alpha", 945, "alpha"},
-	{"beta", 946, "beta"},
-	{"gamma", 947, "gamma"},
-	{"delta", 948, "delta"},
-	{"epsilon", 949, "epsilon"},
-	{"zeta", 950, "zeta"},
-	{"eta", 951, "eta"},
-	{"theta", 952, "theta"},
-	{"iota", 953, "iota"},
-	{"kappa", 954, "kappa"},
-	{"lambda", 955, "lambda"},
-	{"mu", 956, "mu"},
-	{"nu", 957, "nu"},
-	{"xi", 958, "xi"},
-	{"omicron", 959, "omicron"},
-	{"pi", 960, "pi"},
-	{"rho", 961, "rho"},
-	{"sigmaf", 962, "sigmaf"},
-	{"sigma", 963, "sigma"},
-	{"tau", 964, "tau"},
-	{"upsilon", 965, "upsilon"},
-	{"phi", 966, "phi"},
-	{"chi", 967, "chi"},
-	{"psi", 968, "psi"},
-	{"omega", 969, "omega"},
-	{"thetasym", 977, "thetasym"},
-	{"upsih", 978, "upsih"},
-	{"piv", 982, "piv"},
-	{"bull", 8226, "bull"},
-	{"hellip", 8230, "..."},
-	{"prime", 8242, "'"},
-	{"Prime", 8243, "'"},
-	{"oline", 8254, "-"},
-	{"frasl", 8260, NULL},
-	{"weierp", 8472, NULL},
-	{"image", 8465, NULL},
-	{"real", 8476, NULL},
-	{"trade", 8482, NULL},
-	{"alefsym", 8501, "a"},
-	{"larr", 8592, NULL},
-	{"uarr", 8593, NULL},
-	{"rarr", 8594, NULL},
-	{"darr", 8595, NULL},
-	{"harr", 8596, NULL},
-	{"crarr", 8629, NULL},
-	{"lArr", 8656, NULL},
-	{"uArr", 8657, NULL},
-	{"rArr", 8658, NULL},
-	{"dArr", 8659, NULL},
-	{"hArr", 8660, NULL},
-	{"forall", 8704, NULL},
-	{"part", 8706, NULL},
-	{"exist", 8707, NULL},
-	{"empty", 8709, NULL},
-	{"nabla", 8711, NULL},
-	{"isin", 8712, NULL},
-	{"notin", 8713, NULL},
-	{"ni", 8715, NULL},
-	{"prod", 8719, NULL},
-	{"sum", 8721, "E"},
-	{"minus", 8722, "-"},
-	{"lowast", 8727, NULL},
-	{"radic", 8730, NULL},
-	{"prop", 8733, NULL},
-	{"infin", 8734, NULL},
-	{"ang", 8736, "'"},
-	{"and", 8743, "&"},
-	{"or", 8744, "|"},
-	{"cap", 8745, NULL},
-	{"cup", 8746, NULL},
-	{"gint", 8747, NULL},
-	{"there4", 8756, NULL},
-	{"sim", 8764, NULL},
-	{"cong", 8773, NULL},
-	{"asymp", 8776, NULL},
-	{"ne", 8800, "!="},
-	{"equiv", 8801, "=="},
-	{"le", 8804, "<="},
-	{"ge", 8805, ">="},
-	{"sub", 8834, NULL},
-	{"sup", 8835, NULL},
-	{"nsub", 8836, NULL},
-	{"sube", 8838, NULL},
-	{"supe", 8839, NULL},
-	{"oplus", 8853, NULL},
-	{"otimes", 8855, NULL},
-	{"perp", 8869, NULL},
-	{"sdot", 8901, NULL},
-	{"lceil", 8968, NULL},
-	{"rceil", 8969, NULL},
-	{"lfloor", 8970, NULL},
-	{"rfloor", 8971, NULL},
-	{"lang", 9001, NULL},
-	{"rang", 9002, NULL},
-	{"loz", 9674, NULL},
-	{"spades", 9824, NULL},
-	{"clubs", 9827, NULL},
-	{"hearts", 9829, NULL},
-	{"diams", 9830, NULL},
-
-	/*
-	** Extended Entities defined in HTML 4: Special (less Markup at top)
-	*/
-	{"OElig", 338, NULL},
-	{"oelig", 339, NULL},
-	{"Scaron", 352, NULL},
-	{"scaron", 353, NULL},
-	{"Yuml", 376, NULL},
-	{"circ", 710, NULL},
-	{"tilde", 732, NULL},
-	{"ensp", 8194, NULL},
-	{"emsp", 8195, NULL},
-	{"thinsp", 8201, NULL},
-	{"zwnj", 8204, NULL},
-	{"zwj", 8205, NULL},
-	{"lrm", 8206, NULL},
-	{"rlm", 8207, NULL},
-	{"ndash", 8211, "-"},
-	{"mdash", 8212, "-"},
-	{"lsquo", 8216, "'"},
-	{"rsquo", 8217, "'"},
-	{"sbquo", 8218, "\""},
-	{"ldquo", 8220, "\""},
-	{"rdquo", 8221, "\""},
-	{"bdquo", 8222, "\""},
-	{"dagger", 8224, "T"},
-	{"Dagger", 8225, "T"},
-	{"permil", 8240, NULL},
-	{"lsaquo", 8249, "\""},
-	{"rsaquo", 8250, "\""},
-	{"euro", 8364, "E"},
-};
-
-static GHashTable *html_colors_hash = NULL;
-
-static entity entities_defs_num[ (G_N_ELEMENTS (entities_defs)) ];
-static struct html_tag_def tag_defs_num[ (G_N_ELEMENTS (tag_defs)) ];
-
-static gint
-tag_cmp (const void *m1, const void *m2)
-{
-	const struct html_tag_def *p1 = m1;
-	const struct html_tag_def *p2 = m2;
-
-	if (p1->len == p2->len) {
-		return rspamd_lc_cmp (p1->name, p2->name, p1->len);
-	}
-
-	return p1->len - p2->len;
-}
-
-static gint
-tag_cmp_id (const void *m1, const void *m2)
-{
-	const struct html_tag_def *p1 = m1;
-	const struct html_tag_def *p2 = m2;
-
-	return p1->id - p2->id;
-}
-
-static gint
-tag_find_id (const void *skey, const void *elt)
-{
-	const struct html_tag *tag = skey;
-	const struct html_tag_def *d = elt;
-
-	return tag->id - d->id;
-}
-
-static gint
-tag_find (const void *skey, const void *elt)
-{
-	const struct html_tag *tag = skey;
-	const struct html_tag_def *d = elt;
-
-	if (d->len == tag->name.len) {
-		return rspamd_lc_cmp (tag->name.start, d->name, tag->name.len);
-	}
-
-	return tag->name.len - d->len;
-}
-
-static gint
-entity_cmp (const void *m1, const void *m2)
-{
-	const entity *p1 = m1;
-	const entity *p2 = m2;
-
-	return g_ascii_strcasecmp (p1->name, p2->name);
-}
-
-static gint
-entity_cmp_num (const void *m1, const void *m2)
-{
-	const entity *p1 = m1;
-	const entity *p2 = m2;
-
-	return p1->code - p2->code;
-}
+khash_t(entity_by_number) *html_entity_by_number;
+khash_t(entity_by_name) *html_entity_by_name;
+khash_t(tag_by_name) *html_tag_by_name;
+khash_t(tag_by_id) *html_tag_by_id;
+khash_t(color_by_name) *html_color_by_name;
 
 static void
 rspamd_html_library_init (void)
 {
+	guint i;
+	khiter_t k;
+	gint rc;
+
 	if (!tags_sorted) {
-		qsort (tag_defs, G_N_ELEMENTS (
-				tag_defs), sizeof (struct html_tag_def), tag_cmp);
-		memcpy (tag_defs_num, tag_defs, sizeof (tag_defs));
-		qsort (tag_defs_num, G_N_ELEMENTS (tag_defs_num),
-				sizeof (struct html_tag_def), tag_cmp_id);
+		html_tag_by_id = kh_init (tag_by_id);
+		html_tag_by_name = kh_init (tag_by_name);
+		kh_resize (tag_by_id, html_tag_by_id, G_N_ELEMENTS (tag_defs));
+		kh_resize (tag_by_name, html_tag_by_name, G_N_ELEMENTS (tag_defs));
+
+		for (i = 0; i < G_N_ELEMENTS (tag_defs); i++) {
+			k = kh_put (tag_by_id, html_tag_by_id, tag_defs[i].id, &rc);
+			kh_val (html_tag_by_id, k) = tag_defs[i];
+
+			k = kh_put (tag_by_name, html_tag_by_name, tag_defs[i].name, &rc);
+			kh_val (html_tag_by_name, k) = tag_defs[i];
+		}
+
 		tags_sorted = 1;
 	}
 
 	if (!entities_sorted) {
-		qsort (entities_defs, G_N_ELEMENTS (
-				entities_defs), sizeof (entity), entity_cmp);
-		memcpy (entities_defs_num, entities_defs, sizeof (entities_defs));
-		qsort (entities_defs_num, G_N_ELEMENTS (
-				entities_defs), sizeof (entity), entity_cmp_num);
-		entities_sorted = 1;
+		html_entity_by_number = kh_init (entity_by_number);
+		html_entity_by_name = kh_init (entity_by_name);
+		kh_resize (entity_by_number, html_entity_by_number,
+				G_N_ELEMENTS (entities_defs));
+		kh_resize (entity_by_name, html_entity_by_name,
+				G_N_ELEMENTS (entities_defs));
+
+		for (i = 0; i < G_N_ELEMENTS (entities_defs); i++) {
+			k = kh_put (entity_by_number, html_entity_by_number,
+					entities_defs[i].code, &rc);
+			kh_val (html_entity_by_number, k) = entities_defs[i].replacement;
+
+			k = kh_put (entity_by_name, html_entity_by_name,
+					entities_defs[i].name, &rc);
+			kh_val (html_entity_by_name, k) = entities_defs[i].replacement;
 	}
 
-	if (html_colors_hash == NULL) {
-		guint i;
+		html_color_by_name = kh_init (color_by_name);
+		kh_resize (color_by_name, html_color_by_name,
+				G_N_ELEMENTS (html_colornames));
 
-		html_colors_hash = g_hash_table_new_full (rspamd_ftok_icase_hash,
-				rspamd_ftok_icase_equal, g_free, g_free);
+		rspamd_ftok_t *keys;
+
+		keys = g_malloc0 (sizeof (rspamd_ftok_t) *
+				G_N_ELEMENTS (html_colornames));
 
 		for (i = 0; i < G_N_ELEMENTS (html_colornames); i ++) {
-			struct html_color *color;
-			rspamd_ftok_t *key;
+			struct html_color c;
 
-			color = g_malloc0 (sizeof (*color));
-			color->d.comp.alpha = 255;
-			color->d.comp.r = html_colornames[i].rgb.r;
-			color->d.comp.g = html_colornames[i].rgb.g;
-			color->d.comp.b = html_colornames[i].rgb.b;
-			color->valid = TRUE;
-			key = g_malloc0 (sizeof (*key));
-			key->begin = html_colornames[i].name;
-			key->len = strlen (html_colornames[i].name);
+			keys[i].begin = html_colornames[i].name;
+			keys[i].len = strlen (html_colornames[i].name);
+			k = kh_put (color_by_name, html_color_by_name,
+					&keys[i], &rc);
+			c.valid = true;
+			c.d.comp.r = html_colornames[i].rgb.r;
+			c.d.comp.g = html_colornames[i].rgb.g;
+			c.d.comp.b = html_colornames[i].rgb.b;
+			c.d.comp.alpha = 255;
+			kh_val (html_color_by_name, k) = c;
 
-			g_hash_table_insert (html_colors_hash, key, color);
 		}
+
+		entities_sorted = 1;
 	}
 }
 
@@ -599,17 +290,12 @@ rspamd_html_check_balance (GNode * node, GNode ** cur_level)
 gint
 rspamd_html_tag_by_name (const gchar *name)
 {
-	struct html_tag tag;
-	struct html_tag_def *found;
+	khiter_t k;
 
-	tag.name.start = name;
-	tag.name.len = strlen (name);
+	k = kh_get (tag_by_name, html_tag_by_name, name);
 
-	found = bsearch (&tag, tag_defs, G_N_ELEMENTS (tag_defs),
-			sizeof (tag_defs[0]), tag_find);
-
-	if (found) {
-		return found->id;
+	if (k != kh_end (html_tag_by_name)) {
+		return kh_val (html_tag_by_name, k).id;
 	}
 
 	return -1;
@@ -632,19 +318,15 @@ rspamd_html_tag_seen (struct html_content *hc, const gchar *tagname)
 	return FALSE;
 }
 
-const gchar*
+const gchar *
 rspamd_html_tag_by_id (gint id)
 {
-	struct html_tag tag;
-	struct html_tag_def *found;
+	khiter_t k;
 
-	tag.id = id;
-	/* Should work as IDs monotonically increase */
-	found = bsearch (&tag, tag_defs_num, G_N_ELEMENTS (tag_defs_num),
-				sizeof (tag_defs_num[0]), tag_find_id);
+	k = kh_get (tag_by_id, html_tag_by_id, id);
 
-	if (found) {
-		return found->name;
+	if (k != kh_end (html_tag_by_id)) {
+		return kh_val (html_tag_by_id, k).name;
 	}
 
 	return NULL;
@@ -656,8 +338,10 @@ rspamd_html_decode_entitles_inplace (gchar *s, guint len)
 {
 	guint l, rep_len;
 	gchar *t = s, *h = s, *e = s, *end_ptr;
+	const gchar *end;
+	const gchar *entity;
 	gint state = 0, val, base;
-	entity *found, key;
+	khiter_t k;
 
 	if (len == 0) {
 		l = strlen (s);
@@ -666,9 +350,11 @@ rspamd_html_decode_entitles_inplace (gchar *s, guint len)
 		l = len;
 	}
 
+	end = s + l;
+
 	while (h - s < (gint)l) {
 		switch (state) {
-		/* Out of entitle */
+		/* Out of entity */
 		case 0:
 			if (*h == '&') {
 				state = 1;
@@ -686,21 +372,27 @@ rspamd_html_decode_entitles_inplace (gchar *s, guint len)
 			if (*h == ';' && h > e) {
 				/* Determine base */
 				/* First find in entities table */
-
-				key.name = e + 1;
 				*h = '\0';
-				if (*(e + 1) != '#' &&
-					(found =
-					bsearch (&key, entities_defs, G_N_ELEMENTS (entities_defs),
-							sizeof (entity), entity_cmp)) != NULL) {
-					if (found->replacement) {
-						rep_len = strlen (found->replacement);
-						memcpy (t, found->replacement, rep_len);
-						t += rep_len;
-					}
-					else {
-						memmove (t, e, h - e);
-						t += h - e;
+				entity = e + 1;
+
+				if (*entity != '#') {
+					k = kh_get (entity_by_name, html_entity_by_name, entity);
+
+					if (k != kh_end (html_entity_by_name)) {
+						if (kh_val (html_entity_by_name, k)) {
+							rep_len = strlen (kh_val (html_entity_by_name, k));
+
+							if (end - t >= rep_len) {
+								memcpy (t, kh_val (html_entity_by_name, k),
+										rep_len);
+								t += rep_len;
+							}
+						} else {
+							if (end - t >= h - e) {
+								memmove (t, e, h - e);
+								t += h - e;
+							}
+						}
 					}
 				}
 				else if (e + 2 < h) {
@@ -719,23 +411,32 @@ rspamd_html_decode_entitles_inplace (gchar *s, guint len)
 					else {
 						val = strtoul ((e + 3), &end_ptr, base);
 					}
+
 					if (end_ptr != NULL && *end_ptr != '\0') {
 						/* Skip undecoded */
-						memmove (t, e, h - e);
-						t += h - e;
+						if (end - t >= h - e) {
+							memmove (t, e, h - e);
+							t += h - e;
+						}
 					}
 					else {
 						/* Search for a replacement */
-						key.code = val;
-						found =
-							bsearch (&key, entities_defs_num, G_N_ELEMENTS (
-									entities_defs), sizeof (entity),
-								entity_cmp_num);
-						if (found) {
-							if (found->replacement) {
-								rep_len = strlen (found->replacement);
-								memcpy (t, found->replacement, rep_len);
-								t += rep_len;
+						k = kh_get (entity_by_number, html_entity_by_number, val);
+
+						if (k != kh_end (html_entity_by_number)) {
+							if (kh_val (html_entity_by_number, k)) {
+								rep_len = strlen (kh_val (html_entity_by_number, k));
+
+								if (end - t >= rep_len) {
+									memcpy (t, kh_val (html_entity_by_number, k),
+											rep_len);
+									t += rep_len;
+								}
+							} else {
+								if (end - t >= h - e) {
+									memmove (t, e, h - e);
+									t += h - e;
+								}
 							}
 						}
 						else {
@@ -1135,7 +836,8 @@ rspamd_html_parse_tag_component (rspamd_mempool_t *pool,
 			NEW_COMPONENT (RSPAMD_HTML_COMPONENT_HREF);
 		}
 	}
-	else if (tag->id == Tag_IMG) {
+
+	if (tag->id == Tag_IMG) {
 		/* Check width and height if presented */
 		if (len == 5 && g_ascii_strncasecmp (p, "width", len) == 0) {
 			NEW_COMPONENT (RSPAMD_HTML_COMPONENT_WIDTH);
@@ -1147,6 +849,29 @@ rspamd_html_parse_tag_component (rspamd_mempool_t *pool,
 			NEW_COMPONENT (RSPAMD_HTML_COMPONENT_STYLE);
 		}
 	}
+	else if (tag->id == Tag_FONT) {
+		if (len == 5){
+			if (g_ascii_strncasecmp (p, "color", len) == 0) {
+				NEW_COMPONENT (RSPAMD_HTML_COMPONENT_COLOR);
+			}
+			else if (g_ascii_strncasecmp (p, "style", len) == 0) {
+				NEW_COMPONENT (RSPAMD_HTML_COMPONENT_STYLE);
+			}
+			else if (g_ascii_strncasecmp (p, "class", len) == 0) {
+				NEW_COMPONENT (RSPAMD_HTML_COMPONENT_CLASS);
+			}
+		}
+		else if (len == 7) {
+			if (g_ascii_strncasecmp (p, "bgcolor", len) == 0) {
+				NEW_COMPONENT (RSPAMD_HTML_COMPONENT_BGCOLOR);
+			}
+		}
+		else if (len == 4) {
+			if (g_ascii_strncasecmp (p, "size", len) == 0) {
+				NEW_COMPONENT (RSPAMD_HTML_COMPONENT_SIZE);
+			}
+		}
+	}
 	else if (tag->flags & FL_BLOCK) {
 		if (len == 5){
 			if (g_ascii_strncasecmp (p, "color", len) == 0) {
@@ -1230,24 +955,29 @@ rspamd_html_parse_tag_content (rspamd_mempool_t *pool,
 			}
 			else {
 				gchar *s;
+				khiter_t k;
 				/* We CANNOT safely modify tag's name here, as it is already parsed */
 
-				s = rspamd_mempool_alloc (pool, tag->name.len);
+				s = rspamd_mempool_alloc (pool, tag->name.len + 1);
 				memcpy (s, tag->name.start, tag->name.len);
 				tag->name.len = rspamd_html_decode_entitles_inplace (s,
 						tag->name.len);
 				tag->name.start = s;
+				s[tag->name.len] = '\0';
+				rspamd_str_lc_utf8 (s, tag->name.len);
 
-				found = bsearch (tag, tag_defs, G_N_ELEMENTS (tag_defs),
-					sizeof (tag_defs[0]), tag_find);
-				if (found == NULL) {
+				k = kh_get (tag_by_name, html_tag_by_name, s);
+
+				if (k == kh_end (html_tag_by_name)) {
 					hc->flags |= RSPAMD_HTML_FLAG_UNKNOWN_ELEMENTS;
 					tag->id = -1;
 				}
 				else {
+					found = &kh_val (html_tag_by_name, k);
 					tag->id = found->id;
 					tag->flags = found->flags;
 				}
+
 				state = spaces_after_name;
 			}
 		}
@@ -1490,7 +1220,7 @@ rspamd_html_process_url (rspamd_mempool_t *pool, const gchar *start, guint len,
 
 	/* Strip spaces from the url */
 	/* Head spaces */
-	while ( p < start + len && g_ascii_isspace (*p)) {
+	while (p < start + len && g_ascii_isspace (*p)) {
 		p ++;
 		start ++;
 		len --;
@@ -1605,11 +1335,14 @@ rspamd_html_process_url (rspamd_mempool_t *pool, const gchar *start, guint len,
 }
 
 static struct rspamd_url *
-rspamd_html_process_url_tag (rspamd_mempool_t *pool, struct html_tag *tag)
+rspamd_html_process_url_tag (rspamd_mempool_t *pool, struct html_tag *tag,
+		struct html_content *hc)
 {
 	struct html_tag_component *comp;
 	GList *cur;
 	struct rspamd_url *url;
+	const gchar *start;
+	gsize len;
 
 	cur = tag->params->head;
 
@@ -1617,7 +1350,54 @@ rspamd_html_process_url_tag (rspamd_mempool_t *pool, struct html_tag *tag)
 		comp = cur->data;
 
 		if (comp->type == RSPAMD_HTML_COMPONENT_HREF && comp->len > 0) {
-			url = rspamd_html_process_url (pool, comp->start, comp->len, comp);
+			start = comp->start;
+			len = comp->len;
+
+			/* Check base url */
+			if (hc && hc->base_url && comp->len > 2) {
+				/*
+				 * Relative url canot start from the following:
+				 * schema://
+				 * slash
+				 */
+				gchar *buf;
+				gsize orig_len;
+
+				if (rspamd_substring_search (start, len, "://", 3) == -1) {
+					/* Assume relative url */
+
+					gboolean need_slash = FALSE;
+
+					orig_len = len;
+					len += hc->base_url->urllen;
+
+					if (hc->base_url->string[hc->base_url->urllen - 1] != '/') {
+						need_slash = TRUE;
+						len ++;
+					}
+
+					buf = rspamd_mempool_alloc (pool, len + 1);
+					rspamd_snprintf (buf, len + 1, "%*s%s%*s",
+							hc->base_url->urllen, hc->base_url->string,
+							need_slash ? "/" : "",
+							(gint)orig_len, start);
+					start = buf;
+				}
+				else if (start[0] == '/' && start[1] != '/') {
+					/* Relative to the hostname */
+					orig_len = len;
+					len += hc->base_url->hostlen + hc->base_url->protocollen +
+							3 /* for :// */;
+					buf = rspamd_mempool_alloc (pool, len + 1);
+					rspamd_snprintf (buf, len + 1, "%*s://%*s/%*s",
+							hc->base_url->protocollen, hc->base_url->string,
+							hc->base_url->hostlen, hc->base_url->host,
+							(gint)orig_len, start);
+					start = buf;
+				}
+			}
+
+			url = rspamd_html_process_url (pool, start, len, comp);
 
 			if (url && tag->extra == NULL) {
 				tag->extra = url;
@@ -1819,6 +1599,7 @@ rspamd_html_process_color (const gchar *line, guint len, struct html_color *cl)
 		p ++;
 		rspamd_strlcpy (hexbuf, p, MIN ((gint)sizeof(hexbuf), end - p + 1));
 		cl->d.val = strtoul (hexbuf, NULL, 16);
+		cl->d.comp.alpha = 255;
 		cl->valid = TRUE;
 	}
 	else if (len > 4 && rspamd_lc_cmp (p, "rgb", 3) == 0) {
@@ -1828,9 +1609,10 @@ rspamd_html_process_color (const gchar *line, guint len, struct html_color *cl)
 			num1,
 			num2,
 			num3,
+			num4,
 			skip_spaces
 		} state = skip_spaces, next_state = obrace;
-		gulong r = 0, g = 0, b = 0;
+		gulong r = 0, g = 0, b = 0, opacity = 255;
 		const gchar *c;
 		gboolean valid = FALSE;
 
@@ -1898,6 +1680,40 @@ rspamd_html_process_color (const gchar *line, guint len, struct html_color *cl)
 						goto stop;
 					}
 
+					valid = TRUE;
+					p ++;
+					state = skip_spaces;
+					next_state = num4;
+				}
+				else if (*p == ')') {
+					if (!rspamd_strtoul (c, p - c, &b)) {
+						goto stop;
+					}
+
+					valid = TRUE;
+					goto stop;
+				}
+				else if (!g_ascii_isdigit (*p)) {
+					goto stop;
+				}
+				else {
+					p ++;
+				}
+				break;
+			case num4:
+				if (*p == ',') {
+					if (!rspamd_strtoul (c, p - c, &opacity)) {
+						goto stop;
+					}
+
+					valid = TRUE;
+					goto stop;
+				}
+				else if (*p == ')') {
+					if (!rspamd_strtoul (c, p - c, &opacity)) {
+						goto stop;
+					}
+
 					valid = TRUE;
 					goto stop;
 				}
@@ -1923,23 +1739,205 @@ rspamd_html_process_color (const gchar *line, guint len, struct html_color *cl)
 		stop:
 
 		if (valid) {
-			cl->d.val = b + (g << 8) + (r << 16);
+			cl->d.comp.r = r;
+			cl->d.comp.g = g;
+			cl->d.comp.b = b;
+			cl->d.comp.alpha = opacity;
 			cl->valid = TRUE;
 		}
 	}
 	else {
+		khiter_t k;
 		/* Compare color by name */
 		search.begin = line;
 		search.len = len;
 
-		el = g_hash_table_lookup (html_colors_hash, &search);
+		k = kh_get (color_by_name, html_color_by_name, &search);
 
-		if (el != NULL) {
+		if (k != kh_end (html_color_by_name)) {
+			el = &kh_val (html_color_by_name, k);
 			memcpy (cl, el, sizeof (*cl));
+			cl->d.comp.alpha = 255; /* Non transparent */
 		}
 	}
 }
 
+/*
+ * Target is used for in and out if this function returns TRUE
+ */
+static gboolean
+rspamd_html_process_css_size (const gchar *suffix, gsize len,
+		gdouble *tgt)
+{
+	gdouble sz = *tgt;
+	gboolean ret = FALSE;
+
+	if (len >= 2) {
+		if (memcmp (suffix, "px", 2) == 0) {
+			sz = (guint) sz; /* Round to number */
+			ret = TRUE;
+		}
+		else if (memcmp (suffix, "em", 2) == 0) {
+			/* EM is 16 px, so multiply and round */
+			sz = (guint) (sz * 16.0);
+			ret = TRUE;
+		}
+		else if (len >= 3 && memcmp (suffix, "rem", 3) == 0) {
+			/* equal to EM in our case */
+			sz = (guint) (sz * 16.0);
+			ret = TRUE;
+		}
+		else if (memcmp (suffix, "ex", 2) == 0) {
+			/*
+			 * Represents the x-height of the element's font.
+			 * On fonts with the "x" letter, this is generally the height
+			 * of lowercase letters in the font; 1ex = 0.5em in many fonts.
+			 */
+			sz = (guint) (sz * 8.0);
+			ret = TRUE;
+		}
+		else if (memcmp (suffix, "vw", 2) == 0) {
+			/*
+			 * Vewport width in percentages:
+			 * we assume 1% of viewport width as 8px
+			 */
+			sz = (guint) (sz * 8.0);
+			ret = TRUE;
+		}
+		else if (memcmp (suffix, "vh", 2) == 0) {
+			/*
+			 * Vewport height in percentages
+			 * we assume 1% of viewport width as 6px
+			 */
+			sz = (guint) (sz * 6.0);
+			ret = TRUE;
+		}
+		else if (len >= 4 && memcmp (suffix, "vmax", 4) == 0) {
+			/*
+			 * Vewport width in percentages
+			 * we assume 1% of viewport width as 6px
+			 */
+			sz = (guint) (sz * 8.0);
+			ret = TRUE;
+		}
+		else if (len >= 4 && memcmp (suffix, "vmin", 4) == 0) {
+			/*
+			 * Vewport height in percentages
+			 * we assume 1% of viewport width as 6px
+			 */
+			sz = (guint) (sz * 6.0);
+			ret = TRUE;
+		}
+		else if (memcmp (suffix, "pt", 2) == 0) {
+			sz = (guint) (sz * 96.0 / 72.0); /* One point. 1pt = 1/72nd of 1in */
+			ret = TRUE;
+		}
+		else if (memcmp (suffix, "cm", 2) == 0) {
+			sz = (guint) (sz * 96.0 / 2.54); /* 96px/2.54 */
+			ret = TRUE;
+		}
+		else if (memcmp (suffix, "mm", 2) == 0) {
+			sz = (guint) (sz * 9.6 / 2.54); /* 9.6px/2.54 */
+			ret = TRUE;
+		}
+		else if (memcmp (suffix, "in", 2) == 0) {
+			sz = (guint) (sz * 96.0); /* 96px */
+			ret = TRUE;
+		}
+		else if (memcmp (suffix, "pc", 2) == 0) {
+			sz = (guint) (sz * 96.0 / 6.0); /* 1pc = 12pt = 1/6th of 1in. */
+			ret = TRUE;
+		}
+	}
+	else if (suffix[0] == '%') {
+		/* Percentages from 16 px */
+		sz = (guint)(sz / 100.0 * 16.0);
+		ret = TRUE;
+	}
+
+	if (ret) {
+		*tgt = sz;
+	}
+
+	return ret;
+}
+
+static void
+rspamd_html_process_font_size (const gchar *line, guint len, guint *fs,
+							   gboolean is_css)
+{
+	const gchar *p = line, *end = line + len;
+	gchar *err = NULL, numbuf[64];
+	gdouble sz = 0;
+	gboolean failsafe = FALSE;
+
+	while (p < end && g_ascii_isspace (*p)) {
+		p ++;
+		len --;
+	}
+
+	if (g_ascii_isdigit (*p)) {
+		rspamd_strlcpy (numbuf, p, MIN (sizeof (numbuf), len + 1));
+		sz = strtod (numbuf, &err);
+
+		/* Now check leftover */
+		if (sz < 0) {
+			sz = 0;
+		}
+	}
+	else {
+		/* Ignore the rest */
+		failsafe = TRUE;
+		sz = is_css ? 16 : 1;
+		/* TODO: add textual fonts descriptions */
+	}
+
+	if (err && *err != '\0') {
+		const gchar *e = err;
+		gsize slen;
+
+		/* Skip spaces */
+		while (*e && g_ascii_isspace (*e)) {
+			e ++;
+		}
+
+		/* Lowercase */
+		slen = strlen (e);
+		rspamd_str_lc ((gchar *)e, slen);
+
+		if (!rspamd_html_process_css_size (e, slen, &sz)) {
+			failsafe = TRUE;
+		}
+	}
+	else {
+		/* Failsafe naked number */
+		failsafe = TRUE;
+	}
+
+	if (failsafe) {
+		if (is_css) {
+			/*
+			 * In css mode we usually ignore sizes, but let's treat
+			 * small sizes specially
+			 */
+			if (sz < 1) {
+				sz = 0;
+			} else {
+				sz = 16; /* Ignore */
+			}
+		} else {
+			/* In non-css mode we have to check legacy size */
+			sz = sz >= 1 ? sz * 16 : 16;
+		}
+	}
+
+	if (sz > 32) {
+		sz = 32;
+	}
+
+	*fs = sz;
+}
+
 static void
 rspamd_html_process_style (rspamd_mempool_t *pool, struct html_block *bl,
 		struct html_content *hc, const gchar *style, guint len)
@@ -1952,6 +1950,7 @@ rspamd_html_process_style (rspamd_mempool_t *pool, struct html_block *bl,
 		skip_spaces,
 	} state = skip_spaces, next_state = read_key;
 	guint klen = 0;
+	gdouble opacity = 1.0;
 
 	p = style;
 	c = p;
@@ -2009,6 +2008,38 @@ rspamd_html_process_style (rspamd_mempool_t *pool, struct html_block *bl,
 							msg_debug_html ("tag is not visible");
 						}
 					}
+					else if (klen == 9 &&
+							 g_ascii_strncasecmp (key, "font-size", 9) == 0) {
+						rspamd_html_process_font_size (c, p - c,
+								&bl->font_size, TRUE);
+						msg_debug_html ("got font size: %ud", bl->font_size);
+					}
+					else if (klen == 7 &&
+							 g_ascii_strncasecmp (key, "opacity", 7) == 0) {
+						gchar numbuf[64];
+
+						rspamd_strlcpy (numbuf, c,
+								MIN (sizeof (numbuf), p - c + 1));
+						opacity = strtod (numbuf, NULL);
+
+						if (opacity > 1) {
+							opacity = 1;
+						}
+						else if (opacity < 0) {
+							opacity = 0;
+						}
+
+						bl->font_color.d.comp.alpha = (guint8)(opacity * 255.0);
+					}
+					else if (klen == 10 &&
+							 g_ascii_strncasecmp (key, "visibility", 10) == 0) {
+						if (p - c >= 6 && rspamd_substring_search_caseless (c,
+								p - c,
+								"hidden", 6) != -1) {
+							bl->visible = FALSE;
+							msg_debug_html ("tag is not visible");
+						}
+					}
 				}
 
 				key = NULL;
@@ -2039,96 +2070,71 @@ rspamd_html_process_block_tag (rspamd_mempool_t *pool, struct html_tag *tag,
 		struct html_content *hc)
 {
 	struct html_tag_component *comp;
-	struct html_block *bl, *bl_parent;
+	struct html_block *bl;
 	rspamd_ftok_t fstr;
 	GList *cur;
-	GNode *parent;
-	struct html_tag *parent_tag;
 
 	cur = tag->params->head;
 	bl = rspamd_mempool_alloc0 (pool, sizeof (*bl));
 	bl->tag = tag;
 	bl->visible = TRUE;
+	bl->font_size = (guint)-1;
+	bl->font_color.d.comp.alpha = 255;
 
 	while (cur) {
 		comp = cur->data;
 
-		if (comp->type == RSPAMD_HTML_COMPONENT_COLOR && comp->len > 0) {
-			fstr.begin = (gchar *)comp->start;
-			fstr.len = comp->len;
-			rspamd_html_process_color (comp->start, comp->len, &bl->font_color);
-			msg_debug_html ("got color: %xd", bl->font_color.d.val);
-		}
-		else if (comp->type == RSPAMD_HTML_COMPONENT_BGCOLOR && comp->len > 0) {
-			fstr.begin = (gchar *)comp->start;
-			fstr.len = comp->len;
-			rspamd_html_process_color (comp->start, comp->len, &bl->background_color);
-			msg_debug_html ("got color: %xd", bl->font_color.d.val);
+		if (comp->len > 0) {
+			switch (comp->type) {
+			case RSPAMD_HTML_COMPONENT_COLOR:
+				fstr.begin = (gchar *) comp->start;
+				fstr.len = comp->len;
+				rspamd_html_process_color (comp->start, comp->len,
+						&bl->font_color);
+				msg_debug_html ("got color: %xd", bl->font_color.d.val);
+				break;
+			case RSPAMD_HTML_COMPONENT_BGCOLOR:
+				fstr.begin = (gchar *) comp->start;
+				fstr.len = comp->len;
+				rspamd_html_process_color (comp->start, comp->len,
+						&bl->background_color);
+				msg_debug_html ("got color: %xd", bl->font_color.d.val);
 
-			if (tag->id == Tag_BODY) {
-				/* Set global background color */
-				memcpy (&hc->bgcolor, &bl->background_color, sizeof (hc->bgcolor));
+				if (tag->id == Tag_BODY) {
+					/* Set global background color */
+					memcpy (&hc->bgcolor, &bl->background_color,
+							sizeof (hc->bgcolor));
+				}
+				break;
+			case RSPAMD_HTML_COMPONENT_STYLE:
+				bl->style.len = comp->len;
+				bl->style.start = comp->start;
+				msg_debug_html ("got style: %*s", (gint) bl->style.len,
+						bl->style.start);
+				rspamd_html_process_style (pool, bl, hc, comp->start, comp->len);
+				break;
+			case RSPAMD_HTML_COMPONENT_CLASS:
+				fstr.begin = (gchar *) comp->start;
+				fstr.len = comp->len;
+				bl->class = rspamd_mempool_ftokdup (pool, &fstr);
+				msg_debug_html ("got class: %s", bl->class);
+				break;
+			case RSPAMD_HTML_COMPONENT_SIZE:
+				fstr.begin = (gchar *) comp->start;
+				fstr.len = comp->len;
+				rspamd_html_process_color (comp->start, comp->len,
+						&bl->font_color);
+				msg_debug_html ("got color: %xd", bl->font_color.d.val);
+				break;
+			default:
+				/* NYI */
+				break;
 			}
 		}
-		else if (comp->type == RSPAMD_HTML_COMPONENT_STYLE && comp->len > 0) {
-			bl->style.len = comp->len;
-			bl->style.start =  comp->start;
-			msg_debug_html ("got style: %*s", (gint)bl->style.len, bl->style.start);
-			rspamd_html_process_style (pool, bl, hc, comp->start, comp->len);
-		}
-		else if (comp->type == RSPAMD_HTML_COMPONENT_CLASS && comp->len > 0) {
-			fstr.begin = (gchar *)comp->start;
-			fstr.len = comp->len;
-			bl->class = rspamd_mempool_ftokdup (pool, &fstr);
-			msg_debug_html ("got class: %s", bl->class);
-		}
 
 		cur = g_list_next (cur);
 	}
 
-	if (!bl->background_color.valid) {
-		/* Try to propagate background color from parent nodes */
-		for (parent = tag->parent; parent != NULL; parent = parent->parent) {
-			parent_tag = parent->data;
-
-			if (parent_tag && (parent_tag->flags & FL_BLOCK) && parent_tag->extra) {
-				bl_parent = parent_tag->extra;
-
-				if (bl_parent->background_color.valid) {
-					memcpy (&bl->background_color, &bl_parent->background_color,
-							sizeof (bl->background_color));
-					break;
-				}
-			}
-		}
-	}
-	if (!bl->font_color.valid) {
-		/* Try to propagate background color from parent nodes */
-		for (parent = tag->parent; parent != NULL; parent = parent->parent) {
-			parent_tag = parent->data;
-
-			if (parent_tag && (parent_tag->flags & FL_BLOCK) && parent_tag->extra) {
-				bl_parent = parent_tag->extra;
-
-				if (bl_parent->font_color.valid) {
-					memcpy (&bl->font_color, &bl_parent->font_color,
-							sizeof (bl->font_color));
-					break;
-				}
-			}
-		}
-	}
-
-	/* Set bgcolor to the html bgcolor and font color to black as a last resort */
-	if (!bl->font_color.valid) {
-		bl->font_color.d.val = 0;
-		bl->font_color.d.comp.alpha = 255;
-		bl->font_color.valid = TRUE;
-	}
-	if (!bl->background_color.valid) {
-		memcpy (&bl->background_color, &hc->bgcolor, sizeof (hc->bgcolor));
-	}
-
 	if (hc->blocks == NULL) {
 		hc->blocks = g_ptr_array_sized_new (64);
 		rspamd_mempool_add_destructor (pool, rspamd_ptr_array_free_hard,
@@ -2209,6 +2215,104 @@ rspamd_html_check_displayed_url (rspamd_mempool_t *pool,
 	}
 }
 
+static gboolean
+rspamd_html_propagate_lengths (GNode *node, gpointer _unused)
+{
+	GNode *child;
+	struct html_tag *tag = node->data, *cld_tag;
+
+	if (tag) {
+		child = node->children;
+
+		/* Summarize content length from children */
+		while (child) {
+			cld_tag = child->data;
+			tag->content_length += cld_tag->content_length;
+			child = child->next;
+		}
+	}
+
+	return FALSE;
+}
+
+static void
+rspamd_html_propagate_style (struct html_content *hc,
+							 struct html_tag *tag,
+							 struct html_block *bl,
+							 GQueue *blocks)
+{
+	struct html_block *bl_parent;
+	gboolean push_block = FALSE;
+
+
+	/* Propagate from the parent if needed */
+	bl_parent = g_queue_peek_tail (blocks);
+
+	if (bl_parent) {
+		if (!bl->background_color.valid) {
+			/* Try to propagate background color from parent nodes */
+			if (bl_parent->background_color.valid) {
+				memcpy (&bl->background_color, &bl_parent->background_color,
+						sizeof (bl->background_color));
+			}
+		}
+		else {
+			push_block = TRUE;
+		}
+
+		if (!bl->font_color.valid) {
+			/* Try to propagate background color from parent nodes */
+			if (bl_parent->font_color.valid) {
+				memcpy (&bl->font_color, &bl_parent->font_color,
+						sizeof (bl->font_color));
+			}
+		}
+		else {
+			push_block = TRUE;
+		}
+
+		/* Propagate font size */
+		if (bl->font_size == (guint)-1) {
+			if (bl_parent->font_size != (guint)-1) {
+				bl->font_size = bl_parent->font_size;
+			}
+		}
+		else {
+			push_block = TRUE;
+		}
+	}
+
+	/* Set bgcolor to the html bgcolor and font color to black as a last resort */
+	if (!bl->font_color.valid) {
+		/* Don't touch opacity as it can be set separately */
+		bl->font_color.d.comp.r = 0;
+		bl->font_color.d.comp.g = 0;
+		bl->font_color.d.comp.b = 0;
+		bl->font_color.valid = TRUE;
+	}
+	else {
+		push_block = TRUE;
+	}
+
+	if (!bl->background_color.valid) {
+		memcpy (&bl->background_color, &hc->bgcolor, sizeof (hc->bgcolor));
+	}
+	else {
+		push_block = TRUE;
+	}
+
+	if (bl->font_size == (guint)-1) {
+		bl->font_size = 16; /* Default for browsers */
+	}
+	else {
+		push_block = TRUE;
+	}
+
+	if (push_block && !(tag->flags & FL_CLOSED)) {
+		g_queue_push_tail (blocks, bl);
+	}
+}
+
 GByteArray*
 rspamd_html_process_part_full (rspamd_mempool_t *pool, struct html_content *hc,
 		GByteArray *in, GList **exceptions, GHashTable *urls,  GHashTable *emails)
@@ -2224,6 +2328,8 @@ rspamd_html_process_part_full (rspamd_mempool_t *pool, struct html_content *hc,
 	gint substate = 0, len, href_offset = -1;
 	struct html_tag *cur_tag = NULL, *content_tag = NULL;
 	struct rspamd_url *url = NULL, *turl;
+	GQueue *styles_blocks;
+
 	enum {
 		parse_start = 0,
 		tag_begin,
@@ -2256,6 +2362,7 @@ rspamd_html_process_part_full (rspamd_mempool_t *pool, struct html_content *hc,
 	hc->bgcolor.valid = TRUE;
 
 	dest = g_byte_array_sized_new (in->len / 3 * 2);
+	styles_blocks = g_queue_new ();
 
 	p = in->data;
 	c = p;
@@ -2399,6 +2506,13 @@ rspamd_html_process_part_full (rspamd_mempool_t *pool, struct html_content *hc,
 				p ++;
 			}
 			else {
+				if (content_tag) {
+					if (content_tag->content == NULL) {
+						content_tag->content = c;
+					}
+
+					content_tag->content_length += p - c;
+				}
 				state = tag_begin;
 			}
 			break;
@@ -2570,10 +2684,10 @@ rspamd_html_process_part_full (rspamd_mempool_t *pool, struct html_content *hc,
 					}
 					save_space = FALSE;
 				}
-				else if ((cur_tag->flags & (FL_CLOSED|FL_CLOSING)) &&
-						(cur_tag->id == Tag_P ||
+
+				if ((cur_tag->id == Tag_P ||
 						cur_tag->id == Tag_TR ||
-						cur_tag->id == Tag_DIV) && balanced) {
+						cur_tag->id == Tag_DIV)) {
 					if (dest->len > 0 && dest->data[dest->len - 1] != '\n') {
 						g_byte_array_append (dest, "\r\n", 2);
 					}
@@ -2582,7 +2696,7 @@ rspamd_html_process_part_full (rspamd_mempool_t *pool, struct html_content *hc,
 
 				if (cur_tag->id == Tag_A || cur_tag->id == Tag_IFRAME) {
 					if (!(cur_tag->flags & (FL_CLOSING))) {
-						url = rspamd_html_process_url_tag (pool, cur_tag);
+						url = rspamd_html_process_url_tag (pool, cur_tag, hc);
 
 						if (url != NULL) {
 
@@ -2651,21 +2765,62 @@ rspamd_html_process_part_full (rspamd_mempool_t *pool, struct html_content *hc,
 					}
 				}
 				else if (cur_tag->id == Tag_LINK) {
-					url = rspamd_html_process_url_tag (pool, cur_tag);
+					url = rspamd_html_process_url_tag (pool, cur_tag, hc);
+				}
+				else if (cur_tag->id == Tag_BASE && !(cur_tag->flags & (FL_CLOSING))) {
+					struct html_tag *prev_tag = NULL;
+
+					if (cur_level && cur_level->parent) {
+						prev_tag = cur_level->parent->data;
+					}
+
+					/*
+					 * Base is allowed only within head tag but we slightly
+					 * relax that
+					 */
+					if (!prev_tag || prev_tag->id == Tag_HEAD ||
+						prev_tag->id == Tag_HTML) {
+						url = rspamd_html_process_url_tag (pool, cur_tag, hc);
+
+						if (url != NULL && hc->base_url == NULL) {
+							/* We have a base tag available */
+							hc->base_url = url;
+						}
+					}
 				}
 
 				if (cur_tag->id == Tag_IMG && !(cur_tag->flags & FL_CLOSING)) {
 					rspamd_html_process_img_tag (pool, cur_tag, hc);
 				}
-				else if (!(cur_tag->flags & FL_CLOSING) &&
-						(cur_tag->flags & FL_BLOCK)) {
+				else if (cur_tag->flags & FL_BLOCK) {
 					struct html_block *bl;
 
-					rspamd_html_process_block_tag (pool, cur_tag, hc);
-					bl = cur_tag->extra;
+					if (cur_tag->flags & FL_CLOSING) {
+						/* Just remove block element from the queue if any */
+						if (styles_blocks->length > 0) {
+							g_queue_pop_tail (styles_blocks);
+						}
+					}
+					else {
+						rspamd_html_process_block_tag (pool, cur_tag, hc);
+						bl = cur_tag->extra;
 
-					if (bl && !bl->visible) {
-						state = content_ignore;
+						if (bl) {
+							rspamd_html_propagate_style (hc, cur_tag,
+									cur_tag->extra, styles_blocks);
+
+							/* Check visibility */
+							if (bl->font_size < 3 ||
+								bl->font_color.d.comp.alpha < 10) {
+
+								bl->visible = FALSE;
+								msg_debug_html ("tag is not visible");
+							}
+
+							if (!bl->visible) {
+								state = content_ignore;
+							}
+						}
 					}
 				}
 			}
@@ -2681,6 +2836,13 @@ rspamd_html_process_part_full (rspamd_mempool_t *pool, struct html_content *hc,
 		}
 	}
 
+	if (hc->html_tags) {
+		g_node_traverse (hc->html_tags, G_POST_ORDER, G_TRAVERSE_ALL, -1,
+				rspamd_html_propagate_lengths, NULL);
+	}
+
+	g_queue_free (styles_blocks);
+
 	return dest;
 }
 
