@@ -69,6 +69,7 @@ rspamd_task_new (struct rspamd_worker *worker, struct rspamd_config *cfg,
 
 	new_task = g_malloc0 (sizeof (struct rspamd_task));
 	new_task->worker = worker;
+	new_task->lang_det = lang_det;
 
 	if (cfg) {
 		new_task->cfg = cfg;
@@ -78,7 +79,14 @@ rspamd_task_new (struct rspamd_worker *worker, struct rspamd_config *cfg,
 			new_task->flags |= RSPAMD_TASK_FLAG_PASS_ALL;
 		}
 
-		new_task->re_rt = rspamd_re_cache_runtime_new (cfg->re_cache);
+
+		if (cfg->re_cache) {
+			new_task->re_rt = rspamd_re_cache_runtime_new (cfg->re_cache);
+		}
+
+		if (new_task->lang_det == NULL && cfg->lang_det != NULL) {
+			new_task->lang_det = cfg->lang_det;
+		}
 	}
 
 	gettimeofday (&new_task->tv, NULL);
@@ -86,7 +94,6 @@ rspamd_task_new (struct rspamd_worker *worker, struct rspamd_config *cfg,
 	new_task->time_virtual = rspamd_get_virtual_ticks ();
 	new_task->time_real_finish = NAN;
 	new_task->time_virtual_finish = NAN;
-	new_task->lang_det = lang_det;
 
 	if (pool == NULL) {
 		new_task->task_pool =
@@ -235,18 +242,18 @@ rspamd_task_free (struct rspamd_task *task)
 		for (i = 0; i < task->text_parts->len; i ++) {
 			tp = g_ptr_array_index (task->text_parts, i);
 
-			if (tp->normalized_words) {
-				g_array_free (tp->normalized_words, TRUE);
+			if (tp->utf_words) {
+				g_array_free (tp->utf_words, TRUE);
 			}
 			if (tp->normalized_hashes) {
 				g_array_free (tp->normalized_hashes, TRUE);
 			}
-			if (tp->ucs32_words) {
-				g_array_free (tp->ucs32_words, TRUE);
-			}
 			if (tp->languages) {
 				g_ptr_array_unref (tp->languages);
 			}
+			if (tp->unicode_raw_content) {
+				g_array_free (tp->unicode_raw_content, TRUE);
+			}
 		}
 
 		if (task->rcpt_envelope) {
@@ -263,7 +270,10 @@ rspamd_task_free (struct rspamd_task *task)
 		}
 
 		ucl_object_unref (task->messages);
-		rspamd_re_cache_runtime_destroy (task->re_rt);
+
+		if (task->re_rt) {
+			rspamd_re_cache_runtime_destroy (task->re_rt);
+		}
 
 		if (task->http_conn != NULL) {
 			rspamd_http_connection_reset (task->http_conn);
@@ -738,6 +748,12 @@ rspamd_task_process (struct rspamd_task *task, guint stages)
 				RSPAMD_TASK_STAGE_PRE_FILTERS);
 		break;
 
+	case RSPAMD_TASK_STAGE_PROCESS_MESSAGE:
+		if (!(task->flags & RSPAMD_TASK_FLAG_SKIP_PROCESS)) {
+			rspamd_message_process (task);
+		}
+		break;
+
 	case RSPAMD_TASK_STAGE_FILTERS:
 		rspamd_symbols_cache_process_symbols (task, task->cfg->cache,
 				RSPAMD_TASK_STAGE_FILTERS);
@@ -825,7 +841,8 @@ rspamd_task_process (struct rspamd_task *task, guint stages)
 	}
 
 	if (RSPAMD_TASK_IS_SKIPPED (task)) {
-		task->processed_stages |= RSPAMD_TASK_STAGE_DONE;
+		/* Set all bits except idempotent filters */
+		task->processed_stages |= 0x7FFF;
 	}
 
 	task->flags &= ~RSPAMD_TASK_FLAG_PROCESSING;
@@ -1029,9 +1046,7 @@ rspamd_task_log_metric_res (struct rspamd_task *task,
 	static gchar scorebuf[32];
 	rspamd_ftok_t res = {.begin = NULL, .len = 0};
 	struct rspamd_metric_result *mres;
-	GHashTableIter it;
 	gboolean first = TRUE;
-	gpointer k, v;
 	rspamd_fstring_t *symbuf;
 	struct rspamd_symbol_result *sym;
 	GPtrArray *sorted_symbols;
@@ -1067,12 +1082,13 @@ rspamd_task_log_metric_res (struct rspamd_task *task,
 			break;
 		case RSPAMD_LOG_SYMBOLS:
 			symbuf = rspamd_fstring_sized_new (128);
-			g_hash_table_iter_init (&it, mres->symbols);
-			sorted_symbols = g_ptr_array_sized_new (g_hash_table_size (mres->symbols));
+			sorted_symbols = g_ptr_array_sized_new (kh_size (mres->symbols));
 
-			while (g_hash_table_iter_next (&it, &k, &v)) {
-				g_ptr_array_add (sorted_symbols, v);
-			}
+			kh_foreach_value_ptr (mres->symbols, sym, {
+				if (!(sym->flags & RSPAMD_SYMBOL_RESULT_IGNORED)) {
+					g_ptr_array_add (sorted_symbols, (gpointer)sym);
+				}
+			});
 
 			g_ptr_array_sort (sorted_symbols, rspamd_task_compare_log_sym);
 
