@@ -41,25 +41,27 @@ enum rspamd_task_stage {
 	RSPAMD_TASK_STAGE_ENVELOPE = (1 << 1),
 	RSPAMD_TASK_STAGE_READ_MESSAGE = (1 << 2),
 	RSPAMD_TASK_STAGE_PRE_FILTERS = (1 << 3),
-	RSPAMD_TASK_STAGE_FILTERS = (1 << 4),
-	RSPAMD_TASK_STAGE_CLASSIFIERS_PRE = (1 << 5),
-	RSPAMD_TASK_STAGE_CLASSIFIERS = (1 << 6),
-	RSPAMD_TASK_STAGE_CLASSIFIERS_POST = (1 << 7),
-	RSPAMD_TASK_STAGE_COMPOSITES = (1 << 8),
-	RSPAMD_TASK_STAGE_POST_FILTERS = (1 << 9),
-	RSPAMD_TASK_STAGE_LEARN_PRE = (1 << 10),
-	RSPAMD_TASK_STAGE_LEARN = (1 << 11),
-	RSPAMD_TASK_STAGE_LEARN_POST = (1 << 12),
-	RSPAMD_TASK_STAGE_COMPOSITES_POST = (1 << 13),
-	RSPAMD_TASK_STAGE_IDEMPOTENT = (1 << 14),
-	RSPAMD_TASK_STAGE_DONE = (1 << 15),
-	RSPAMD_TASK_STAGE_REPLIED = (1 << 16)
+	RSPAMD_TASK_STAGE_PROCESS_MESSAGE = (1 << 4),
+	RSPAMD_TASK_STAGE_FILTERS = (1 << 5),
+	RSPAMD_TASK_STAGE_CLASSIFIERS_PRE = (1 << 6),
+	RSPAMD_TASK_STAGE_CLASSIFIERS = (1 << 7),
+	RSPAMD_TASK_STAGE_CLASSIFIERS_POST = (1 << 8),
+	RSPAMD_TASK_STAGE_COMPOSITES = (1 << 9),
+	RSPAMD_TASK_STAGE_POST_FILTERS = (1 << 10),
+	RSPAMD_TASK_STAGE_LEARN_PRE = (1 << 11),
+	RSPAMD_TASK_STAGE_LEARN = (1 << 12),
+	RSPAMD_TASK_STAGE_LEARN_POST = (1 << 13),
+	RSPAMD_TASK_STAGE_COMPOSITES_POST = (1 << 14),
+	RSPAMD_TASK_STAGE_IDEMPOTENT = (1 << 15),
+	RSPAMD_TASK_STAGE_DONE = (1 << 16),
+	RSPAMD_TASK_STAGE_REPLIED = (1 << 17)
 };
 
 #define RSPAMD_TASK_PROCESS_ALL (RSPAMD_TASK_STAGE_CONNECT | \
 		RSPAMD_TASK_STAGE_ENVELOPE | \
 		RSPAMD_TASK_STAGE_READ_MESSAGE | \
 		RSPAMD_TASK_STAGE_PRE_FILTERS | \
+		RSPAMD_TASK_STAGE_PROCESS_MESSAGE | \
 		RSPAMD_TASK_STAGE_FILTERS | \
 		RSPAMD_TASK_STAGE_CLASSIFIERS_PRE | \
 		RSPAMD_TASK_STAGE_CLASSIFIERS | \
@@ -75,6 +77,7 @@ enum rspamd_task_stage {
 #define RSPAMD_TASK_PROCESS_LEARN (RSPAMD_TASK_STAGE_CONNECT | \
 		RSPAMD_TASK_STAGE_ENVELOPE | \
 		RSPAMD_TASK_STAGE_READ_MESSAGE | \
+		RSPAMD_TASK_STAGE_PROCESS_MESSAGE | \
 		RSPAMD_TASK_STAGE_CLASSIFIERS_PRE | \
 		RSPAMD_TASK_STAGE_CLASSIFIERS | \
 		RSPAMD_TASK_STAGE_CLASSIFIERS_POST | \
@@ -85,7 +88,7 @@ enum rspamd_task_stage {
 
 #define RSPAMD_TASK_FLAG_MIME (1 << 0)
 #define RSPAMD_TASK_FLAG_JSON (1 << 1)
-#define RSPAMD_TASK_FLAG_SKIP_EXTRA (1 << 2)
+#define RSPAMD_TASK_FLAG_SKIP_PROCESS (1 << 2)
 #define RSPAMD_TASK_FLAG_SKIP (1 << 3)
 #define RSPAMD_TASK_FLAG_EXT_URLS (1 << 4)
 #define RSPAMD_TASK_FLAG_SPAMC (1 << 5)
