@@ -95,7 +95,7 @@ rspamd_redis_cache_timeout (gint fd, short what, gpointer d)
 
 	msg_err_task ("connection to redis server %s timed out",
 			rspamd_upstream_name (rt->selected));
-	rspamd_upstream_fail (rt->selected);
+	rspamd_upstream_fail (rt->selected, FALSE);
 
 	if (rt->has_event) {
 		rspamd_session_remove_event (task->s, rspamd_redis_cache_fin, d);
@@ -147,7 +147,7 @@ rspamd_stat_cache_redis_get (redisAsyncContext *c, gpointer r, gpointer priv)
 		rspamd_upstream_ok (rt->selected);
 	}
 	else {
-		rspamd_upstream_fail (rt->selected);
+		rspamd_upstream_fail (rt->selected, FALSE);
 	}
 
 	if (rt->has_event) {
@@ -169,7 +169,7 @@ rspamd_stat_cache_redis_set (redisAsyncContext *c, gpointer r, gpointer priv)
 		rspamd_upstream_ok (rt->selected);
 	}
 	else {
-		rspamd_upstream_fail (rt->selected);
+		rspamd_upstream_fail (rt->selected, FALSE);
 	}
 
 	if (rt->has_event) {
@@ -438,6 +438,10 @@ rspamd_stat_cache_redis_check (struct rspamd_task *task,
 	struct timeval tv;
 	gchar *h;
 
+	if (rspamd_session_blocked (task->s)) {
+		return RSPAMD_LEARN_INGORE;
+	}
+
 	h = rspamd_mempool_get_variable (task->task_pool, "words_hash");
 
 	if (h == NULL) {
@@ -449,8 +453,7 @@ rspamd_stat_cache_redis_check (struct rspamd_task *task,
 	if (redisAsyncCommand (rt->redis, rspamd_stat_cache_redis_get, rt,
 			"HGET %s %s",
 			rt->ctx->redis_object, h) == REDIS_OK) {
-		rspamd_session_add_event (task->s, rspamd_redis_cache_fin, rt,
-				rspamd_stat_cache_redis_quark ());
+		rspamd_session_add_event (task->s, NULL, rspamd_redis_cache_fin, rt, rspamd_stat_cache_redis_quark ());
 		event_add (&rt->timeout_event, &tv);
 		rt->has_event = TRUE;
 	}
@@ -469,6 +472,10 @@ rspamd_stat_cache_redis_learn (struct rspamd_task *task,
 	gchar *h;
 	gint flag;
 
+	if (rspamd_session_blocked (task->s)) {
+		return RSPAMD_LEARN_INGORE;
+	}
+
 	h = rspamd_mempool_get_variable (task->task_pool, "words_hash");
 	g_assert (h != NULL);
 
@@ -478,8 +485,7 @@ rspamd_stat_cache_redis_learn (struct rspamd_task *task,
 	if (redisAsyncCommand (rt->redis, rspamd_stat_cache_redis_set, rt,
 			"HSET %s %s %d",
 			rt->ctx->redis_object, h, flag) == REDIS_OK) {
-		rspamd_session_add_event (task->s, rspamd_redis_cache_fin, rt,
-				rspamd_stat_cache_redis_quark ());
+		rspamd_session_add_event (task->s, NULL, rspamd_redis_cache_fin, rt, rspamd_stat_cache_redis_quark ());
 		event_add (&rt->timeout_event, &tv);
 		rt->has_event = TRUE;
 	}
