@@ -40,21 +40,23 @@
 #define MAP_RELEASE(x, t) REF_RELEASE(x)
 #endif
 
-static void free_http_cbdata_common (struct http_callback_data *cbd, gboolean plan_new);
+static void free_http_cbdata_common (struct http_callback_data *cbd,
+									 gboolean plan_new);
 static void free_http_cbdata_dtor (gpointer p);
 static void free_http_cbdata (struct http_callback_data *cbd);
 static void rspamd_map_periodic_callback (gint fd, short what, void *ud);
 static void rspamd_map_schedule_periodic (struct rspamd_map *map, gboolean locked,
-		gboolean initial, gboolean errored);
-
-struct rspamd_http_map_cached_cbdata {
-	struct event timeout;
-	struct rspamd_storage_shmem *shm;
-	struct rspamd_map *map;
-	struct http_map_data *data;
-	guint64 gen;
-	time_t last_checked;
-};
+										  gboolean initial, gboolean errored);
+static gboolean read_map_file_chunks (struct rspamd_map *map,
+									  struct map_cb_data *cbdata,
+									  const gchar *fname,
+									  gsize len,
+									  goffset off);
+static gboolean rspamd_map_save_http_cached_file (struct rspamd_map *map,
+												  struct rspamd_map_backend *bk,
+												  struct http_map_data *htdata,
+												  const guchar *data,
+												  gsize len);
 
 guint rspamd_map_log_id = (guint)-1;
 RSPAMD_CONSTRUCTOR(rspamd_map_log_init)
@@ -150,7 +152,7 @@ rspamd_map_check_sig_pk_mem (const guchar *sig,
 		ret = FALSE;
 	}
 
-	if (ret && !rspamd_cryptobox_verify (sig, input, inlen,
+	if (ret && !rspamd_cryptobox_verify (sig, siglen, input, inlen,
 			rspamd_pubkey_get_pk (pk, NULL), RSPAMD_CRYPTOBOX_MODE_25519)) {
 		msg_err_map ("can't verify signature for %s: incorrect signature", map->name);
 
@@ -285,8 +287,8 @@ free_http_cbdata_common (struct http_callback_data *cbd, gboolean plan_new)
 static void
 free_http_cbdata (struct http_callback_data *cbd)
 {
-	cbd->map->dtor = NULL;
-	cbd->map->dtor_data = NULL;
+	cbd->map->tmp_dtor = NULL;
+	cbd->map->tmp_dtor_data = NULL;
 
 	free_http_cbdata_common (cbd, TRUE);
 }
@@ -338,15 +340,21 @@ rspamd_map_cache_cb (gint fd, short what, gpointer ud)
 {
 	struct rspamd_http_map_cached_cbdata *cache_cbd = ud;
 	struct rspamd_map *map;
+	struct http_map_data *data;
 	struct timeval tv;
 
 	map = cache_cbd->map;
+	data = cache_cbd->data;
 
 	if (cache_cbd->gen != cache_cbd->data->gen) {
 		/* We have another update, so this cache element is obviously expired */
-		/* Important: we do not set cache availability to zero here */
+		/*
+		 * Important!: we do not set cache availability to zero here, as there
+		 * might be fresh cache
+		 */
+		msg_info_map ("cached data is now expired (gen mismatch %L != %L) for %s",
+				cache_cbd->gen, cache_cbd->data->gen, map->name);
 		MAP_RELEASE (cache_cbd->shm, "rspamd_http_map_cached_cbdata");
-		msg_info_map ("cached data is now expired (gen mismatch) for %s", map->name);
 		event_del (&cache_cbd->timeout);
 		g_free (cache_cbd);
 	}
@@ -361,7 +369,8 @@ rspamd_map_cache_cb (gint fd, short what, gpointer ud)
 		event_add (&cache_cbd->timeout, &tv);
 	}
 	else {
-		g_atomic_int_set (&map->cache->available, 0);
+		data->cur_cache_cbd = NULL;
+		g_atomic_int_set (&data->cache->available, 0);
 		MAP_RELEASE (cache_cbd->shm, "rspamd_http_map_cached_cbdata");
 		msg_info_map ("cached data is now expired for %s", map->name);
 		event_del (&cache_cbd->timeout);
@@ -431,6 +440,7 @@ http_map_finish (struct rspamd_http_connection *conn,
 	struct http_callback_data *cbd = conn->ud;
 	struct rspamd_map *map;
 	struct rspamd_map_backend *bk;
+	struct http_map_data *data;
 	struct rspamd_http_map_cached_cbdata *cache_cbd;
 	struct timeval tv;
 	const rspamd_ftok_t *expires_hdr, *etag_hdr;
@@ -440,6 +450,7 @@ http_map_finish (struct rspamd_http_connection *conn,
 
 	map = cbd->map;
 	bk = cbd->bk;
+	data = bk->data.hd;
 
 	if (msg->code == 200) {
 
@@ -447,6 +458,10 @@ http_map_finish (struct rspamd_http_connection *conn,
 			cbd->periodic->need_modify = TRUE;
 			/* Reset the whole chain */
 			cbd->periodic->cur_backend = 0;
+			/* Reset cache, old cached data will be cleaned on timeout */
+			g_atomic_int_set (&data->cache->available, 0);
+			data->cur_cache_cbd = NULL;
+
 			rspamd_map_periodic_callback (-1, EV_TIMEOUT, cbd->periodic);
 			MAP_RELEASE (cbd, "http_callback_data");
 
@@ -652,12 +667,12 @@ read_data:
 		/*
 		 * We know that a map is in the locked state
 		 */
-		g_atomic_int_set (&map->cache->available, 1);
+		g_atomic_int_set (&data->cache->available, 1);
 		/* Store cached data */
-		rspamd_strlcpy (map->cache->shmem_name, cbd->shmem_data->shm_name,
-				sizeof (map->cache->shmem_name));
-		map->cache->len = cbd->data_len;
-		map->cache->last_modified = cbd->data->last_modified;
+		rspamd_strlcpy (data->cache->shmem_name, cbd->shmem_data->shm_name,
+				sizeof (data->cache->shmem_name));
+		data->cache->len = cbd->data_len;
+		data->cache->last_modified = cbd->data->last_modified;
 		cache_cbd = g_malloc0 (sizeof (*cache_cbd));
 		cache_cbd->shm = cbd->shmem_data;
 		cache_cbd->map = map;
@@ -670,6 +685,7 @@ read_data:
 				cache_cbd);
 		event_base_set (cbd->ev_base, &cache_cbd->timeout);
 		event_add (&cache_cbd->timeout, &tv);
+		data->cur_cache_cbd = cache_cbd;
 
 		if (map->next_check) {
 			rspamd_http_date_format (next_check_date, sizeof (next_check_date),
@@ -734,6 +750,7 @@ read_data:
 					rspamd_inet_address_to_string_pretty (cbd->addr),
 					dlen, zout.pos, next_check_date);
 			map->read_callback (out, zout.pos, &cbd->periodic->cbdata, TRUE);
+			rspamd_map_save_http_cached_file (map, bk, cbd->data, out, zout.pos);
 			g_free (out);
 		}
 		else {
@@ -741,6 +758,7 @@ read_data:
 					cbd->bk->uri,
 					rspamd_inet_address_to_string_pretty (cbd->addr),
 					dlen, next_check_date);
+			rspamd_map_save_http_cached_file (map, bk, cbd->data, in, cbd->data_len);
 			map->read_callback (in, cbd->data_len, &cbd->periodic->cbdata, TRUE);
 		}
 
@@ -819,6 +837,80 @@ err:
 	return 0;
 }
 
+static gboolean
+read_map_file_chunks (struct rspamd_map *map, struct map_cb_data *cbdata,
+		const gchar *fname, gsize len, goffset off)
+{
+	gint fd;
+	gssize r, avail;
+	gsize buflen = 1024 * 1024;
+	gchar *pos, *bytes;
+
+	fd = rspamd_file_xopen (fname, O_RDONLY, 0, TRUE);
+
+	if (fd == -1) {
+		msg_err_map ("can't open map for buffered reading %s: %s",
+				fname, strerror (errno));
+		return FALSE;
+	}
+
+	if (lseek (fd, off, SEEK_SET) == -1) {
+		msg_err_map ("can't seek in map to pos %d for buffered reading %s: %s",
+				(gint)off, fname, strerror (errno));
+		return FALSE;
+	}
+
+	buflen = MIN (len, buflen);
+	bytes = g_malloc (buflen);
+	avail = buflen;
+	pos = bytes;
+
+	while ((r = read (fd, pos, avail)) > 0) {
+		gchar *end = bytes + (pos - bytes) + r;
+		msg_info_map ("%s: read map chunk, %z bytes", fname,
+				r);
+		pos = map->read_callback (bytes, end - bytes, cbdata, r == len);
+
+		if (pos && pos > bytes && pos < end) {
+			guint remain = end - pos;
+
+			memmove (bytes, pos, remain);
+			pos = bytes + remain;
+			/* Need to preserve the remain */
+			avail = ((gssize)buflen) - remain;
+
+			if (avail <= 0) {
+				/* Try realloc, too large element */
+				g_assert (buflen >= remain);
+				bytes = g_realloc (bytes, buflen * 2);
+
+				pos = bytes + remain; /* Adjust */
+				avail += buflen;
+				buflen *= 2;
+			}
+		}
+		else {
+			avail = buflen;
+			pos = bytes;
+		}
+
+		len -= r;
+	}
+
+	if (r == -1) {
+		msg_err_map ("can't read from map %s: %s", fname, strerror (errno));
+		close (fd);
+		g_free (bytes);
+
+		return FALSE;
+	}
+
+	close (fd);
+	g_free (bytes);
+
+	return TRUE;
+}
+
 /**
  * Callback for reading data from file
  */
@@ -826,8 +918,9 @@ static gboolean
 read_map_file (struct rspamd_map *map, struct file_map_data *data,
 		struct rspamd_map_backend *bk, struct map_periodic_cbdata *periodic)
 {
-	guchar *bytes;
+	gchar *bytes;
 	gsize len;
+	struct stat st;
 
 	if (map->read_callback == NULL || map->fin_callback == NULL) {
 		msg_err_map ("%s: bad callback for reading map file",
@@ -835,7 +928,7 @@ read_map_file (struct rspamd_map *map, struct file_map_data *data,
 		return FALSE;
 	}
 
-	if (access (data->filename, R_OK) == -1) {
+	if (stat (data->filename, &st) == -1) {
 		/* File does not exist, skipping */
 		if (errno != ENOENT) {
 			msg_err_map ("%s: map file is unavailable for reading: %s",
@@ -844,29 +937,41 @@ read_map_file (struct rspamd_map *map, struct file_map_data *data,
 			return FALSE;
 		}
 		else {
-			msg_info_map ("%s: map file is not found",
+			msg_info_map ("%s: map file is not found; "
+						  "it will be read automatically if created",
 					data->filename);
 			return TRUE;
 		}
 	}
 
-	bytes = rspamd_file_xmap (data->filename, PROT_READ, &len, TRUE);
-
-	if (bytes == NULL) {
-		msg_err_map ("can't open map %s: %s", data->filename, strerror (errno));
-		return FALSE;
-	}
+	len = st.st_size;
 
 	if (bk->is_signed) {
+		bytes = rspamd_file_xmap (data->filename, PROT_READ, &len, TRUE);
+
+		if (bytes == NULL) {
+			msg_err_map ("can't open map %s: %s", data->filename, strerror (errno));
+			return FALSE;
+		}
+
 		if (!rspamd_map_check_file_sig (data->filename, map, bk, bytes, len)) {
 			munmap (bytes, len);
 
 			return FALSE;
 		}
+
+		munmap (bytes, len);
 	}
 
 	if (len > 0) {
 		if (bk->is_compressed) {
+			bytes = rspamd_file_xmap (data->filename, PROT_READ, &len, TRUE);
+
+			if (bytes == NULL) {
+				msg_err_map ("can't open map %s: %s", data->filename, strerror (errno));
+				return FALSE;
+			}
+
 			ZSTD_DStream *zstream;
 			ZSTD_inBuffer zin;
 			ZSTD_outBuffer zout;
@@ -917,18 +1022,24 @@ read_map_file (struct rspamd_map *map, struct file_map_data *data,
 					len, zout.pos);
 			map->read_callback (out, zout.pos, &periodic->cbdata, TRUE);
 			g_free (out);
+
+			munmap (bytes, len);
 		}
 		else {
-			msg_info_map ("%s: read map dat, %z bytes", data->filename,
-					len);
-			map->read_callback (bytes, len, &periodic->cbdata, TRUE);
+			/* Perform buffered read: fail-safe */
+			if (!read_map_file_chunks (map, &periodic->cbdata, data->filename,
+					len, 0)) {
+				return FALSE;
+			}
 		}
 	}
 	else {
+		/* Empty map */
 		map->read_callback (NULL, 0, &periodic->cbdata, TRUE);
 	}
 
-	munmap (bytes, len);
+	/* Also update at the read time */
+	memcpy (&data->st, &st, sizeof (struct stat));
 
 	return TRUE;
 }
@@ -1057,7 +1168,7 @@ rspamd_map_schedule_periodic (struct rspamd_map *map,
 	gdouble timeout;
 	struct map_periodic_cbdata *cbd;
 
-	if (map->scheduled_check) {
+	if (map->scheduled_check || (map->wrk && map->wrk->wanna_die)) {
 		/* Do not schedule check if some check is already scheduled */
 		return;
 	}
@@ -1189,20 +1300,24 @@ rspamd_map_read_cached (struct rspamd_map *map, struct rspamd_map_backend *bk,
 {
 	gsize len;
 	gpointer in;
+	struct http_map_data *data;
 
-	in = rspamd_shmem_xmap (map->cache->shmem_name, PROT_READ, &len);
+	data = bk->data.hd;
+
+	in = rspamd_shmem_xmap (data->cache->shmem_name, PROT_READ, &len);
 
 	if (in == NULL) {
-		msg_err ("cannot map cache from %s: %s", map->cache->shmem_name,
+		msg_err ("cannot map cache from %s: %s", data->cache->shmem_name,
 				strerror (errno));
 		return FALSE;
 	}
 
-	if (len < map->cache->len) {
+	if (len < data->cache->len) {
 		msg_err ("cannot map cache from %s: bad length %z, %z expected",
-				map->cache->shmem_name,
-				len, map->cache->len);
+				data->cache->shmem_name,
+				len, data->cache->len);
 		munmap (in, len);
+
 		return FALSE;
 	}
 
@@ -1269,6 +1384,168 @@ rspamd_map_read_cached (struct rspamd_map *map, struct rspamd_map_backend *bk,
 	return TRUE;
 }
 
+static gboolean
+rspamd_map_has_http_cached_file (struct rspamd_map *map,
+								 struct rspamd_map_backend *bk)
+{
+	gchar path[PATH_MAX];
+	guchar digest[rspamd_cryptobox_HASHBYTES];
+	struct rspamd_config *cfg = map->cfg;
+	struct stat st;
+
+	if (cfg->maps_cache_dir == NULL || cfg->maps_cache_dir[0] == '\0') {
+		return FALSE;
+	}
+
+	rspamd_cryptobox_hash (digest, bk->uri, strlen (bk->uri), NULL, 0);
+	rspamd_snprintf (path, sizeof (path), "%s%c%*xs.map", cfg->maps_cache_dir,
+			G_DIR_SEPARATOR, 20, digest);
+
+	if (stat (path, &st) != -1 && st.st_size >
+								  sizeof (struct rspamd_http_file_data)) {
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
+static gboolean
+rspamd_map_save_http_cached_file (struct rspamd_map *map,
+								  struct rspamd_map_backend *bk,
+								  struct http_map_data *htdata,
+								  const guchar *data,
+								  gsize len)
+{
+	gchar path[PATH_MAX];
+	guchar digest[rspamd_cryptobox_HASHBYTES];
+	struct rspamd_config *cfg = map->cfg;
+	gint fd;
+	struct rspamd_http_file_data header;
+
+	if (cfg->maps_cache_dir == NULL || cfg->maps_cache_dir[0] == '\0') {
+		return FALSE;
+	}
+
+	rspamd_cryptobox_hash (digest, bk->uri, strlen (bk->uri), NULL, 0);
+	rspamd_snprintf (path, sizeof (path), "%s%c%*xs.map", cfg->maps_cache_dir,
+			G_DIR_SEPARATOR, 20, digest);
+
+	fd = rspamd_file_xopen (path, O_WRONLY | O_TRUNC | O_CREAT,
+			00600, FALSE);
+
+	if (fd == -1) {
+		return FALSE;
+	}
+
+	if (!rspamd_file_lock (fd, FALSE)) {
+		msg_err_map ("cannot lock file %s: %s", path, strerror (errno));
+		close (fd);
+
+		return FALSE;
+	}
+
+	memcpy (header.magic, rspamd_http_file_magic, sizeof (rspamd_http_file_magic));
+	header.mtime = htdata->last_modified;
+	header.next_check = map->next_check;
+	header.data_off = sizeof (header);
+
+	if (write (fd, &header, sizeof (header)) != sizeof (header)) {
+		msg_err_map ("cannot write file %s: %s", path, strerror (errno));
+		rspamd_file_unlock (fd, FALSE);
+		close (fd);
+
+		return FALSE;
+	}
+
+	/* Now write the rest */
+	if (write (fd, data, len) != len) {
+		msg_err_map ("cannot write file %s: %s", path, strerror (errno));
+		rspamd_file_unlock (fd, FALSE);
+		close (fd);
+
+		return FALSE;
+	}
+
+	rspamd_file_unlock (fd, FALSE);
+	close (fd);
+
+	msg_info_map ("saved data from %s in %s, %uz bytes", bk->uri, path, len);
+
+	return TRUE;
+}
+
+static gboolean
+rspamd_map_read_http_cached_file (struct rspamd_map *map,
+								  struct rspamd_map_backend *bk,
+								  struct http_map_data *htdata,
+								  struct map_cb_data *cbdata)
+{
+	gchar path[PATH_MAX];
+	guchar digest[rspamd_cryptobox_HASHBYTES];
+	struct rspamd_config *cfg = map->cfg;
+	gint fd;
+	struct stat st;
+	struct rspamd_http_file_data header;
+
+	if (cfg->maps_cache_dir == NULL || cfg->maps_cache_dir[0] == '\0') {
+		return FALSE;
+	}
+
+	rspamd_cryptobox_hash (digest, bk->uri, strlen (bk->uri), NULL, 0);
+	rspamd_snprintf (path, sizeof (path), "%s%c%*xs.map", cfg->maps_cache_dir,
+			G_DIR_SEPARATOR, 20, digest);
+
+	fd = rspamd_file_xopen (path, O_RDONLY, 00600, FALSE);
+
+	if (fd == -1) {
+		return FALSE;
+	}
+
+	if (!rspamd_file_lock (fd, FALSE)) {
+		msg_err_map ("cannot lock file %s: %s", path, strerror (errno));
+		close (fd);
+
+		return FALSE;
+	}
+
+	(void)fstat (fd, &st);
+
+	if (read (fd, &header, sizeof (header)) != sizeof (header)) {
+		msg_err_map ("cannot read file %s: %s", path, strerror (errno));
+		rspamd_file_unlock (fd, FALSE);
+		close (fd);
+
+		return FALSE;
+	}
+
+	if (memcmp (header.magic, rspamd_http_file_magic,
+			sizeof (rspamd_http_file_magic)) != 0) {
+		msg_err_map ("invalid magic in file %s: %s", path, strerror (errno));
+		rspamd_file_unlock (fd, FALSE);
+		close (fd);
+
+		return FALSE;
+	}
+
+	rspamd_file_unlock (fd, FALSE);
+	close (fd);
+
+	map->next_check = header.next_check;
+	htdata->last_modified = header.mtime;
+
+	/* Now read file data */
+	/* Perform buffered read: fail-safe */
+	if (!read_map_file_chunks (map, cbdata, path,
+			st.st_size - header.data_off, header.data_off)) {
+		return FALSE;
+	}
+
+	msg_info_map ("read cached data for %s from %s, %uz bytes", bk->uri, path,
+			st.st_size - header.data_off);
+
+	return TRUE;
+}
+
 /**
  * Async HTTP callback
  */
@@ -1284,10 +1561,10 @@ rspamd_map_common_http_callback (struct rspamd_map *map,
 
 	data = bk->data.hd;
 
-	if (g_atomic_int_get (&map->cache->available) == 1) {
+	if (g_atomic_int_get (&data->cache->available) == 1) {
 		/* Read cached data */
 		if (check) {
-			if (data->last_modified < map->cache->last_modified) {
+			if (data->last_modified < data->cache->last_modified) {
 				periodic->need_modify = TRUE;
 				/* Reset the whole chain */
 				periodic->cur_backend = 0;
@@ -1309,13 +1586,13 @@ rspamd_map_common_http_callback (struct rspamd_map *map,
 		}
 		else {
 			if (map->active_http &&
-					data->last_modified > map->cache->last_modified) {
+					data->last_modified > data->cache->last_modified) {
 				goto check;
 			}
 			else if (rspamd_map_read_cached (map, bk, periodic, data->host)) {
 				/* Switch to the next backend */
 				periodic->cur_backend++;
-				data->last_modified = map->cache->last_modified;
+				data->last_modified = data->cache->last_modified;
 				rspamd_map_periodic_callback (-1, EV_TIMEOUT, periodic);
 
 				return;
@@ -1389,8 +1666,8 @@ check:
 			MAP_RETAIN (cbd, "http_callback_data");
 		}
 
-		map->dtor = free_http_cbdata_dtor;
-		map->dtor_data = cbd;
+		map->tmp_dtor = free_http_cbdata_dtor;
+		map->tmp_dtor_data = cbd;
 	}
 	else {
 		msg_warn_map ("cannot load map: DNS resolver is not initialized");
@@ -1582,45 +1859,49 @@ rspamd_map_periodic_callback (gint fd, short what, void *ud)
 		return;
 	}
 
-	bk = g_ptr_array_index (cbd->map->backends, cbd->cur_backend);
-	g_assert (bk != NULL);
+	if (!(cbd->map->wrk && cbd->map->wrk->wanna_die)) {
+		bk = g_ptr_array_index (cbd->map->backends, cbd->cur_backend);
+		g_assert (bk != NULL);
 
-	if (cbd->need_modify) {
-		/* Load data from the next backend */
-		switch (bk->protocol) {
-		case MAP_PROTO_HTTP:
-		case MAP_PROTO_HTTPS:
-			rspamd_map_http_read_callback (fd, what, cbd);
-			break;
-		case MAP_PROTO_FILE:
-			rspamd_map_file_read_callback (fd, what, cbd);
-			break;
-		case MAP_PROTO_STATIC:
-			rspamd_map_static_read_callback (fd, what, cbd);
-			break;
-		}
-	}
-	else {
-		/* Check the next backend */
-		switch (bk->protocol) {
-		case MAP_PROTO_HTTP:
-		case MAP_PROTO_HTTPS:
-			rspamd_map_http_check_callback (fd, what, cbd);
-			break;
-		case MAP_PROTO_FILE:
-			rspamd_map_file_check_callback (fd, what, cbd);
-			break;
-		case MAP_PROTO_STATIC:
-			rspamd_map_static_check_callback (fd, what, cbd);
-			break;
+		if (cbd->need_modify) {
+			/* Load data from the next backend */
+			switch (bk->protocol) {
+			case MAP_PROTO_HTTP:
+			case MAP_PROTO_HTTPS:
+				rspamd_map_http_read_callback (fd, what, cbd);
+				break;
+			case MAP_PROTO_FILE:
+				rspamd_map_file_read_callback (fd, what, cbd);
+				break;
+			case MAP_PROTO_STATIC:
+				rspamd_map_static_read_callback (fd, what, cbd);
+				break;
+			}
+		} else {
+			/* Check the next backend */
+			switch (bk->protocol) {
+			case MAP_PROTO_HTTP:
+			case MAP_PROTO_HTTPS:
+				rspamd_map_http_check_callback (fd, what, cbd);
+				break;
+			case MAP_PROTO_FILE:
+				rspamd_map_file_check_callback (fd, what, cbd);
+				break;
+			case MAP_PROTO_STATIC:
+				rspamd_map_static_check_callback (fd, what, cbd);
+				break;
+			}
 		}
 	}
 }
 
 /* Start watching event for all maps */
 void
-rspamd_map_watch (struct rspamd_config *cfg, struct event_base *ev_base,
-		struct rspamd_dns_resolver *resolver, gboolean active_http)
+rspamd_map_watch (struct rspamd_config *cfg,
+				  struct event_base *ev_base,
+				  struct rspamd_dns_resolver *resolver,
+				  struct rspamd_worker *worker,
+				  gboolean active_http)
 {
 	GList *cur = cfg->maps;
 	struct rspamd_map *map;
@@ -1630,6 +1911,7 @@ rspamd_map_watch (struct rspamd_config *cfg, struct event_base *ev_base,
 		map = cur->data;
 		map->ev_base = ev_base;
 		map->r = resolver;
+		map->wrk = worker;
 
 		if (active_http) {
 			map->active_http = active_http;
@@ -1640,7 +1922,8 @@ rspamd_map_watch (struct rspamd_config *cfg, struct event_base *ev_base,
 
 			if (map->poll_timeout >= cfg->map_timeout &&
 					cfg->map_file_watch_multiplier < 1.0) {
-				map->poll_timeout = map->poll_timeout * cfg->map_file_watch_multiplier;
+				map->poll_timeout =
+						map->poll_timeout * cfg->map_file_watch_multiplier;
 			}
 		}
 
@@ -1650,28 +1933,145 @@ rspamd_map_watch (struct rspamd_config *cfg, struct event_base *ev_base,
 	}
 }
 
+void
+rspamd_map_preload (struct rspamd_config *cfg)
+{
+	GList *cur = cfg->maps;
+	struct rspamd_map *map;
+	struct rspamd_map_backend *bk;
+	guint i;
+	gboolean map_ok;
+
+	/* First of all do synced read of data */
+	while (cur) {
+		map = cur->data;
+		map_ok = TRUE;
+
+		PTR_ARRAY_FOREACH (map->backends, i, bk) {
+			if (!(bk->protocol == MAP_PROTO_FILE ||
+				  bk->protocol == MAP_PROTO_STATIC)) {
+
+				if (bk->protocol == MAP_PROTO_HTTP ||
+						bk->protocol == MAP_PROTO_HTTPS) {
+					if (!rspamd_map_has_http_cached_file (map, bk)) {
+
+						if (!map->fallback_backend) {
+							map_ok = FALSE;
+						}
+						break;
+					}
+					else {
+						continue; /* We are yet fine */
+					}
+				}
+				map_ok = FALSE;
+				break;
+			}
+		}
+
+		if (map_ok) {
+			struct map_periodic_cbdata fake_cbd;
+			gboolean succeed = TRUE;
+
+			memset (&fake_cbd, 0, sizeof (fake_cbd));
+			fake_cbd.cbdata.state = 0;
+			fake_cbd.cbdata.prev_data = *map->user_data;
+			fake_cbd.cbdata.cur_data = NULL;
+			fake_cbd.cbdata.map = map;
+			fake_cbd.map = map;
+
+			PTR_ARRAY_FOREACH (map->backends, i, bk) {
+				fake_cbd.cur_backend = i;
+
+				if (bk->protocol == MAP_PROTO_FILE) {
+					if (!read_map_file (map, bk->data.fd, bk, &fake_cbd)) {
+						succeed = FALSE;
+						break;
+					}
+				}
+				else if (bk->protocol == MAP_PROTO_STATIC) {
+					if (!read_map_static (map, bk->data.sd, bk, &fake_cbd)) {
+						succeed = FALSE;
+						break;
+					}
+				}
+				else if (bk->protocol == MAP_PROTO_HTTP ||
+						 bk->protocol == MAP_PROTO_HTTPS) {
+					if (!rspamd_map_read_http_cached_file (map, bk, bk->data.hd,
+							&fake_cbd.cbdata)) {
+
+						if (map->fallback_backend) {
+							/* Try fallback */
+							g_assert (map->fallback_backend->protocol ==
+									  MAP_PROTO_FILE);
+							if (!read_map_file (map,
+									map->fallback_backend->data.fd,
+									map->fallback_backend, &fake_cbd)) {
+								succeed = FALSE;
+								break;
+							}
+						}
+						else {
+							succeed = FALSE;
+							break;
+						}
+					}
+				}
+				else {
+					g_assert_not_reached ();
+				}
+			}
+
+			if (succeed) {
+				map->fin_callback (&fake_cbd.cbdata);
+
+				if (fake_cbd.cbdata.cur_data) {
+					*map->user_data = fake_cbd.cbdata.cur_data;
+				}
+			}
+			else {
+				msg_info_map ("preload of %s failed", map->name);
+			}
+
+		}
+
+		cur = g_list_next (cur);
+	}
+}
+
 void
 rspamd_map_remove_all (struct rspamd_config *cfg)
 {
 	struct rspamd_map *map;
 	GList *cur;
 	struct rspamd_map_backend *bk;
+	struct map_cb_data cbdata;
 	guint i;
 
 	for (cur = cfg->maps; cur != NULL; cur = g_list_next (cur)) {
 		map = cur->data;
 
-		for (i = 0; i < map->backends->len; i ++) {
-			bk = g_ptr_array_index (map->backends, i);
-			MAP_RELEASE (bk, "rspamd_map_backend");
-		}
-
-		if (g_atomic_int_compare_and_exchange (&map->cache->available, 1, 0)) {
-			unlink (map->cache->shmem_name);
+		if (map->tmp_dtor) {
+			map->tmp_dtor (map->tmp_dtor_data);
 		}
 
 		if (map->dtor) {
-			map->dtor (map->dtor_data);
+			cbdata.prev_data = NULL;
+			cbdata.map = map;
+			cbdata.cur_data = *map->user_data;
+
+			map->dtor (&cbdata);
+			*map->user_data = NULL;
+		}
+
+		for (i = 0; i < map->backends->len; i ++) {
+			bk = g_ptr_array_index (map->backends, i);
+
+			MAP_RELEASE (bk, "rspamd_map_backend");
+		}
+
+		if (map->fallback_backend) {
+			MAP_RELEASE (map->fallback_backend, "rspamd_map_backend");
 		}
 	}
 
@@ -1690,6 +2090,7 @@ rspamd_map_check_proto (struct rspamd_config *cfg,
 
 	end = pos + strlen (pos);
 
+	/* Static check */
 	if (g_ascii_strcasecmp (pos, "static") == 0) {
 		bk->protocol = MAP_PROTO_STATIC;
 		bk->uri = g_strdup (pos);
@@ -1704,46 +2105,53 @@ rspamd_map_check_proto (struct rspamd_config *cfg,
 		return pos + 4;
 	}
 
-	if (g_ascii_strncasecmp (pos, "sign+", sizeof ("sign+") - 1) == 0) {
-		bk->is_signed = TRUE;
-		pos += sizeof ("sign+") - 1;
-	}
-
-	if (g_ascii_strncasecmp (pos, "key=", sizeof ("key=") - 1) == 0) {
-		pos += sizeof ("key=") - 1;
-		end_key = memchr (pos, '+', end - pos);
-
-		if (end_key != NULL) {
-			bk->trusted_pubkey = rspamd_pubkey_from_base32 (pos, end_key - pos,
-					RSPAMD_KEYPAIR_SIGN, RSPAMD_CRYPTOBOX_MODE_25519);
-
-			if (bk->trusted_pubkey == NULL) {
-				msg_err_config ("cannot read pubkey from map: %s",
-						map_line);
-				return NULL;
-			}
-			pos = end_key + 1;
+	for (;;) {
+		if (g_ascii_strncasecmp (pos, "sign+", sizeof ("sign+") - 1) == 0) {
+			bk->is_signed = TRUE;
+			pos += sizeof ("sign+") - 1;
 		}
-		else if (end - pos > 64) {
-			/* Try hex encoding */
-			bk->trusted_pubkey = rspamd_pubkey_from_hex (pos, 64,
-					RSPAMD_KEYPAIR_SIGN, RSPAMD_CRYPTOBOX_MODE_25519);
+		else if (g_ascii_strncasecmp (pos, "fallback+", sizeof ("fallback+") - 1) == 0) {
+			bk->is_fallback = TRUE;
+			pos += sizeof ("fallback+") - 1;
+		}
+		else if (g_ascii_strncasecmp (pos, "key=", sizeof ("key=") - 1) == 0) {
+			pos += sizeof ("key=") - 1;
+			end_key = memchr (pos, '+', end - pos);
 
-			if (bk->trusted_pubkey == NULL) {
+			if (end_key != NULL) {
+				bk->trusted_pubkey = rspamd_pubkey_from_base32 (pos, end_key - pos,
+						RSPAMD_KEYPAIR_SIGN, RSPAMD_CRYPTOBOX_MODE_25519);
+
+				if (bk->trusted_pubkey == NULL) {
+					msg_err_config ("cannot read pubkey from map: %s",
+							map_line);
+					return NULL;
+				}
+				pos = end_key + 1;
+			} else if (end - pos > 64) {
+				/* Try hex encoding */
+				bk->trusted_pubkey = rspamd_pubkey_from_hex (pos, 64,
+						RSPAMD_KEYPAIR_SIGN, RSPAMD_CRYPTOBOX_MODE_25519);
+
+				if (bk->trusted_pubkey == NULL) {
+					msg_err_config ("cannot read pubkey from map: %s",
+							map_line);
+					return NULL;
+				}
+				pos += 64;
+			} else {
 				msg_err_config ("cannot read pubkey from map: %s",
 						map_line);
 				return NULL;
 			}
-			pos += 64;
+
+			if (*pos == '+' || *pos == ':') {
+				pos++;
+			}
 		}
 		else {
-			msg_err_config ("cannot read pubkey from map: %s",
-					map_line);
-			return NULL;
-		}
-
-		if (*pos == '+' || *pos == ':') {
-			pos ++;
+			/* No known flags */
+			break;
 		}
 	}
 
@@ -1776,7 +2184,6 @@ rspamd_map_check_proto (struct rspamd_config *cfg,
 		return NULL;
 	}
 
-
 	return pos;
 }
 
@@ -1793,6 +2200,9 @@ rspamd_map_is_map (const gchar *map_line)
 	else if (g_ascii_strncasecmp (map_line, "sign+", sizeof ("sign+") - 1) == 0) {
 		ret = TRUE;
 	}
+	else if (g_ascii_strncasecmp (map_line, "fallback+", sizeof ("fallback+") - 1) == 0) {
+		ret = TRUE;
+	}
 	else if (g_ascii_strncasecmp (map_line, "file://", sizeof ("file://") - 1) == 0) {
 		ret = TRUE;
 	}
@@ -1823,16 +2233,32 @@ rspamd_map_backend_dtor (struct rspamd_map_backend *bk)
 			if (bk->data.sd->data) {
 				g_free (bk->data.sd->data);
 			}
+
+			g_free (bk->data.sd);
 		}
 		break;
 	case MAP_PROTO_HTTP:
 	case MAP_PROTO_HTTPS:
 		if (bk->data.hd) {
-			g_free (bk->data.hd->host);
-			g_free (bk->data.hd->path);
+			struct http_map_data *data = bk->data.hd;
 
-			if (bk->data.hd->etag) {
-				rspamd_fstring_free (bk->data.hd->etag);
+			g_free (data->host);
+			g_free (data->path);
+
+			if (data->etag) {
+				rspamd_fstring_free (data->etag);
+			}
+
+			if (g_atomic_int_compare_and_exchange (&data->cache->available, 1, 0)) {
+				if (data->cur_cache_cbd) {
+					MAP_RELEASE (data->cur_cache_cbd->shm,
+							"rspamd_http_map_cached_cbdata");
+					event_del (&data->cur_cache_cbd->timeout);
+					g_free (data->cur_cache_cbd);
+					data->cur_cache_cbd = NULL;
+				}
+
+				unlink (data->cache->shmem_name);
 			}
 
 			g_free (bk->data.hd);
@@ -1865,6 +2291,12 @@ rspamd_map_parse_backend (struct rspamd_config *cfg, const gchar *map_line)
 		goto err;
 	}
 
+	if (bk->is_fallback && bk->protocol != MAP_PROTO_FILE) {
+		msg_err_config ("fallback backend must be file for %s", bk->uri);
+
+		goto err;
+	}
+
 	end = map_line + strlen (map_line);
 	if (end - map_line > 5) {
 		p = end - 5;
@@ -1885,9 +2317,9 @@ rspamd_map_parse_backend (struct rspamd_config *cfg, const gchar *map_line)
 		if (access (bk->uri, R_OK) == -1) {
 			if (errno != ENOENT) {
 				msg_err_config ("cannot open file '%s': %s", bk->uri, strerror (errno));
-				return NULL;
-
+				goto err;
 			}
+
 			msg_info_config (
 					"map '%s' is not found, but it can be loaded automatically later",
 					bk->uri);
@@ -1908,7 +2340,7 @@ rspamd_map_parse_backend (struct rspamd_config *cfg, const gchar *map_line)
 		else {
 			if (!(up.field_set & 1 << UF_HOST)) {
 				msg_err_config ("cannot parse HTTP url: %s: no host", bk->uri);
-				return NULL;
+				goto err;
 			}
 
 			tok.begin = bk->uri + up.field_data[UF_HOST].off;
@@ -1935,8 +2367,12 @@ rspamd_map_parse_backend (struct rspamd_config *cfg, const gchar *map_line)
 			}
 		}
 
+		hdata->cache = rspamd_mempool_alloc0_shared (cfg->cfg_pool,
+						sizeof (*hdata->cache));
+
 		bk->data.hd = hdata;
-	}else if (bk->protocol == MAP_PROTO_STATIC) {
+	}
+	else if (bk->protocol == MAP_PROTO_STATIC) {
 		sdata = g_malloc0 (sizeof (*sdata));
 		bk->data.sd = sdata;
 	}
@@ -2008,11 +2444,12 @@ rspamd_map_add_static_string (struct rspamd_config *cfg,
 
 struct rspamd_map *
 rspamd_map_add (struct rspamd_config *cfg,
-	const gchar *map_line,
-	const gchar *description,
-	map_cb_t read_callback,
-	map_fin_cb_t fin_callback,
-	void **user_data)
+				const gchar *map_line,
+				const gchar *description,
+				map_cb_t read_callback,
+				map_fin_cb_t fin_callback,
+				map_dtor_t dtor,
+				void **user_data)
 {
 	struct rspamd_map *map;
 	struct rspamd_map_backend *bk;
@@ -2022,16 +2459,22 @@ rspamd_map_add (struct rspamd_config *cfg,
 		return NULL;
 	}
 
+	if (bk->is_fallback) {
+		msg_err_config ("cannot add map with fallback only backend: %s", bk->uri);
+		REF_RELEASE (bk);
+
+		return NULL;
+	}
+
 	map = rspamd_mempool_alloc0 (cfg->cfg_pool, sizeof (struct rspamd_map));
 	map->read_callback = read_callback;
 	map->fin_callback = fin_callback;
+	map->dtor = dtor;
 	map->user_data = user_data;
 	map->cfg = cfg;
 	map->id = rspamd_random_uint64_fast ();
 	map->locked =
 		rspamd_mempool_alloc0_shared (cfg->cfg_pool, sizeof (gint));
-	map->cache =
-			rspamd_mempool_alloc0_shared (cfg->cfg_pool, sizeof (*map->cache));
 	map->backends = g_ptr_array_sized_new (1);
 	rspamd_mempool_add_destructor (cfg->cfg_pool, rspamd_ptr_array_free_hard,
 			map->backends);
@@ -2056,13 +2499,30 @@ rspamd_map_add (struct rspamd_config *cfg,
 	return map;
 }
 
+static inline void
+rspamd_map_add_backend (struct rspamd_map *map, struct rspamd_map_backend *bk)
+{
+	if (bk->is_fallback) {
+		if (map->fallback_backend) {
+			msg_warn_map ("redefining fallback backend from %s to %s",
+					map->fallback_backend->uri, bk->uri);
+		}
+
+		map->fallback_backend = bk;
+	}
+	else {
+		g_ptr_array_add (map->backends, bk);
+	}
+}
+
 struct rspamd_map*
 rspamd_map_add_from_ucl (struct rspamd_config *cfg,
-	const ucl_object_t *obj,
-	const gchar *description,
-	map_cb_t read_callback,
-	map_fin_cb_t fin_callback,
-	void **user_data)
+						 const ucl_object_t *obj,
+						 const gchar *description,
+						 map_cb_t read_callback,
+						 map_fin_cb_t fin_callback,
+						 map_dtor_t dtor,
+						 void **user_data)
 {
 	ucl_object_iter_t it = NULL;
 	const ucl_object_t *cur, *elt;
@@ -2075,19 +2535,18 @@ rspamd_map_add_from_ucl (struct rspamd_config *cfg,
 	if (ucl_object_type (obj) == UCL_STRING) {
 		/* Just a plain string */
 		return rspamd_map_add (cfg, ucl_object_tostring (obj), description,
-				read_callback, fin_callback, user_data);
+				read_callback, fin_callback, dtor, user_data);
 	}
 
 	map = rspamd_mempool_alloc0 (cfg->cfg_pool, sizeof (struct rspamd_map));
 	map->read_callback = read_callback;
 	map->fin_callback = fin_callback;
+	map->dtor = dtor;
 	map->user_data = user_data;
 	map->cfg = cfg;
 	map->id = rspamd_random_uint64_fast ();
 	map->locked =
 			rspamd_mempool_alloc0_shared (cfg->cfg_pool, sizeof (gint));
-	map->cache =
-				rspamd_mempool_alloc0_shared (cfg->cfg_pool, sizeof (*map->cache));
 	map->backends = g_ptr_array_new ();
 	rspamd_mempool_add_destructor (cfg->cfg_pool, rspamd_ptr_array_free_hard,
 			map->backends);
@@ -2104,11 +2563,7 @@ rspamd_map_add_from_ucl (struct rspamd_config *cfg,
 				bk = rspamd_map_parse_backend (cfg, ucl_object_tostring (cur));
 
 				if (bk != NULL) {
-					if (bk->protocol == MAP_PROTO_FILE) {
-						map->poll_timeout = (map->poll_timeout *
-								cfg->map_file_watch_multiplier);
-					}
-					g_ptr_array_add (map->backends, bk);
+					rspamd_map_add_backend (map, bk);
 
 					if (!map->name) {
 						map->name = rspamd_mempool_strdup (cfg->cfg_pool,
@@ -2161,10 +2616,7 @@ rspamd_map_add_from_ucl (struct rspamd_config *cfg,
 					bk = rspamd_map_parse_backend (cfg, ucl_object_tostring (cur));
 
 					if (bk != NULL) {
-						if (bk->protocol == MAP_PROTO_FILE) {
-							map->poll_timeout = (map->poll_timeout * cfg->map_file_watch_multiplier);
-						}
-						g_ptr_array_add (map->backends, bk);
+						rspamd_map_add_backend (map, bk);
 
 						if (!map->name) {
 							map->name = rspamd_mempool_strdup (cfg->cfg_pool,
@@ -2191,10 +2643,7 @@ rspamd_map_add_from_ucl (struct rspamd_config *cfg,
 			bk = rspamd_map_parse_backend (cfg, ucl_object_tostring (elt));
 
 			if (bk != NULL) {
-				if (bk->protocol == MAP_PROTO_FILE) {
-					map->poll_timeout = (map->poll_timeout * cfg->map_file_watch_multiplier);
-				}
-				g_ptr_array_add (map->backends, bk);
+				rspamd_map_add_backend (map, bk);
 
 				if (!map->name) {
 					map->name = rspamd_mempool_strdup (cfg->cfg_pool,
@@ -2207,54 +2656,6 @@ rspamd_map_add_from_ucl (struct rspamd_config *cfg,
 			msg_err_config ("map has no urls to be loaded: no valid backends");
 			goto err;
 		}
-
-		PTR_ARRAY_FOREACH (map->backends, i, bk) {
-			if (bk->protocol == MAP_PROTO_STATIC) {
-				GString *map_data;
-				/* We need data field in ucl */
-				elt = ucl_object_lookup (obj, "data");
-
-				if (elt == NULL) {
-					msg_err_config ("map has static backend but no `data` field");
-					goto err;
-				}
-
-
-				if (ucl_object_type (elt) == UCL_STRING) {
-					map_data = g_string_sized_new (32);
-
-					if (rspamd_map_add_static_string (cfg, elt, map_data)) {
-						bk->data.sd->data = map_data->str;
-						bk->data.sd->len = map_data->len;
-						g_string_free (map_data, FALSE);
-					}
-					else {
-						g_string_free (map_data, TRUE);
-						msg_err_config ("map has static backend with invalid `data` field");
-						goto err;
-					}
-				}
-				else if (ucl_object_type (elt) == UCL_ARRAY) {
-					map_data = g_string_sized_new (32);
-					it = ucl_object_iterate_new (elt);
-
-					while ((cur = ucl_object_iterate_safe (it, true))) {
-						if (!rspamd_map_add_static_string (cfg, cur, map_data)) {
-							g_string_free (map_data, TRUE);
-							msg_err_config ("map has static backend with invalid "
-											"`data` field");
-							ucl_object_iterate_free (it);
-							goto err;
-						}
-					}
-
-					ucl_object_iterate_free (it);
-					bk->data.sd->data = map_data->str;
-					bk->data.sd->len = map_data->len;
-					g_string_free (map_data, FALSE);
-				}
-			}
-		}
 	}
 	else {
 		msg_err_config ("map has invalid type for value: %s",
@@ -2262,6 +2663,64 @@ rspamd_map_add_from_ucl (struct rspamd_config *cfg,
 		goto err;
 	}
 
+	gboolean all_local = TRUE;
+
+	PTR_ARRAY_FOREACH (map->backends, i, bk) {
+		if (bk->protocol == MAP_PROTO_STATIC) {
+			GString *map_data;
+			/* We need data field in ucl */
+			elt = ucl_object_lookup (obj, "data");
+
+			if (elt == NULL) {
+				msg_err_config ("map has static backend but no `data` field");
+				goto err;
+			}
+
+
+			if (ucl_object_type (elt) == UCL_STRING) {
+				map_data = g_string_sized_new (32);
+
+				if (rspamd_map_add_static_string (cfg, elt, map_data)) {
+					bk->data.sd->data = map_data->str;
+					bk->data.sd->len = map_data->len;
+					g_string_free (map_data, FALSE);
+				}
+				else {
+					g_string_free (map_data, TRUE);
+					msg_err_config ("map has static backend with invalid `data` field");
+					goto err;
+				}
+			}
+			else if (ucl_object_type (elt) == UCL_ARRAY) {
+				map_data = g_string_sized_new (32);
+				it = ucl_object_iterate_new (elt);
+
+				while ((cur = ucl_object_iterate_safe (it, true))) {
+					if (!rspamd_map_add_static_string (cfg, cur, map_data)) {
+						g_string_free (map_data, TRUE);
+						msg_err_config ("map has static backend with invalid "
+										"`data` field");
+						ucl_object_iterate_free (it);
+						goto err;
+					}
+				}
+
+				ucl_object_iterate_free (it);
+				bk->data.sd->data = map_data->str;
+				bk->data.sd->len = map_data->len;
+				g_string_free (map_data, FALSE);
+			}
+		}
+		else if (bk->protocol != MAP_PROTO_FILE) {
+			all_local = FALSE;
+		}
+	}
+
+	if (all_local) {
+		map->poll_timeout = (map->poll_timeout *
+							 cfg->map_file_watch_multiplier);
+	}
+
 	rspamd_map_calculate_hash (map);
 	msg_debug_map ("added map from ucl");
 
@@ -2271,6 +2730,12 @@ rspamd_map_add_from_ucl (struct rspamd_config *cfg,
 
 err:
 
+	if (map) {
+		PTR_ARRAY_FOREACH (map->backends, i, bk) {
+			MAP_RELEASE (bk, "rspamd_map_backend");
+		}
+	}
+
 	return NULL;
 }
 
