@@ -150,7 +150,7 @@ rspamd_parse_kv_list (
 		case map_read_key:
 			/* read key */
 			/* Check here comments, eol and end of buffer */
-			if (*p == '#') {
+			if (*p == '#' && (p == c || *(p - 1) != '\\')) {
 				if (p - c > 0) {
 					/* Store a single key */
 					MAP_STORE_KEY;
@@ -813,6 +813,17 @@ rspamd_kv_list_fin (struct map_cb_data *data)
 	}
 }
 
+void
+rspamd_kv_list_dtor (struct map_cb_data *data)
+{
+	struct rspamd_hash_map_helper *htb;
+
+	if (data->cur_data) {
+		htb = (struct rspamd_hash_map_helper *)data->cur_data;
+		rspamd_map_helper_destroy_hash (htb);
+	}
+}
+
 gchar *
 rspamd_radix_read (
 		gchar * chunk,
@@ -858,6 +869,17 @@ rspamd_radix_fin (struct map_cb_data *data)
 	}
 }
 
+void
+rspamd_radix_dtor (struct map_cb_data *data)
+{
+	struct rspamd_radix_map_helper *r;
+
+	if (data->cur_data) {
+		r = (struct rspamd_radix_map_helper *)data->cur_data;
+		rspamd_map_helper_destroy_radix (r);
+	}
+}
+
 static void
 rspamd_re_map_finalize (struct rspamd_regexp_map_helper *re_map)
 {
@@ -1005,7 +1027,8 @@ rspamd_regexp_list_read_multiple (
 	struct rspamd_regexp_map_helper *re_map;
 
 	if (data->cur_data == NULL) {
-		re_map = rspamd_map_helper_new_regexp (data->map, RSPAMD_REGEXP_MAP_FLAG_MULTIPLE);
+		re_map = rspamd_map_helper_new_regexp (data->map,
+				RSPAMD_REGEXP_MAP_FLAG_MULTIPLE);
 		data->cur_data = re_map;
 	}
 
@@ -1018,6 +1041,31 @@ rspamd_regexp_list_read_multiple (
 			final);
 }
 
+gchar *
+rspamd_glob_list_read_multiple (
+		gchar *chunk,
+		gint len,
+		struct map_cb_data *data,
+		gboolean final)
+{
+	struct rspamd_regexp_map_helper *re_map;
+
+	if (data->cur_data == NULL) {
+		re_map = rspamd_map_helper_new_regexp (data->map,
+				RSPAMD_REGEXP_MAP_FLAG_GLOB|RSPAMD_REGEXP_MAP_FLAG_MULTIPLE);
+		data->cur_data = re_map;
+	}
+
+	return rspamd_parse_kv_list (
+			chunk,
+			len,
+			data,
+			rspamd_map_helper_insert_re,
+			hash_fill,
+			final);
+}
+
+
 void
 rspamd_regexp_list_fin (struct map_cb_data *data)
 {
@@ -1037,6 +1085,13 @@ rspamd_regexp_list_fin (struct map_cb_data *data)
 		data->map->digest = rspamd_cryptobox_fast_hash_final (&re_map->hst);
 	}
 }
+void
+rspamd_regexp_list_dtor (struct map_cb_data *data)
+{
+	if (data->cur_data) {
+		rspamd_map_helper_destroy_regexp (data->cur_data);
+	}
+}
 
 #ifdef WITH_HYPERSCAN
 static int
