@@ -20,6 +20,7 @@
 #include "worker_util.h"
 #include "ottery.h"
 #include "rspamd_control.h"
+#include "lua_thread_pool.h"
 #include <math.h>
 #include <sys/wait.h>
 #include <src/libserver/rspamd_control.h>
@@ -27,6 +28,10 @@
 /* Lua module init function */
 #define MODULE_INIT_FUNC "module_init"
 
+#ifdef WITH_LUA_TRACE
+ucl_object_t *lua_traces;
+#endif
+
 const luaL_reg null_reg[] = {
 	{"__tostring", rspamd_lua_class_tostring},
 	{NULL, NULL}
@@ -375,6 +380,330 @@ rspamd_lua_set_path (lua_State *L, const ucl_object_t *cfg_obj, GHashTable *vars
 	lua_pop (L, 1);
 }
 
+static gint
+rspamd_lua_cmp_version_components (const gchar *comp1, const gchar *comp2)
+{
+	guint v1, v2;
+
+	v1 = strtoul (comp1, NULL, 10);
+	v2 = strtoul (comp2, NULL, 10);
+
+	return v1 - v2;
+}
+
+static int
+rspamd_lua_rspamd_version_cmp (lua_State *L)
+{
+	const gchar *ver;
+	gchar **components;
+	gint ret = 0;
+
+	if (lua_type (L, 2) == LUA_TSTRING) {
+		ver = lua_tostring (L, 2);
+
+		components = g_strsplit_set (ver, ".-_", -1);
+
+		if (!components) {
+			return luaL_error (L, "invalid arguments to 'cmp': %s", ver);
+		}
+
+		if (components[0]) {
+			ret = rspamd_lua_cmp_version_components (components[0],
+					RSPAMD_VERSION_MAJOR);
+		}
+
+		if (ret) {
+			goto set;
+		}
+
+		if (components[1]) {
+			ret = rspamd_lua_cmp_version_components (components[1],
+					RSPAMD_VERSION_MINOR);
+		}
+
+		if (ret) {
+			goto set;
+		}
+
+		if (components[2]) {
+			ret = rspamd_lua_cmp_version_components (components[2],
+					RSPAMD_VERSION_PATCH);
+		}
+
+		/*
+		 * XXX: we don't compare git releases assuming that it is meaningless
+		 */
+	}
+	else {
+		return luaL_error (L, "invalid arguments to 'cmp'");
+	}
+
+set:
+	g_strfreev (components);
+	lua_pushinteger (L, ret);
+
+	return 1;
+}
+
+static int
+rspamd_lua_rspamd_version_numeric (lua_State *L)
+{
+	static gint64 version_num = RSPAMD_VERSION_NUM;
+	const gchar *type;
+
+	if (lua_gettop (L) >= 2 && lua_type (L, 1) == LUA_TSTRING) {
+		type = lua_tostring (L, 1);
+		if (g_ascii_strcasecmp (type, "short") == 0) {
+			version_num = RSPAMD_VERSION_MAJOR_NUM * 1000 +
+						  RSPAMD_VERSION_MINOR_NUM * 100 +
+						  RSPAMD_VERSION_PATCH_NUM * 10;
+		}
+		else if (g_ascii_strcasecmp (type, "main") == 0) {
+			version_num = RSPAMD_VERSION_MAJOR_NUM * 1000 +
+						  RSPAMD_VERSION_MINOR_NUM * 100;
+		}
+		else if (g_ascii_strcasecmp (type, "major") == 0) {
+			version_num = RSPAMD_VERSION_MAJOR_NUM;
+		}
+		else if (g_ascii_strcasecmp (type, "minor") == 0) {
+			version_num = RSPAMD_VERSION_MINOR_NUM;
+		}
+		else if (g_ascii_strcasecmp (type, "patch") == 0) {
+			version_num = RSPAMD_VERSION_PATCH_NUM;
+		}
+	}
+
+	lua_pushinteger (L, version_num);
+
+	return 1;
+}
+
+static int
+rspamd_lua_rspamd_version (lua_State *L)
+{
+	const gchar *result = NULL, *type;
+
+	if (lua_gettop (L) == 0) {
+		result = RVERSION;
+	}
+	else if (lua_gettop (L) >= 1 && lua_type (L, 1) == LUA_TSTRING) {
+		/* We got something like string */
+		type = lua_tostring (L, 1);
+
+		if (g_ascii_strcasecmp (type, "short") == 0) {
+			result = RSPAMD_VERSION_MAJOR
+					 "." RSPAMD_VERSION_MINOR
+					 "." RSPAMD_VERSION_PATCH;
+		}
+		else if (g_ascii_strcasecmp (type, "main") == 0) {
+			result = RSPAMD_VERSION_MAJOR "." RSPAMD_VERSION_MINOR;
+		}
+		else if (g_ascii_strcasecmp (type, "major") == 0) {
+			result = RSPAMD_VERSION_MAJOR;
+		}
+		else if (g_ascii_strcasecmp (type, "minor") == 0) {
+			result = RSPAMD_VERSION_MINOR;
+		}
+		else if (g_ascii_strcasecmp (type, "patch") == 0) {
+			result = RSPAMD_VERSION_PATCH;
+		}
+		else if (g_ascii_strcasecmp (type, "id") == 0) {
+			result = RID;
+		}
+		else if (g_ascii_strcasecmp (type, "num") == 0) {
+			return rspamd_lua_rspamd_version_numeric (L);
+		}
+		else if (g_ascii_strcasecmp (type, "cmp") == 0) {
+			return rspamd_lua_rspamd_version_cmp (L);
+		}
+	}
+
+	lua_pushstring (L, result);
+
+	return 1;
+}
+
+void
+rspamd_lua_set_globals (struct rspamd_config *cfg, lua_State *L,
+							GHashTable *vars)
+{
+	struct rspamd_config **pcfg;
+	gint orig_top = lua_gettop (L);
+
+	/* First check for global variable 'config' */
+	lua_getglobal (L, "config");
+	if (lua_isnil (L, -1)) {
+		/* Assign global table to set up attributes */
+		lua_newtable (L);
+		lua_setglobal (L, "config");
+	}
+
+	lua_getglobal (L, "metrics");
+	if (lua_isnil (L, -1)) {
+		lua_newtable (L);
+		lua_setglobal (L, "metrics");
+	}
+
+	lua_getglobal (L, "composites");
+	if (lua_isnil (L, -1)) {
+		lua_newtable (L);
+		lua_setglobal (L, "composites");
+	}
+
+	lua_getglobal (L, "rspamd_classifiers");
+	if (lua_isnil (L, -1)) {
+		lua_newtable (L);
+		lua_setglobal (L, "rspamd_classifiers");
+	}
+
+	lua_getglobal (L, "classifiers");
+	if (lua_isnil (L, -1)) {
+		lua_newtable (L);
+		lua_setglobal (L, "classifiers");
+	}
+
+	lua_getglobal (L, "rspamd_version");
+	if (lua_isnil (L, -1)) {
+		lua_pushcfunction (L, rspamd_lua_rspamd_version);
+		lua_setglobal (L, "rspamd_version");
+	}
+
+	if (cfg != NULL) {
+		pcfg = lua_newuserdata (L, sizeof (struct rspamd_config *));
+		rspamd_lua_setclass (L, "rspamd{config}", -1);
+		*pcfg = cfg;
+		lua_setglobal (L, "rspamd_config");
+	}
+
+	lua_settop (L, orig_top);
+
+	/* Set known paths as rspamd_paths global */
+	lua_getglobal (L, "rspamd_paths");
+	if (lua_isnil (L, -1)) {
+		const gchar *confdir = RSPAMD_CONFDIR, *rundir = RSPAMD_RUNDIR,
+				*dbdir = RSPAMD_DBDIR, *logdir = RSPAMD_LOGDIR,
+				*wwwdir = RSPAMD_WWWDIR, *pluginsdir = RSPAMD_PLUGINSDIR,
+				*rulesdir = RSPAMD_RULESDIR, *lualibdir = RSPAMD_LUALIBDIR,
+				*prefix = RSPAMD_PREFIX;
+		const gchar *t;
+
+		/* Try environment */
+		t = getenv ("PLUGINSDIR");
+		if (t) {
+			pluginsdir = t;
+		}
+
+		t = getenv ("RULESDIR");
+		if (t) {
+			rulesdir = t;
+		}
+
+		t = getenv ("DBDIR");
+		if (t) {
+			dbdir = t;
+		}
+
+		t = getenv ("RUNDIR");
+		if (t) {
+			rundir = t;
+		}
+
+		t = getenv ("LUALIBDIR");
+		if (t) {
+			lualibdir = t;
+		}
+
+		t = getenv ("LOGDIR");
+		if (t) {
+			logdir = t;
+		}
+
+		t = getenv ("WWWDIR");
+		if (t) {
+			wwwdir = t;
+		}
+
+		t = getenv ("CONFDIR");
+		if (t) {
+			confdir = t;
+		}
+
+
+		if (vars) {
+			t = g_hash_table_lookup (vars, "PLUGINSDIR");
+			if (t) {
+				pluginsdir = t;
+			}
+
+			t = g_hash_table_lookup (vars, "RULESDIR");
+			if (t) {
+				rulesdir = t;
+			}
+
+			t = g_hash_table_lookup (vars, "LUALIBDIR");
+			if (t) {
+				lualibdir = t;
+			}
+
+			t = g_hash_table_lookup (vars, "RUNDIR");
+			if (t) {
+				rundir = t;
+			}
+
+			t = g_hash_table_lookup (vars, "WWWDIR");
+			if (t) {
+				wwwdir = t;
+			}
+
+			t = g_hash_table_lookup (vars, "CONFDIR");
+			if (t) {
+				confdir = t;
+			}
+
+			t = g_hash_table_lookup (vars, "DBDIR");
+			if (t) {
+				dbdir = t;
+			}
+
+			t = g_hash_table_lookup (vars, "LOGDIR");
+			if (t) {
+				logdir = t;
+			}
+		}
+
+		lua_createtable (L, 0, 9);
+
+		rspamd_lua_table_set (L, RSPAMD_CONFDIR_INDEX, confdir);
+		rspamd_lua_table_set (L, RSPAMD_RUNDIR_INDEX, rundir);
+		rspamd_lua_table_set (L, RSPAMD_DBDIR_INDEX, dbdir);
+		rspamd_lua_table_set (L, RSPAMD_LOGDIR_INDEX, logdir);
+		rspamd_lua_table_set (L, RSPAMD_WWWDIR_INDEX, wwwdir);
+		rspamd_lua_table_set (L, RSPAMD_PLUGINSDIR_INDEX, pluginsdir);
+		rspamd_lua_table_set (L, RSPAMD_RULESDIR_INDEX, rulesdir);
+		rspamd_lua_table_set (L, RSPAMD_LUALIBDIR_INDEX, lualibdir);
+		rspamd_lua_table_set (L, RSPAMD_PREFIX_INDEX, prefix);
+
+		lua_setglobal (L, "rspamd_paths");
+	}
+
+	lua_settop (L, orig_top);
+}
+
+#ifdef WITH_LUA_TRACE
+static gint
+lua_push_trace_data (lua_State *L)
+{
+	if (lua_traces) {
+		ucl_object_push_lua (L, lua_traces, true);
+	}
+	else {
+		lua_pushnil (L);
+	}
+
+	return 1;
+}
+#endif
+
 lua_State *
 rspamd_lua_init ()
 {
@@ -412,6 +741,7 @@ rspamd_lua_init ()
 	luaopen_fann (L);
 	luaopen_sqlite3 (L);
 	luaopen_cryptobox (L);
+	luaopen_dns (L);
 
 	luaL_newmetatable (L, "rspamd{ev_base}");
 	lua_pushstring (L, "class");
@@ -438,7 +768,7 @@ rspamd_lua_init ()
 	lua_getglobal (L, "math");
 	lua_pushstring (L, "randomseed");
 	lua_gettable (L, -2);
-	lua_pushnumber (L, ottery_rand_uint64 ());
+	lua_pushinteger (L, ottery_rand_uint64 ());
 	lua_pcall (L, 1, 0, 0);
 	lua_pop (L, 1); /* math table */
 
@@ -470,6 +800,11 @@ rspamd_lua_init ()
 #undef ADD_TABLE
 	lua_setglobal (L, rspamd_modules_state_global);
 
+#ifdef WITH_LUA_TRACE
+	lua_pushcfunction (L, lua_push_trace_data);
+	lua_setglobal (L, "get_traces");
+#endif
+
 	return L;
 }
 
@@ -587,9 +922,6 @@ rspamd_init_lua_filters (struct rspamd_config *cfg, gboolean force_load)
 		cur = g_list_next (cur);
 	}
 
-	/* Assign state */
-	cfg->lua_state = L;
-
 	return TRUE;
 }
 
@@ -1164,16 +1496,28 @@ rspamd_lua_traceback (lua_State *L)
 {
 
 	GString *tb;
-	const gchar *msg = lua_tostring (L, 1);
 
-	tb = g_string_sized_new (100);
-	g_string_append_printf (tb, "%s; trace:", msg);
-	rspamd_lua_traceback_string (L, tb);
+	tb = rspamd_lua_get_traceback_string (L);
+
 	lua_pushlightuserdata (L, tb);
 
 	return 1;
 }
 
+GString *
+rspamd_lua_get_traceback_string (lua_State *L)
+{
+	GString *tb;
+	const gchar *msg = lua_tostring (L, -1);
+
+	tb = g_string_sized_new (100);
+	g_string_append_printf (tb, "%s; trace:", msg);
+
+	rspamd_lua_traceback_string (L, tb);
+
+	return tb;
+}
+
 guint
 rspamd_lua_table_size (lua_State *L, gint tbl_pos)
 {
@@ -1225,19 +1569,19 @@ err:
 	if (fatal) {
 		const gchar *actual_classname = NULL;
 
-		if (lua_getmetatable (L, pos)) {
+		if (lua_type (L, pos) == LUA_TUSERDATA && lua_getmetatable (L, pos)) {
 			lua_pushstring (L, "__index");
-			lua_gettable (L, -3);
+			lua_gettable (L, -2);
 			lua_pushstring (L, "class");
 			lua_gettable (L, -2);
 			actual_classname = lua_tostring (L, -1);
 		}
 		else {
-			actual_classname = lua_typename (L, lua_type (L, -1));
+			actual_classname = lua_typename (L, lua_type (L, pos));
 		}
 
 		err_msg = g_string_sized_new (100);
-		rspamd_printf_gstring (err_msg, "expected %s at %d, but userdata has "
+		rspamd_printf_gstring (err_msg, "expected %s at position %d, but userdata has "
 										"%s metatable; trace: ",
 				classname, pos, actual_classname);
 		rspamd_lua_traceback_string (L, err_msg);
@@ -1245,15 +1589,21 @@ err:
 
 		for (i = 1; i <= MIN (top, 10); i ++) {
 			if (lua_type (L, i) == LUA_TUSERDATA) {
-				lua_getmetatable (L, i);
-				lua_pushstring (L, "__index");
-				lua_gettable (L, -3);
-				lua_pushstring (L, "class");
-				lua_gettable (L, -2);
+				const char *clsname;
+
+				if (lua_getmetatable (L, i)) {
+					lua_pushstring (L, "__index");
+					lua_gettable (L, -2);
+					lua_pushstring (L, "class");
+					lua_gettable (L, -2);
+					clsname = lua_tostring (L, -1);
+				}
+				else {
+					clsname = lua_typename (L, lua_type (L, i));
+				}
 
 				rspamd_printf_gstring (err_msg, "[%d: ud=%s] ", i,
-						lua_tostring (L, -1));
-				lua_pop (L, 3);
+						clsname);
 			}
 			else {
 				rspamd_printf_gstring (err_msg, "[%d: %s] ", i,
@@ -1298,7 +1648,9 @@ lua_check_ev_base (lua_State * L, gint pos)
 	return ud ? *((struct event_base **)ud) : NULL;
 }
 
-gboolean
+static void rspamd_lua_run_postloads_error (struct thread_entry *thread, int ret, const char *msg);
+
+void
 rspamd_lua_run_postloads (lua_State *L, struct rspamd_config *cfg,
 		struct event_base *ev_base, struct rspamd_worker *w)
 {
@@ -1306,42 +1658,40 @@ rspamd_lua_run_postloads (lua_State *L, struct rspamd_config *cfg,
 	struct rspamd_config **pcfg;
 	struct event_base **pev_base;
 	struct rspamd_worker **pw;
-	gint err_idx;
-	GString *tb;
 
 	/* Execute post load scripts */
 	LL_FOREACH (cfg->on_load, sc) {
-		lua_pushcfunction (L, &rspamd_lua_traceback);
-		err_idx = lua_gettop (L);
+		struct thread_entry *thread = lua_thread_pool_get_for_config (cfg);
+		thread->error_callback = rspamd_lua_run_postloads_error;
+		L = thread->lua_state;
 
-		lua_rawgeti (cfg->lua_state, LUA_REGISTRYINDEX, sc->cbref);
-		pcfg = lua_newuserdata (cfg->lua_state, sizeof (*pcfg));
+		lua_rawgeti (L, LUA_REGISTRYINDEX, sc->cbref);
+		pcfg = lua_newuserdata (L, sizeof (*pcfg));
 		*pcfg = cfg;
-		rspamd_lua_setclass (cfg->lua_state, "rspamd{config}", -1);
+		rspamd_lua_setclass (L, "rspamd{config}", -1);
 
-		pev_base = lua_newuserdata (cfg->lua_state, sizeof (*pev_base));
+		pev_base = lua_newuserdata (L, sizeof (*pev_base));
 		*pev_base = ev_base;
-		rspamd_lua_setclass (cfg->lua_state, "rspamd{ev_base}", -1);
+		rspamd_lua_setclass (L, "rspamd{ev_base}", -1);
 
-		pw = lua_newuserdata (cfg->lua_state, sizeof (*pw));
+		pw = lua_newuserdata (L, sizeof (*pw));
 		*pw = w;
-		rspamd_lua_setclass (cfg->lua_state, "rspamd{worker}", -1);
+		rspamd_lua_setclass (L, "rspamd{worker}", -1);
 
-		if (lua_pcall (cfg->lua_state, 3, 0, err_idx) != 0) {
-			tb = lua_touserdata (L, -1);
-			msg_err_config ("error executing post load code: %v", tb);
-			g_string_free (tb, TRUE);
-			lua_settop (L, err_idx - 1);
-
-			return FALSE;
-		}
-
-		lua_settop (L, err_idx - 1);
+		lua_thread_call (thread, 3);
 	}
-
-	return TRUE;
 }
 
+
+static void
+rspamd_lua_run_postloads_error (struct thread_entry *thread, int ret, const char *msg)
+{
+	struct rspamd_config *cfg = thread->cfg;
+
+	msg_err_config ("error executing post load code: %s", msg);
+}
+
+
 static struct rspamd_worker *
 lua_check_worker (lua_State *L, gint pos)
 {
@@ -1455,7 +1805,7 @@ lua_worker_get_index (lua_State *L)
 	struct rspamd_worker *w = lua_check_worker (L, 1);
 
 	if (w) {
-		lua_pushnumber (L, w->index);
+		lua_pushinteger (L, w->index);
 	}
 	else {
 		return luaL_error (L, "invalid arguments");
@@ -1470,7 +1820,7 @@ lua_worker_get_pid (lua_State *L)
 	struct rspamd_worker *w = lua_check_worker (L, 1);
 
 	if (w) {
-		lua_pushnumber (L, w->pid);
+		lua_pushinteger (L, w->pid);
 	}
 	else {
 		return luaL_error (L, "invalid arguments");
