@@ -22,7 +22,7 @@
 
 #define LUA_ENUM(L, name, val) \
 	lua_pushlstring (L, # name, sizeof(# name) - 1); \
-	lua_pushnumber (L, val); \
+	lua_pushinteger (L, val); \
 	lua_settable (L, -3);
 
 #if LUA_VERSION_NUM > 501 && !defined LUA_COMPAT_MODULE
@@ -208,18 +208,22 @@ struct rspamd_lua_ip * lua_check_ip (lua_State * L, gint pos);
 
 struct rspamd_lua_text * lua_check_text (lua_State * L, gint pos);
 
+enum rspamd_lua_task_header_type {
+	RSPAMD_TASK_HEADER_PUSH_SIMPLE = 0,
+	RSPAMD_TASK_HEADER_PUSH_RAW,
+	RSPAMD_TASK_HEADER_PUSH_FULL,
+	RSPAMD_TASK_HEADER_PUSH_COUNT,
+};
 
 gint rspamd_lua_push_header (lua_State *L,
-		struct rspamd_mime_header *h,
-		gboolean full,
-		gboolean raw);
+							 struct rspamd_mime_header *h,
+							 enum rspamd_lua_task_header_type how);
 /**
  * Push specific header to lua
  */
 gint rspamd_lua_push_header_array (lua_State *L,
-		GPtrArray *hdrs,
-		gboolean full,
-		gboolean raw);
+								   GPtrArray *hdrs,
+								   enum rspamd_lua_task_header_type how);
 
 /**
  * Check for task at the specified position
@@ -276,6 +280,7 @@ void luaopen_html (lua_State * L);
 void luaopen_fann (lua_State *L);
 void luaopen_sqlite3 (lua_State *L);
 void luaopen_cryptobox (lua_State *L);
+void luaopen_dns (lua_State *L);
 
 void rspamd_lua_dostring (const gchar *line);
 
@@ -298,6 +303,10 @@ void rspamd_lua_dumpstack (lua_State *L);
 void rspamd_lua_set_path (lua_State *L, const ucl_object_t *cfg_obj,
 		GHashTable *vars);
 
+/* Set some lua globals */
+void rspamd_lua_set_globals (struct rspamd_config *cfg, lua_State *L,
+							 GHashTable *vars);
+
 struct memory_pool_s * rspamd_lua_check_mempool (lua_State * L, gint pos);
 struct rspamd_config * lua_check_config (lua_State * L, gint pos);
 struct rspamd_async_session* lua_check_session (lua_State * L, gint pos);
@@ -328,6 +337,14 @@ gboolean rspamd_lua_parse_table_arguments (lua_State *L, gint pos,
 
 gint rspamd_lua_traceback (lua_State *L);
 
+/**
+ * Returns stack trace as a string. Caller should clear memory.
+ * @param L
+ * @return
+ */
+GString *
+rspamd_lua_get_traceback_string (lua_State *L);
+
 /**
  * Returns size of table at position `tbl_pos`
  */
@@ -335,6 +352,14 @@ guint rspamd_lua_table_size (lua_State *L, gint tbl_pos);
 
 void lua_push_emails_address_list (lua_State *L, GPtrArray *addrs);
 
+
+#define TRACE_POINTS 6
+
+struct lua_logger_trace {
+	gint cur_level;
+	gconstpointer traces[TRACE_POINTS];
+};
+
 /**
  * Log lua object to string
  * @param L
@@ -344,7 +369,7 @@ void lua_push_emails_address_list (lua_State *L, GPtrArray *addrs);
  * @return
  */
 gsize lua_logger_out_type (lua_State *L, gint pos, gchar *outbuf,
-		gsize len);
+		gsize len, struct lua_logger_trace *trace);
 
 /**
  * Safely checks userdata to match specified class
@@ -364,12 +389,10 @@ void *rspamd_lua_check_udata_maybe (lua_State *L, gint pos, const gchar *classna
 
 /**
  * Call finishing script with the specified task
- * @param L
  * @param sc
  * @param task
  */
-void lua_call_finish_script (lua_State *L, struct
-		rspamd_config_post_load_script *sc,
+void lua_call_finish_script (struct rspamd_config_post_load_script *sc,
 		struct rspamd_task *task);
 
 /**
@@ -378,7 +401,7 @@ void lua_call_finish_script (lua_State *L, struct
  * @param cfg
  * @param ev_base
  */
-gboolean rspamd_lua_run_postloads (lua_State *L, struct rspamd_config *cfg,
+void rspamd_lua_run_postloads (lua_State *L, struct rspamd_config *cfg,
 		struct event_base *ev_base, struct rspamd_worker *w);
 
 /**
@@ -400,5 +423,33 @@ void rspamd_lua_add_ref_dtor (lua_State *L, rspamd_mempool_t *pool,
 gboolean rspamd_lua_require_function (lua_State *L, const gchar *modname,
 		const gchar *funcname);
 
+/* Paths defs */
+#define RSPAMD_CONFDIR_INDEX "CONFDIR"
+#define RSPAMD_RUNDIR_INDEX "RUNDIR"
+#define RSPAMD_DBDIR_INDEX "DBDIR"
+#define RSPAMD_LOGDIR_INDEX "LOGDIR"
+#define RSPAMD_PLUGINSDIR_INDEX "PLUGINSDIR"
+#define RSPAMD_RULESDIR_INDEX "RULESDIR"
+#define RSPAMD_LUALIBDIR_INDEX "LUALIBDIR"
+#define RSPAMD_WWWDIR_INDEX "WWWDIR"
+#define RSPAMD_PREFIX_INDEX "PREFIX"
+#define RSPAMD_VERSION_INDEX "VERSION"
+
+#ifdef WITH_LUA_TRACE
+extern ucl_object_t *lua_traces;
+#define LUA_TRACE_POINT do { \
+ ucl_object_t *func_obj; \
+ if (lua_traces == NULL) { lua_traces = ucl_object_typed_new (UCL_OBJECT); } \
+ func_obj = (ucl_object_t *)ucl_object_lookup (lua_traces, G_STRFUNC); \
+ if (func_obj == NULL) { \
+   func_obj = ucl_object_typed_new (UCL_INT); \
+   ucl_object_insert_key (lua_traces, func_obj, G_STRFUNC, 0, false); \
+ } \
+ func_obj->value.iv ++; \
+} while(0)
+#else
+#define LUA_TRACE_POINT
+#endif
+
 #endif /* WITH_LUA */
 #endif /* RSPAMD_LUA_H */
