@@ -14,6 +14,7 @@
  * limitations under the License.
  */
 #include "lua_common.h"
+#include "lua_thread_pool.h"
 #include "http_private.h"
 #include "unix-std.h"
 #include "zlib.h"
@@ -54,9 +55,9 @@ static const struct luaL_reg httplib_m[] = {
 
 #define RSPAMD_LUA_HTTP_FLAG_TEXT (1 << 0)
 #define RSPAMD_LUA_HTTP_FLAG_NOVERIFY (1 << 1)
+#define RSPAMD_LUA_HTTP_FLAG_RESOLVED (1 << 2)
 
 struct lua_http_cbdata {
-	lua_State *L;
 	struct rspamd_http_connection *conn;
 	struct rspamd_async_session *session;
 	struct rspamd_async_watcher *w;
@@ -69,13 +70,13 @@ struct lua_http_cbdata {
 	rspamd_inet_addr_t *addr;
 	gchar *mime_type;
 	gchar *host;
+	gchar *auth;
 	const gchar *url;
 	gsize max_size;
 	gint flags;
 	gint fd;
 	gint cbref;
-	gint bodyref;
-	gboolean gzip;
+	struct thread_entry *thread;
 };
 
 static const int default_http_timeout = 5000;
@@ -97,7 +98,10 @@ lua_http_fin (gpointer arg)
 {
 	struct lua_http_cbdata *cbd = (struct lua_http_cbdata *)arg;
 
-	luaL_unref (cbd->L, LUA_REGISTRYINDEX, cbd->cbref);
+	if (cbd->cbref != -1) {
+		luaL_unref (cbd->cfg->lua_state, LUA_REGISTRYINDEX, cbd->cbref);
+	}
+
 	if (cbd->conn) {
 		/* Here we already have a connection, so we need to unref it */
 		rspamd_http_connection_unref (cbd->conn);
@@ -123,6 +127,10 @@ lua_http_fin (gpointer arg)
 		g_free (cbd->host);
 	}
 
+	if (cbd->auth) {
+		g_free (cbd->auth);
+	}
+
 	if (cbd->local_kp) {
 		rspamd_keypair_unref (cbd->local_kp);
 	}
@@ -138,8 +146,15 @@ static void
 lua_http_maybe_free (struct lua_http_cbdata *cbd)
 {
 	if (cbd->session) {
-		rspamd_session_watcher_pop (cbd->session, cbd->w);
-		rspamd_session_remove_event (cbd->session, lua_http_fin, cbd);
+		if (cbd->w) {
+			/* We still need to clear watcher */
+			rspamd_session_watcher_pop (cbd->session, cbd->w);
+		}
+
+		if (cbd->flags & RSPAMD_LUA_HTTP_FLAG_RESOLVED) {
+			/* Event is added merely for resolved events */
+			rspamd_session_remove_event (cbd->session, lua_http_fin, cbd);
+		}
 	}
 	else {
 		lua_http_fin (cbd);
@@ -149,21 +164,37 @@ lua_http_maybe_free (struct lua_http_cbdata *cbd)
 static void
 lua_http_push_error (struct lua_http_cbdata *cbd, const char *err)
 {
-	lua_rawgeti (cbd->L, LUA_REGISTRYINDEX, cbd->cbref);
-	lua_pushstring (cbd->L, err);
+	struct lua_callback_state lcbd;
+	lua_State *L;
 
-	if (lua_pcall (cbd->L, 1, 0, 0) != 0) {
-		msg_info ("callback call failed: %s", lua_tostring (cbd->L, -1));
-		lua_pop (cbd->L, 1);
+	lua_thread_pool_prepare_callback (cbd->cfg->lua_thread_pool, &lcbd);
+
+	L = lcbd.L;
+
+	lua_rawgeti (L, LUA_REGISTRYINDEX, cbd->cbref);
+	lua_pushstring (L, err);
+
+	if (lua_pcall (L, 1, 0, 0) != 0) {
+		msg_info ("callback call failed: %s", lua_tostring (L, -1));
+		lua_pop (L, 1);
 	}
+
+	lua_thread_pool_restore_callback (&lcbd);
 }
 
+static void lua_http_resume_handler (struct rspamd_http_connection *conn,
+						 struct rspamd_http_message *msg, const char *err);
+
 static void
 lua_http_error_handler (struct rspamd_http_connection *conn, GError *err)
 {
 	struct lua_http_cbdata *cbd = (struct lua_http_cbdata *)conn->ud;
-
-	lua_http_push_error (cbd, err->message);
+	if (cbd->cbref == -1) {
+		lua_http_resume_handler (conn, NULL, err->message);
+	}
+	else {
+		lua_http_push_error (cbd, err->message);
+	}
 	lua_http_maybe_free (cbd);
 }
 
@@ -173,57 +204,170 @@ lua_http_finish_handler (struct rspamd_http_connection *conn,
 {
 	struct lua_http_cbdata *cbd = (struct lua_http_cbdata *)conn->ud;
 	struct rspamd_http_header *h, *htmp;
+	struct rspamd_async_watcher *existing_watcher = NULL;
 	const gchar *body;
 	gsize body_len;
 
-	lua_rawgeti (cbd->L, LUA_REGISTRYINDEX, cbd->cbref);
+	struct lua_callback_state lcbd;
+	lua_State *L;
+
+	if (cbd->cbref == -1) {
+		lua_http_resume_handler (conn, msg, NULL);
+		lua_http_maybe_free (cbd);
+		return 0;
+	}
+	lua_thread_pool_prepare_callback (cbd->cfg->lua_thread_pool, &lcbd);
+
+	L = lcbd.L;
+
+	lua_rawgeti (L, LUA_REGISTRYINDEX, cbd->cbref);
 	/* Error */
-	lua_pushnil (cbd->L);
+	lua_pushnil (L);
 	/* Reply code */
-	lua_pushnumber (cbd->L, msg->code);
+	lua_pushinteger (L, msg->code);
 	/* Body */
 	body = rspamd_http_message_get_body (msg, &body_len);
 
 	if (cbd->flags & RSPAMD_LUA_HTTP_FLAG_TEXT) {
 		struct rspamd_lua_text *t;
 
-		t = lua_newuserdata (cbd->L, sizeof (*t));
-		rspamd_lua_setclass (cbd->L, "rspamd{text}", -1);
+		t = lua_newuserdata (L, sizeof (*t));
+		rspamd_lua_setclass (L, "rspamd{text}", -1);
 		t->start = body;
 		t->len = body_len;
 		t->flags = 0;
 	}
 	else {
 		if (body_len > 0) {
-			lua_pushlstring (cbd->L, body, body_len);
+			lua_pushlstring (L, body, body_len);
 		}
 		else {
-			lua_pushnil (cbd->L);
+			lua_pushnil (L);
 		}
 	}
 	/* Headers */
-	lua_newtable (cbd->L);
+	lua_newtable (L);
 
 	HASH_ITER (hh, msg->headers, h, htmp) {
 		/*
 		 * Lowercase header name, as Lua cannot search in caseless matter
 		 */
 		rspamd_str_lc (h->combined->str, h->name.len);
-		lua_pushlstring (cbd->L, h->name.begin, h->name.len);
-		lua_pushlstring (cbd->L, h->value.begin, h->value.len);
-		lua_settable (cbd->L, -3);
+		lua_pushlstring (L, h->name.begin, h->name.len);
+		lua_pushlstring (L, h->value.begin, h->value.len);
+		lua_settable (L, -3);
 	}
 
-	if (lua_pcall (cbd->L, 4, 0, 0) != 0) {
-		msg_info ("callback call failed: %s", lua_tostring (cbd->L, -1));
-		lua_pop (cbd->L, 1);
+	if (cbd->w) {
+		/* Replace watcher to deal with nested calls */
+		existing_watcher = rspamd_session_replace_watcher (cbd->session, cbd->w);
+	}
+
+	if (lua_pcall (L, 4, 0, 0) != 0) {
+		msg_info ("callback call failed: %s", lua_tostring (L, -1));
+		lua_pop (L, 1);
+	}
+
+	if (cbd->w) {
+		/* Restore existing watcher */
+		rspamd_session_replace_watcher (cbd->session, existing_watcher);
 	}
 
 	lua_http_maybe_free (cbd);
 
+	lua_thread_pool_restore_callback (&lcbd);
+
 	return 0;
 }
 
+/*
+ * resumes yielded thread
+ */
+static void
+lua_http_resume_handler (struct rspamd_http_connection *conn,
+						 struct rspamd_http_message *msg, const char *err)
+{
+	struct lua_http_cbdata *cbd = (struct lua_http_cbdata *)conn->ud;
+	lua_State *L = cbd->thread->lua_state;
+	const gchar *body;
+	gsize body_len;
+	struct rspamd_http_header *h, *htmp;
+	struct rspamd_async_watcher *existing_watcher = NULL;
+
+	if (err) {
+		lua_pushstring (L, err);
+		lua_pushnil (L);
+	}
+	else {
+		/*
+		 * 1 - nil (error)
+		 * 2 - table:
+		 *   code (int)
+		 *   content (string)
+		 *   headers (table: header -> value)
+		 */
+		lua_pushnil (L); // error code
+
+		lua_createtable (L, 0, 3);
+
+		/* code */
+		lua_pushliteral (L, "code");
+		lua_pushinteger (L, msg->code);
+		lua_settable (L, -3);
+
+		/* content */
+		lua_pushliteral (L, "content");
+
+		body = rspamd_http_message_get_body (msg, &body_len);
+		if (cbd->flags & RSPAMD_LUA_HTTP_FLAG_TEXT) {
+			struct rspamd_lua_text *t;
+
+			t = lua_newuserdata (L, sizeof (*t));
+			rspamd_lua_setclass (L, "rspamd{text}", -1);
+			t->start = body;
+			t->len = body_len;
+			t->flags = 0;
+		}
+		else {
+			if (body_len > 0) {
+				lua_pushlstring (L, body, body_len);
+			}
+			else {
+				lua_pushnil (L);
+			}
+		}
+		lua_settable (L, -3);
+
+		/* headers */
+		lua_pushliteral (L, "headers");
+		lua_newtable (L);
+
+		HASH_ITER (hh, msg->headers, h, htmp) {
+			/*
+			 * Lowercase header name, as Lua cannot search in caseless matter
+			 */
+			rspamd_str_lc (h->combined->str, h->name.len);
+			lua_pushlstring (L, h->name.begin, h->name.len);
+			lua_pushlstring (L, h->value.begin, h->value.len);
+			lua_settable (L, -3);
+		}
+
+		lua_settable (L, -3);
+	}
+
+	if (cbd->w) {
+		/* Replace watcher to deal with nested calls */
+		existing_watcher = rspamd_session_replace_watcher (cbd->session, cbd->w);
+	}
+
+	lua_thread_resume (cbd->thread, 2);
+
+	if (cbd->w) {
+		/* Restore existing watcher */
+		rspamd_session_replace_watcher (cbd->session, existing_watcher);
+	}
+}
+
 static gboolean
 lua_http_make_connection (struct lua_http_cbdata *cbd)
 {
@@ -245,7 +389,8 @@ lua_http_make_connection (struct lua_http_cbdata *cbd)
 				RSPAMD_HTTP_CLIENT_SIMPLE,
 				RSPAMD_HTTP_CLIENT,
 				NULL,
-				cbd->cfg->libs_ctx->ssl_ctx);
+				(cbd->flags & RSPAMD_LUA_HTTP_FLAG_NOVERIFY) ?
+				cbd->cfg->libs_ctx->ssl_ctx_noverify : cbd->cfg->libs_ctx->ssl_ctx);
 	}
 	else {
 		cbd->conn = rspamd_http_connection_new (NULL,
@@ -274,12 +419,24 @@ lua_http_make_connection (struct lua_http_cbdata *cbd)
 			rspamd_http_connection_set_max_size (cbd->conn, cbd->max_size);
 		}
 
+		if (cbd->auth) {
+			rspamd_http_message_add_header (cbd->msg, "Authorization",
+					cbd->auth);
+		}
+
 		rspamd_http_connection_write_message (cbd->conn, cbd->msg,
 				cbd->host, cbd->mime_type, cbd, fd,
 				&cbd->tv, cbd->ev_base);
 		/* Message is now owned by a connection object */
 		cbd->msg = NULL;
 
+		if (cbd->session) {
+			rspamd_session_add_event (cbd->session, cbd->w,
+					(event_finalizer_t) lua_http_fin, cbd,
+					g_quark_from_static_string ("lua http"));
+			cbd->flags |= RSPAMD_LUA_HTTP_FLAG_RESOLVED;
+		}
+
 		return TRUE;
 	}
 
@@ -363,11 +520,12 @@ lua_http_push_headers (lua_State *L, struct rspamd_http_message *msg)
 static gint
 lua_http_request (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct event_base *ev_base;
 	struct rspamd_http_message *msg;
 	struct lua_http_cbdata *cbd;
 	struct rspamd_dns_resolver *resolver;
-	struct rspamd_async_session *session;
+	struct rspamd_async_session *session = NULL;
 	struct rspamd_lua_text *t;
 	struct rspamd_task *task = NULL;
 	struct rspamd_config *cfg = NULL;
@@ -376,11 +534,12 @@ lua_http_request (lua_State *L)
 	const gchar *url, *lua_body;
 	rspamd_fstring_t *body = NULL;
 	gchar *to_resolve;
-	gint cbref;
+	gint cbref = -1;
 	gsize bodylen;
 	gdouble timeout = default_http_timeout;
 	gint flags = 0;
 	gchar *mime_type = NULL;
+	gchar *auth = NULL;
 	gsize max_size = 0;
 	gboolean gzip = FALSE;
 
@@ -435,11 +594,9 @@ lua_http_request (lua_State *L)
 		lua_gettable (L, 1);
 		if (url == NULL || lua_type (L, -1) != LUA_TFUNCTION) {
 			lua_pop (L, 1);
-			msg_err ("http request has bad params");
-			lua_pushboolean (L, FALSE);
-			return 1;
+		} else {
+			cbref = luaL_ref (L, LUA_REGISTRYINDEX);
 		}
-		cbref = luaL_ref (L, LUA_REGISTRYINDEX);
 
 		lua_pushstring (L, "task");
 		lua_gettable (L, 1);
@@ -543,6 +700,39 @@ lua_http_request (lua_State *L)
 			if (t) {
 				body = rspamd_fstring_new_init (t->start, t->len);
 			}
+			else {
+				return luaL_error (L, "invalid body argument type: %s",
+						lua_typename (L, lua_type (L, -1)));
+			}
+		}
+		else if (lua_type (L, -1) == LUA_TTABLE) {
+			body = rspamd_fstring_new ();
+
+			for (lua_pushnil (L); lua_next (L, -2); lua_pop (L, 1)) {
+				if (lua_type (L, -1) == LUA_TSTRING) {
+					lua_body = lua_tolstring (L, -1, &bodylen);
+					body = rspamd_fstring_append (body, lua_body, bodylen);
+				}
+				else if (lua_type (L, -1) == LUA_TUSERDATA) {
+					t = lua_check_text (L, -1);
+
+					if (t) {
+						body = rspamd_fstring_append (body, t->start, t->len);
+					}
+					else {
+						return luaL_error (L, "invalid body argument: %s",
+								lua_typename (L, lua_type (L, -1)));
+					}
+				}
+				else {
+					return luaL_error (L, "invalid body argument type: %s",
+							lua_typename (L, lua_type (L, -1)));
+				}
+			}
+		}
+		else if (lua_type (L, -1) != LUA_TNONE && lua_type (L, -1) != LUA_TNIL) {
+			return luaL_error (L, "invalid body argument type: %s",
+					lua_typename (L, lua_type (L, -1)));
 		}
 		lua_pop (L, 1);
 
@@ -616,6 +806,38 @@ lua_http_request (lua_State *L)
 		}
 
 		lua_pop (L, 1);
+
+		lua_pushstring (L, "user");
+		lua_gettable (L, 1);
+
+		if (lua_type (L, -1) == LUA_TSTRING) {
+			const gchar *user = lua_tostring (L, -1);
+
+			lua_pushstring (L, "password");
+			lua_gettable (L, 1);
+
+			if (lua_type (L, -1) == LUA_TSTRING) {
+				const gchar *password = lua_tostring (L, -1);
+				gchar *tmpbuf;
+				gsize tlen;
+
+				tlen = strlen (user) + strlen (password) + 1;
+				tmpbuf = g_malloc (tlen + 1);
+				rspamd_snprintf (tmpbuf, tlen + 1, "%s:%s", user, password);
+				tlen *= 2;
+				tlen += sizeof ("Basic ") - 1;
+				auth = g_malloc (tlen + 1);
+				rspamd_snprintf (auth, tlen + 1, "Basic %Bs", tmpbuf);
+				g_free (tmpbuf);
+			}
+			else {
+				msg_warn ("HTTP user must have password, disabling auth");
+			}
+
+			lua_pop (L, 1); /* password */
+		}
+
+		lua_pop (L, 1); /* username */
 	}
 	else {
 		msg_err ("http request has bad params");
@@ -624,8 +846,16 @@ lua_http_request (lua_State *L)
 		return 1;
 	}
 
+	if (session && rspamd_session_blocked (session)) {
+		lua_pushboolean (L, FALSE);
+
+		return 1;
+	}
+	if (task == NULL && cfg == NULL) {
+		return luaL_error (L, "Bad params to rspamd_http:request(): either task or config should be set");
+	}
+
 	cbd = g_malloc0 (sizeof (*cbd));
-	cbd->L = L;
 	cbd->cbref = cbref;
 	cbd->msg = msg;
 	cbd->ev_base = ev_base;
@@ -638,6 +868,7 @@ lua_http_request (lua_State *L)
 	cbd->flags = flags;
 	cbd->max_size = max_size;
 	cbd->url = url;
+	cbd->auth = auth;
 
 	if (msg->host) {
 		cbd->host = rspamd_fstring_cstr (msg->host);
@@ -653,19 +884,11 @@ lua_http_request (lua_State *L)
 		rspamd_http_message_set_body_from_fstring_steal (msg, body);
 	}
 
-	if (gzip) {
-		cbd->gzip = TRUE;
-		/* TODO: Add client support for gzip */
-	}
-
 	if (session) {
 		cbd->session = session;
-		rspamd_session_add_event (session,
-				(event_finalizer_t)lua_http_fin,
-				cbd,
-				g_quark_from_static_string ("lua http"));
-		cbd->w = rspamd_session_get_watcher (session);
-		rspamd_session_watcher_push_specific (session, cbd->w);
+
+		cbd->w = rspamd_session_get_watcher (cbd->session);
+		rspamd_session_watcher_push_specific (cbd->session, cbd->w);
 	}
 
 	if (rspamd_parse_inet_address (&cbd->addr, msg->host->str, msg->host->len)) {
@@ -674,6 +897,10 @@ lua_http_request (lua_State *L)
 			lua_http_maybe_free (cbd);
 			lua_pushboolean (L, FALSE);
 
+			if (cbd->w) {
+				rspamd_session_watcher_pop (cbd->session, cbd->w);
+			}
+
 			return 1;
 		}
 	}
@@ -689,9 +916,14 @@ lua_http_request (lua_State *L)
 				lua_pushboolean (L, FALSE);
 				g_free (to_resolve);
 
+				if (cbd->w) {
+					rspamd_session_watcher_pop (cbd->session, cbd->w);
+				}
+
 				return 1;
 			}
 
+
 			g_free (to_resolve);
 		}
 		else {
@@ -702,12 +934,24 @@ lua_http_request (lua_State *L)
 				lua_http_maybe_free (cbd);
 				lua_pushboolean (L, FALSE);
 
+				if (cbd->w) {
+					rspamd_session_watcher_pop (cbd->session, cbd->w);
+				}
+
 				return 1;
 			}
 		}
 	}
 
-	lua_pushboolean (L, TRUE);
+	if (cbd->cbref == -1) {
+		cbd->thread = lua_thread_pool_get_running_entry (cfg->lua_thread_pool);
+
+		return lua_thread_yield (cbd->thread, 0);
+	}
+	else {
+		lua_pushboolean (L, TRUE);
+	}
+
 	return 1;
 }
 
