@@ -58,6 +58,7 @@
 #define RSPAMD_FUZZY_PLUGIN_VERSION RSPAMD_FUZZY_VERSION
 
 static const gint rspamd_fuzzy_hash_len = 5;
+struct fuzzy_ctx;
 
 struct fuzzy_mapping {
 	guint64 fuzzy_flag;
@@ -91,6 +92,7 @@ struct fuzzy_rule {
 	gboolean short_text_direct_hash;
 	gint learn_condition_cb;
 	struct rspamd_hash_map_helper *skip_map;
+	struct fuzzy_ctx *ctx;
 };
 
 struct fuzzy_ctx {
@@ -172,7 +174,7 @@ struct fuzzy_cmd_io {
 	struct iovec io;
 };
 
-static struct fuzzy_ctx *fuzzy_module_ctx = NULL;
+
 static const char *default_headers = "Subject,Content-Type,Reply-To,X-Mailer";
 
 static void fuzzy_symbol_callback (struct rspamd_task *task, void *unused);
@@ -188,19 +190,27 @@ static gint fuzzy_lua_learn_handler (lua_State *L);
 static gint fuzzy_lua_unlearn_handler (lua_State *L);
 
 module_t fuzzy_check_module = {
-	"fuzzy_check",
-	fuzzy_check_module_init,
-	fuzzy_check_module_config,
-	fuzzy_check_module_reconfig,
-	fuzzy_attach_controller,
-	RSPAMD_MODULE_VER
+		"fuzzy_check",
+		fuzzy_check_module_init,
+		fuzzy_check_module_config,
+		fuzzy_check_module_reconfig,
+		fuzzy_attach_controller,
+		RSPAMD_MODULE_VER,
+		(guint)-1,
 };
 
+static inline struct fuzzy_ctx *
+fuzzy_get_context (struct rspamd_config *cfg)
+{
+	return (struct fuzzy_ctx *)g_ptr_array_index (cfg->c_modules,
+			fuzzy_check_module.ctx_offset);
+}
+
 static void
 parse_flags (struct fuzzy_rule *rule,
-	struct rspamd_config *cfg,
-	const ucl_object_t *val,
-	gint cb_id)
+			 struct rspamd_config *cfg,
+			 const ucl_object_t *val,
+			 gint cb_id)
 {
 	const ucl_object_t *elt;
 	struct fuzzy_mapping *map;
@@ -217,7 +227,7 @@ parse_flags (struct fuzzy_rule *rule,
 		}
 		if (sym != NULL) {
 			map =
-				rspamd_mempool_alloc (fuzzy_module_ctx->fuzzy_pool,
+				rspamd_mempool_alloc (cfg->cfg_pool,
 					sizeof (struct fuzzy_mapping));
 			map->symbol = sym;
 			elt = ucl_object_lookup (val, "flag");
@@ -257,7 +267,7 @@ parse_flags (struct fuzzy_rule *rule,
 
 
 static GPtrArray *
-parse_mime_types (const gchar *str)
+parse_mime_types (struct rspamd_config *cfg, const gchar *str)
 {
 	gchar **strvec, *p;
 	gint num, i;
@@ -272,24 +282,24 @@ parse_mime_types (const gchar *str)
 		g_strstrip (strvec[i]);
 
 		if ((p = strchr (strvec[i], '/')) != NULL) {
-			type = rspamd_mempool_alloc (fuzzy_module_ctx->fuzzy_pool,
+			type = rspamd_mempool_alloc (cfg->cfg_pool,
 					sizeof (struct fuzzy_mime_type));
 			type->type_re = rspamd_regexp_from_glob (strvec[i], p - strvec[i],
 					NULL);
 			type->subtype_re = rspamd_regexp_from_glob (p + 1, 0, NULL);
-			rspamd_mempool_add_destructor (fuzzy_module_ctx->fuzzy_pool,
+			rspamd_mempool_add_destructor (cfg->cfg_pool,
 						(rspamd_mempool_destruct_t)rspamd_regexp_unref,
 						type->type_re);
-			rspamd_mempool_add_destructor (fuzzy_module_ctx->fuzzy_pool,
+			rspamd_mempool_add_destructor (cfg->cfg_pool,
 					(rspamd_mempool_destruct_t)rspamd_regexp_unref,
 					type->subtype_re);
 			g_ptr_array_add (res, type);
 		}
 		else {
-			type = rspamd_mempool_alloc (fuzzy_module_ctx->fuzzy_pool,
+			type = rspamd_mempool_alloc (cfg->cfg_pool,
 							sizeof (struct fuzzy_mime_type));
 			type->type_re = rspamd_regexp_from_glob (strvec[i], 0, NULL);
-			rspamd_mempool_add_destructor (fuzzy_module_ctx->fuzzy_pool,
+			rspamd_mempool_add_destructor (cfg->cfg_pool,
 					(rspamd_mempool_destruct_t)rspamd_regexp_unref,
 					type->type_re);
 			type->subtype_re = NULL;
@@ -303,7 +313,7 @@ parse_mime_types (const gchar *str)
 }
 
 static GPtrArray *
-parse_fuzzy_headers (const gchar *str)
+parse_fuzzy_headers (struct rspamd_config *cfg, const gchar *str)
 {
 	gchar **strvec;
 	gint num, i;
@@ -316,7 +326,7 @@ parse_fuzzy_headers (const gchar *str)
 	for (i = 0; i < num; i++) {
 		g_strstrip (strvec[i]);
 		g_ptr_array_add (res, rspamd_mempool_strdup (
-				fuzzy_module_ctx->fuzzy_pool, strvec[i]));
+				cfg->cfg_pool, strvec[i]));
 	}
 
 	g_strfreev (strvec);
@@ -412,6 +422,7 @@ fuzzy_parse_rule (struct rspamd_config *cfg, const ucl_object_t *obj,
 	struct fuzzy_rule *rule;
 	ucl_object_iter_t it = NULL;
 	const char *k = NULL, *key_str = NULL, *shingles_key_str = NULL, *lua_script;
+	struct fuzzy_ctx *fuzzy_module_ctx = fuzzy_get_context (cfg);
 
 	if (obj->type != UCL_OBJECT) {
 		msg_err_config ("invalid rule definition");
@@ -419,20 +430,19 @@ fuzzy_parse_rule (struct rspamd_config *cfg, const ucl_object_t *obj,
 	}
 
 	rule = fuzzy_rule_new (fuzzy_module_ctx->default_symbol,
-			fuzzy_module_ctx->fuzzy_pool);
+			cfg->cfg_pool);
+	rule->ctx = fuzzy_module_ctx;
 	rule->learn_condition_cb = -1;
 	rule->alg = RSPAMD_SHINGLES_OLD;
+	rule->skip_map = NULL;
 
 	if ((value = ucl_object_lookup (obj, "skip_hashes")) != NULL) {
 		rspamd_map_add_from_ucl (cfg, value,
-			"Fuzzy hashes whitelist", rspamd_kv_list_read, rspamd_kv_list_fin,
-			(void **)&rule->skip_map);
-		rspamd_mempool_add_destructor (fuzzy_module_ctx->fuzzy_pool,
-				(rspamd_mempool_destruct_t)rspamd_map_helper_destroy_radix,
-				rule->skip_map);
-	}
-	else {
-		rule->skip_map = NULL;
+				"Fuzzy hashes whitelist",
+				rspamd_kv_list_read,
+				rspamd_kv_list_fin,
+				rspamd_kv_list_dtor,
+				(void **)&rule->skip_map);
 	}
 
 	if ((value = ucl_object_lookup (obj, "mime_types")) != NULL) {
@@ -443,7 +453,7 @@ fuzzy_parse_rule (struct rspamd_config *cfg, const ucl_object_t *obj,
 			guint i;
 			gpointer ptr;
 
-			tmp = parse_mime_types (ucl_obj_tostring (cur));
+			tmp = parse_mime_types (cfg, ucl_obj_tostring (cur));
 
 			if (tmp) {
 				if (rule->mime_types) {
@@ -460,7 +470,7 @@ fuzzy_parse_rule (struct rspamd_config *cfg, const ucl_object_t *obj,
 		}
 
 		if (rule->mime_types) {
-			rspamd_mempool_add_destructor (fuzzy_module_ctx->fuzzy_pool,
+			rspamd_mempool_add_destructor (cfg->cfg_pool,
 						rspamd_ptr_array_free_hard, rule->mime_types);
 		}
 	}
@@ -473,7 +483,7 @@ fuzzy_parse_rule (struct rspamd_config *cfg, const ucl_object_t *obj,
 			guint i;
 			gpointer ptr;
 
-			tmp = parse_fuzzy_headers (ucl_obj_tostring (cur));
+			tmp = parse_fuzzy_headers (cfg, ucl_obj_tostring (cur));
 
 			if (tmp) {
 				if (rule->fuzzy_headers) {
@@ -490,11 +500,11 @@ fuzzy_parse_rule (struct rspamd_config *cfg, const ucl_object_t *obj,
 		}
 	}
 	else {
-		rule->fuzzy_headers = parse_fuzzy_headers (default_headers);
+		rule->fuzzy_headers = parse_fuzzy_headers (cfg, default_headers);
 	}
 
 	if (rule->fuzzy_headers != NULL) {
-		rspamd_mempool_add_destructor (fuzzy_module_ctx->fuzzy_pool,
+		rspamd_mempool_add_destructor (cfg->cfg_pool,
 				(rspamd_mempool_destruct_t) rspamd_ptr_array_free_hard,
 				rule->fuzzy_headers);
 	}
@@ -580,7 +590,7 @@ fuzzy_parse_rule (struct rspamd_config *cfg, const ucl_object_t *obj,
 	if ((value = ucl_object_lookup (obj, "servers")) != NULL) {
 		rule->servers = rspamd_upstreams_create (cfg->ups_ctx);
 
-		rspamd_mempool_add_destructor (fuzzy_module_ctx->fuzzy_pool,
+		rspamd_mempool_add_destructor (cfg->cfg_pool,
 				(rspamd_mempool_destruct_t)rspamd_upstreams_destroy,
 				rule->servers);
 		if (!rspamd_upstreams_from_ucl (rule->servers, value, DEFAULT_PORT, NULL)) {
@@ -689,7 +699,7 @@ fuzzy_parse_rule (struct rspamd_config *cfg, const ucl_object_t *obj,
 				rule->algorithm_str);
 	}
 
-	rspamd_mempool_add_destructor (fuzzy_module_ctx->fuzzy_pool, fuzzy_free_rule,
+	rspamd_mempool_add_destructor (cfg->cfg_pool, fuzzy_free_rule,
 			rule);
 
 	return 0;
@@ -698,17 +708,27 @@ fuzzy_parse_rule (struct rspamd_config *cfg, const ucl_object_t *obj,
 gint
 fuzzy_check_module_init (struct rspamd_config *cfg, struct module_ctx **ctx)
 {
-	if (fuzzy_module_ctx == NULL) {
-		fuzzy_module_ctx = g_malloc0 (sizeof (struct fuzzy_ctx));
+	struct fuzzy_ctx *fuzzy_module_ctx;
 
-		fuzzy_module_ctx->fuzzy_pool = rspamd_mempool_new (rspamd_mempool_suggest_size (), NULL);
-		/* TODO: this should match rules count actually */
-		fuzzy_module_ctx->keypairs_cache = rspamd_keypair_cache_new (32);
-		fuzzy_module_ctx->fuzzy_rules = g_ptr_array_new ();
-	}
+	fuzzy_module_ctx = rspamd_mempool_alloc0 (cfg->cfg_pool,
+			sizeof (struct fuzzy_ctx));
 
+	fuzzy_module_ctx->fuzzy_pool = rspamd_mempool_new (rspamd_mempool_suggest_size (), NULL);
+	/* TODO: this should match rules count actually */
+	fuzzy_module_ctx->keypairs_cache = rspamd_keypair_cache_new (32);
+	fuzzy_module_ctx->fuzzy_rules = g_ptr_array_new ();
 	fuzzy_module_ctx->cfg = cfg;
 
+	rspamd_mempool_add_destructor (cfg->cfg_pool,
+			(rspamd_mempool_destruct_t)rspamd_mempool_delete,
+			fuzzy_module_ctx->fuzzy_pool);
+	rspamd_mempool_add_destructor (cfg->cfg_pool,
+			(rspamd_mempool_destruct_t)rspamd_keypair_cache_destroy,
+			fuzzy_module_ctx->keypairs_cache);
+	rspamd_mempool_add_destructor (cfg->cfg_pool,
+			(rspamd_mempool_destruct_t)rspamd_ptr_array_free_hard,
+			fuzzy_module_ctx->fuzzy_rules);
+
 	*ctx = (struct module_ctx *)fuzzy_module_ctx;
 
 	rspamd_rcl_add_doc_by_path (cfg,
@@ -978,6 +998,7 @@ fuzzy_check_module_config (struct rspamd_config *cfg)
 	ucl_object_iter_t it;
 	gint res = TRUE, cb_id, nrules = 0;
 	lua_State *L = cfg->lua_state;
+	struct fuzzy_ctx *fuzzy_module_ctx = fuzzy_get_context (cfg);
 
 	if (!rspamd_config_is_module_enabled (cfg, "fuzzy_check")) {
 		return TRUE;
@@ -1150,19 +1171,6 @@ fuzzy_check_module_config (struct rspamd_config *cfg)
 gint
 fuzzy_check_module_reconfig (struct rspamd_config *cfg)
 {
-	struct module_ctx saved_ctx;
-
-	saved_ctx = fuzzy_module_ctx->ctx;
-	rspamd_mempool_delete (fuzzy_module_ctx->fuzzy_pool);
-	rspamd_keypair_cache_destroy (fuzzy_module_ctx->keypairs_cache);
-	g_ptr_array_free (fuzzy_module_ctx->fuzzy_rules, TRUE);
-	memset (fuzzy_module_ctx, 0, sizeof (*fuzzy_module_ctx));
-	fuzzy_module_ctx->ctx = saved_ctx;
-	fuzzy_module_ctx->fuzzy_pool = rspamd_mempool_new (rspamd_mempool_suggest_size (), NULL);
-	fuzzy_module_ctx->cfg = cfg;
-	fuzzy_module_ctx->fuzzy_rules = g_ptr_array_new ();
-	fuzzy_module_ctx->keypairs_cache = rspamd_keypair_cache_new (32);
-
 	return fuzzy_check_module_config (cfg);
 }
 
@@ -1188,7 +1196,7 @@ fuzzy_io_fin (void *ud)
 static GArray *
 fuzzy_preprocess_words (struct rspamd_mime_text_part *part, rspamd_mempool_t *pool)
 {
-	return part->normalized_words;
+	return part->utf_words;
 }
 
 static void
@@ -1213,10 +1221,11 @@ fuzzy_encrypt_cmd (struct fuzzy_rule *rule,
 	memcpy (hdr->pubkey, pk, MIN (pklen, sizeof (hdr->pubkey)));
 	pk = rspamd_pubkey_get_pk (rule->peer_key, &pklen);
 	memcpy (hdr->key_id, pk, MIN (sizeof (hdr->key_id), pklen));
-	rspamd_keypair_cache_process (fuzzy_module_ctx->keypairs_cache,
+	rspamd_keypair_cache_process (rule->ctx->keypairs_cache,
 			rule->local_key, rule->peer_key);
 	rspamd_cryptobox_encrypt_nm_inplace (data, datalen,
-			hdr->nonce, rspamd_pubkey_get_nm (rule->peer_key), hdr->mac,
+			hdr->nonce, rspamd_pubkey_get_nm (rule->peer_key, rule->local_key),
+			hdr->mac,
 			rspamd_pubkey_alg (rule->peer_key));
 }
 
@@ -1409,8 +1418,8 @@ fuzzy_cmd_from_text_part (struct rspamd_task *task,
 			rspamd_cryptobox_hash_init (&st, rule->hash_key->str,
 					rule->hash_key->len);
 
-			rspamd_cryptobox_hash_update (&st, part->stripped_content->data,
-					part->stripped_content->len);
+			rspamd_cryptobox_hash_update (&st, part->utf_stripped_content->data,
+					part->utf_stripped_content->len);
 
 			if (task->subject) {
 				/* We also include subject */
@@ -1765,13 +1774,13 @@ fuzzy_process_reply (guchar **pos, gint *r, GPtrArray *req,
 		*r -= required_size;
 
 		/* Try to decrypt reply */
-		rspamd_keypair_cache_process (fuzzy_module_ctx->keypairs_cache,
+		rspamd_keypair_cache_process (rule->ctx->keypairs_cache,
 				rule->local_key, rule->peer_key);
 
 		if (!rspamd_cryptobox_decrypt_nm_inplace ((guchar *)&encrep.rep,
 				sizeof (encrep.rep),
 				encrep.hdr.nonce,
-				rspamd_pubkey_get_nm (rule->peer_key),
+				rspamd_pubkey_get_nm (rule->peer_key, rule->local_key),
 				encrep.hdr.mac,
 				rspamd_pubkey_alg (rule->peer_key))) {
 			msg_info ("cannot decrypt reply");
@@ -2169,7 +2178,7 @@ fuzzy_check_io_callback (gint fd, short what, void *arg)
 			session->state == 1 ? "read" : "write",
 			errno,
 			strerror (errno));
-		rspamd_upstream_fail (session->server);
+		rspamd_upstream_fail (session->server, FALSE);
 		rspamd_session_remove_event (session->task->s, fuzzy_io_fin, session);
 	}
 	else {
@@ -2203,12 +2212,12 @@ fuzzy_check_timer_callback (gint fd, short what, void *arg)
 		}
 	}
 
-	if (session->retransmits >= fuzzy_module_ctx->retransmits) {
+	if (session->retransmits >= session->rule->ctx->retransmits) {
 		msg_err_task ("got IO timeout with server %s(%s), after %d retransmits",
 				rspamd_upstream_name (session->server),
 				rspamd_inet_address_to_string_pretty (session->addr),
 				session->retransmits);
-		rspamd_upstream_fail (session->server);
+		rspamd_upstream_fail (session->server, FALSE);
 		rspamd_session_remove_event (session->task->s, fuzzy_io_fin, session);
 	}
 	else {
@@ -2413,7 +2422,7 @@ fuzzy_controller_io_callback (gint fd, short what, void *arg)
 				rspamd_upstream_name (session->server),
 				rspamd_inet_address_to_string_pretty (session->addr),
 				errno, strerror (errno));
-		rspamd_upstream_fail (session->server);
+		rspamd_upstream_fail (session->server, FALSE);
 	}
 
 	/*
@@ -2509,8 +2518,8 @@ fuzzy_controller_timer_callback (gint fd, short what, void *arg)
 
 	task = session->task;
 
-	if (session->retransmits >= fuzzy_module_ctx->retransmits) {
-		rspamd_upstream_fail (session->server);
+	if (session->retransmits >= session->rule->ctx->retransmits) {
+		rspamd_upstream_fail (session->server, FALSE);
 		msg_err_task_check ("got IO timeout with server %s(%s), "
 				"after %d retransmits",
 				rspamd_upstream_name (session->server),
@@ -2582,7 +2591,7 @@ fuzzy_generate_commands (struct rspamd_task *task, struct fuzzy_rule *rule,
 		min_bytes = rule->min_bytes;
 	}
 	else {
-		min_bytes = fuzzy_module_ctx->min_bytes;
+		min_bytes = rule->ctx->min_bytes;
 	}
 
 	if (c == FUZZY_STAT) {
@@ -2606,7 +2615,7 @@ fuzzy_generate_commands (struct rspamd_task *task, struct fuzzy_rule *rule,
 			}
 
 			/* Check length of part */
-			fac = fuzzy_module_ctx->text_multiplier * part->content->len;
+			fac = rule->ctx->text_multiplier * part->utf_content->len;
 			if ((double)min_bytes > fac) {
 				if (!rule->short_text_direct_hash) {
 					msg_info_task (
@@ -2615,8 +2624,8 @@ fuzzy_generate_commands (struct rspamd_task *task, struct fuzzy_rule *rule,
 									"skip fuzzy check",
 							task->message_id, min_bytes,
 							fac,
-							part->content->len,
-							fuzzy_module_ctx->text_multiplier);
+							part->utf_content->len,
+							rule->ctx->text_multiplier);
 					continue;
 				}
 				else {
@@ -2626,28 +2635,28 @@ fuzzy_generate_commands (struct rspamd_task *task, struct fuzzy_rule *rule,
 									"use direct hash",
 							task->message_id, min_bytes,
 							fac,
-							part->content->len,
-							fuzzy_module_ctx->text_multiplier);
+							part->utf_content->len,
+							rule->ctx->text_multiplier);
 					short_text = TRUE;
 				}
 			}
 
-			if (part->normalized_words == NULL ||
-					part->normalized_words->len == 0) {
+			if (part->utf_words == NULL ||
+					part->utf_words->len == 0) {
 				msg_info_task ("<%s>, part hash empty, skip fuzzy check",
 						task->message_id);
 				continue;
 			}
 
-			if (fuzzy_module_ctx->min_hash_len != 0 &&
-					part->normalized_words->len <
-							fuzzy_module_ctx->min_hash_len) {
+			if (rule->ctx->min_hash_len != 0 &&
+					part->utf_words->len <
+							rule->ctx->min_hash_len) {
 				if (!rule->short_text_direct_hash) {
 					msg_info_task (
 							"<%s>, part hash is shorter than %d symbols, "
 									"skip fuzzy check",
 							task->message_id,
-							fuzzy_module_ctx->min_hash_len);
+							rule->ctx->min_hash_len);
 					continue;
 				}
 				else {
@@ -2655,7 +2664,7 @@ fuzzy_generate_commands (struct rspamd_task *task, struct fuzzy_rule *rule,
 							"<%s>, part hash is shorter than %d symbols, "
 									"use direct hash",
 							task->message_id,
-							fuzzy_module_ctx->min_hash_len);
+							rule->ctx->min_hash_len);
 					short_text = TRUE;
 				}
 			}
@@ -2703,10 +2712,10 @@ fuzzy_generate_commands (struct rspamd_task *task, struct fuzzy_rule *rule,
 					 * - min bytes
 					 */
 
-					if ((fuzzy_module_ctx->min_height == 0 ||
-							image->height >= fuzzy_module_ctx->min_height) &&
-						(fuzzy_module_ctx->min_width == 0 ||
-							image->width >= fuzzy_module_ctx->min_width) &&
+					if ((rule->ctx->min_height == 0 ||
+							image->height >= rule->ctx->min_height) &&
+						(rule->ctx->min_width == 0 ||
+							image->width >= rule->ctx->min_width) &&
 						(min_bytes == 0 ||
 								mime_part->parsed_data.len >= min_bytes)) {
 						io = fuzzy_cmd_from_data_part (rule, c, flag, value,
@@ -2821,49 +2830,48 @@ register_fuzzy_client_call (struct rspamd_task *task,
 	rspamd_inet_addr_t *addr;
 	gint sock;
 
-	/* Get upstream */
-	selected = rspamd_upstream_get (rule->servers, RSPAMD_UPSTREAM_ROUND_ROBIN,
-			NULL, 0);
-	if (selected) {
-		addr = rspamd_upstream_addr (selected);
-		if ((sock = rspamd_inet_address_connect (addr, SOCK_DGRAM, TRUE)) == -1) {
-			msg_warn_task ("cannot connect to %s(%s), %d, %s",
-				rspamd_upstream_name (selected),
-					rspamd_inet_address_to_string_pretty (addr),
-				errno,
-				strerror (errno));
-			rspamd_upstream_fail (selected);
-			g_ptr_array_free (commands, TRUE);
-		}
-		else {
-			/* Create session for a socket */
-			session =
-				rspamd_mempool_alloc0 (task->task_pool,
-					sizeof (struct fuzzy_client_session));
-			msec_to_tv (fuzzy_module_ctx->io_timeout, &session->tv);
-			session->state = 0;
-			session->commands = commands;
-			session->task = task;
-			session->fd = sock;
-			session->server = selected;
-			session->rule = rule;
-			session->addr = addr;
-			session->results = g_ptr_array_sized_new (32);
+	if (!rspamd_session_blocked (task->s)) {
+		/* Get upstream */
+		selected = rspamd_upstream_get (rule->servers, RSPAMD_UPSTREAM_ROUND_ROBIN,
+				NULL, 0);
+		if (selected) {
+			addr = rspamd_upstream_addr (selected);
+			if ((sock = rspamd_inet_address_connect (addr, SOCK_DGRAM, TRUE)) == -1) {
+				msg_warn_task ("cannot connect to %s(%s), %d, %s",
+						rspamd_upstream_name (selected),
+						rspamd_inet_address_to_string_pretty (addr),
+						errno,
+						strerror (errno));
+				rspamd_upstream_fail (selected, FALSE);
+				g_ptr_array_free (commands, TRUE);
+			} else {
+				/* Create session for a socket */
+				session =
+						rspamd_mempool_alloc0 (task->task_pool,
+								sizeof (struct fuzzy_client_session));
+				msec_to_tv (rule->ctx->io_timeout, &session->tv);
+				session->state = 0;
+				session->commands = commands;
+				session->task = task;
+				session->fd = sock;
+				session->server = selected;
+				session->rule = rule;
+				session->addr = addr;
+				session->results = g_ptr_array_sized_new (32);
 
-			event_set (&session->ev, sock, EV_WRITE, fuzzy_check_io_callback,
-					session);
-			event_base_set (session->task->ev_base, &session->ev);
-			event_add (&session->ev, NULL);
+				event_set (&session->ev, sock, EV_WRITE, fuzzy_check_io_callback,
+						session);
+				event_base_set (session->task->ev_base, &session->ev);
+				event_add (&session->ev, NULL);
 
-			evtimer_set (&session->timev, fuzzy_check_timer_callback,
-					session);
-			event_base_set (session->task->ev_base, &session->timev);
-			event_add (&session->timev, &session->tv);
+				evtimer_set (&session->timev, fuzzy_check_timer_callback,
+						session);
+				event_base_set (session->task->ev_base, &session->timev);
+				event_add (&session->timev, &session->tv);
 
-			rspamd_session_add_event (task->s,
-				fuzzy_io_fin,
-				session,
-				g_quark_from_static_string ("fuzzy check"));
+				rspamd_session_add_event (task->s, NULL, fuzzy_io_fin, session,
+						g_quark_from_static_string ("fuzzy check"));
+			}
 		}
 	}
 }
@@ -2875,6 +2883,7 @@ fuzzy_symbol_callback (struct rspamd_task *task, void *unused)
 	struct fuzzy_rule *rule;
 	guint i;
 	GPtrArray *commands;
+	struct fuzzy_ctx *fuzzy_module_ctx = fuzzy_get_context (task->cfg);
 
 	if (!fuzzy_module_ctx->enabled) {
 		return;
@@ -2906,6 +2915,7 @@ fuzzy_stat_command (struct rspamd_task *task)
 	struct fuzzy_rule *rule;
 	guint i;
 	GPtrArray *commands;
+	struct fuzzy_ctx *fuzzy_module_ctx = fuzzy_get_context (task->cfg);
 
 	if (!fuzzy_module_ctx->enabled) {
 		return;
@@ -2933,6 +2943,7 @@ register_fuzzy_controller_call (struct rspamd_http_connection_entry *entry,
 	struct rspamd_controller_session *session = entry->ud;
 	gint sock;
 	gint ret = -1;
+	struct fuzzy_ctx *fuzzy_module_ctx = fuzzy_get_context (task->cfg);
 
 	/* Get upstream */
 
@@ -2943,7 +2954,7 @@ register_fuzzy_controller_call (struct rspamd_http_connection_entry *entry,
 
 		if ((sock = rspamd_inet_address_connect (addr,
 				SOCK_DGRAM, TRUE)) == -1) {
-			rspamd_upstream_fail (selected);
+			rspamd_upstream_fail (selected, TRUE);
 		}
 		else {
 			s =
@@ -2995,6 +3006,7 @@ fuzzy_process_handler (struct rspamd_http_connection_entry *conn_ent,
 	GString *tb;
 	lua_State *L;
 	gint r, *saved, rules = 0, err_idx;
+	struct fuzzy_ctx *fuzzy_module_ctx;
 
 	/* Prepare task */
 	task = rspamd_task_new (session->wrk, session->cfg, NULL, session->lang_det);
@@ -3002,6 +3014,7 @@ fuzzy_process_handler (struct rspamd_http_connection_entry *conn_ent,
 	task->ev_base = conn_ent->rt->ev_base;
 	saved = rspamd_mempool_alloc0 (session->pool, sizeof (gint));
 	err = rspamd_mempool_alloc0 (session->pool, sizeof (GError *));
+	fuzzy_module_ctx = fuzzy_get_context (ctx->cfg);
 
 	if (!is_hash) {
 		/* Allocate message from string */
@@ -3019,6 +3032,8 @@ fuzzy_process_handler (struct rspamd_http_connection_entry *conn_ent,
 					"Message processing error");
 			return;
 		}
+
+		rspamd_message_process (task);
 	}
 
 	PTR_ARRAY_FOREACH (fuzzy_module_ctx->fuzzy_rules, i, rule) {
@@ -3188,6 +3203,7 @@ fuzzy_controller_handler (struct rspamd_http_connection_entry *conn_ent,
 {
 	const rspamd_ftok_t *arg;
 	glong value = 1, flag = 0, send_flags = 0;
+	struct fuzzy_ctx *fuzzy_module_ctx = (struct fuzzy_ctx *)ctx;
 
 	if (!fuzzy_module_ctx->enabled) {
 		msg_err ("fuzzy_check module is not enabled");
@@ -3293,48 +3309,45 @@ fuzzy_check_send_lua_learn (struct fuzzy_rule *rule,
 	gint ret = -1;
 
 	/* Get upstream */
+	if (!rspamd_session_blocked (task->s)) {
+		while ((selected = rspamd_upstream_get (rule->servers,
+				RSPAMD_UPSTREAM_SEQUENTIAL, NULL, 0))) {
+			/* Create UDP socket */
+			addr = rspamd_upstream_addr (selected);
 
-	while ((selected = rspamd_upstream_get (rule->servers,
-			RSPAMD_UPSTREAM_SEQUENTIAL, NULL, 0))) {
-		/* Create UDP socket */
-		addr = rspamd_upstream_addr (selected);
+			if ((sock = rspamd_inet_address_connect (addr,
+					SOCK_DGRAM, TRUE)) == -1) {
+				rspamd_upstream_fail (selected, TRUE);
+			} else {
+				s =
+						rspamd_mempool_alloc0 (task->task_pool,
+								sizeof (struct fuzzy_learn_session));
 
-		if ((sock = rspamd_inet_address_connect (addr,
-				SOCK_DGRAM, TRUE)) == -1) {
-			rspamd_upstream_fail (selected);
-		}
-		else {
-			s =
-				rspamd_mempool_alloc0 (task->task_pool,
-					sizeof (struct fuzzy_learn_session));
+				msec_to_tv (rule->ctx->io_timeout, &s->tv);
+				s->task = task;
+				s->addr = addr;
+				s->commands = commands;
+				s->http_entry = NULL;
+				s->server = selected;
+				s->saved = saved;
+				s->fd = sock;
+				s->err = err;
+				s->rule = rule;
+				s->session = task->s;
 
-			msec_to_tv (fuzzy_module_ctx->io_timeout, &s->tv);
-			s->task = task;
-			s->addr = addr;
-			s->commands = commands;
-			s->http_entry = NULL;
-			s->server = selected;
-			s->saved = saved;
-			s->fd = sock;
-			s->err = err;
-			s->rule = rule;
-			s->session = task->s;
+				event_set (&s->ev, sock, EV_WRITE, fuzzy_controller_io_callback, s);
+				event_base_set (task->ev_base, &s->ev);
+				event_add (&s->ev, NULL);
 
-			event_set (&s->ev, sock, EV_WRITE, fuzzy_controller_io_callback, s);
-			event_base_set (task->ev_base, &s->ev);
-			event_add (&s->ev, NULL);
+				evtimer_set (&s->timev, fuzzy_controller_timer_callback, s);
+				event_base_set (s->task->ev_base, &s->timev);
+				event_add (&s->timev, &s->tv);
 
-			evtimer_set (&s->timev, fuzzy_controller_timer_callback, s);
-			event_base_set (s->task->ev_base, &s->timev);
-			event_add (&s->timev, &s->tv);
+				rspamd_session_add_event (task->s, NULL, fuzzy_lua_fin, s, g_quark_from_static_string ("fuzzy check"));
 
-			rspamd_session_add_event (task->s,
-					fuzzy_lua_fin,
-					s,
-					g_quark_from_static_string ("fuzzy check"));
-
-			(*saved)++;
-			ret = 1;
+				(*saved)++;
+				ret = 1;
+			}
 		}
 	}
 
@@ -3351,6 +3364,7 @@ fuzzy_check_lua_process_learn (struct rspamd_task *task,
 	GError **err;
 	GPtrArray *commands;
 	gint *saved, rules = 0;
+	struct fuzzy_ctx *fuzzy_module_ctx = fuzzy_get_context (task->cfg);
 
 	saved = rspamd_mempool_alloc0 (task->task_pool, sizeof (gint));
 	err = rspamd_mempool_alloc0 (task->task_pool, sizeof (GError *));
@@ -3417,6 +3431,7 @@ fuzzy_lua_learn_handler (lua_State *L)
 	struct rspamd_task *task = lua_check_task (L, 1);
 	guint flag = 0, weight = 1.0, send_flags = 0;
 	const gchar *symbol;
+	struct fuzzy_ctx *fuzzy_module_ctx = fuzzy_get_context (task->cfg);
 
 	if (task) {
 		if (lua_type (L, 2) == LUA_TNUMBER) {
@@ -3494,6 +3509,7 @@ fuzzy_lua_unlearn_handler (lua_State *L)
 	struct rspamd_task *task = lua_check_task (L, 1);
 	guint flag = 0, weight = 1.0, send_flags = 0;
 	const gchar *symbol;
+	struct fuzzy_ctx *fuzzy_module_ctx = fuzzy_get_context (task->cfg);
 
 	if (task) {
 		if (lua_type (L, 2) == LUA_TNUMBER) {
