@@ -16,12 +16,12 @@ limitations under the License.
 
 local rspamd_logger = require "rspamd_logger"
 local lua_util = require "lua_util"
-local dkim_sign_tools = require "dkim_sign_tools"
+local dkim_sign_tools = require "lua_dkim_tools"
 local rspamd_util = require "rspamd_util"
 local rspamd_rsa_privkey = require "rspamd_rsa_privkey"
 local rspamd_rsa = require "rspamd_rsa"
 local fun = require "fun"
-local auth_results = require "auth_results"
+local auth_results = require "lua_auth_results"
 local hash = require "rspamd_cryptobox_hash"
 
 if confighelp then
@@ -29,11 +29,22 @@ if confighelp then
 end
 
 local N = 'arc'
+
+if not rspamd_plugins.dkim then
+  rspamd_logger.errx(rspamd_config, "cannot enable arc plugin: dkim is disabled")
+  return
+end
+
 local dkim_verify = rspamd_plugins.dkim.verify
 local dkim_sign = rspamd_plugins.dkim.sign
 local dkim_canonicalize = rspamd_plugins.dkim.canon_header_relaxed
 local redis_params
 
+if not dkim_verify or not dkim_sign or not dkim_canonicalize then
+  rspamd_logger.errx(rspamd_config, "cannot enable arc plugin: dkim is disabled")
+  return
+end
+
 local arc_symbols = {
   allow = 'ARC_ALLOW',
   invalid = 'ARC_INVALID',
@@ -76,6 +87,7 @@ local settings = {
   use_esld = true,
   use_redis = false,
   key_prefix = 'arc_keys', -- default hash name
+  reuse_auth_results = false, -- Reuse the existing authentication results
 }
 
 local function parse_arc_header(hdr, target)
@@ -98,6 +110,8 @@ local function parse_arc_header(hdr, target)
     end, fun.map(function(elt)
       return lua_util.rspamd_str_split(elt, '=')
     end, elts))
+    target[i].header = hdr[i].decoded
+    target[i].raw_header = hdr[i].value
   end
 end
 
@@ -134,11 +148,6 @@ local function arc_validate_seals(task, seals, sigs, seal_headers, sig_headers)
         return false
       end
     end
-
-    sigs[i].header = sig_headers[i].decoded
-    seals[i].header = seal_headers[i].decoded
-    sigs[i].raw_header = sig_headers[i].value
-    seals[i].raw_header = seal_headers[i].value
   end
 
   return true
@@ -189,7 +198,7 @@ local function arc_callback(task)
     return (e1.i or 0) < (e2.i or 0)
   end)
 
-  rspamd_logger.debugm(N, task, 'got %s arc sections', #cbdata.seals)
+  lua_util.debugm(N, task, 'got %s arc sections', #cbdata.seals)
 
   -- Now check sanity of what we have
   if not arc_validate_seals(task, cbdata.seals, cbdata.sigs,
@@ -202,7 +211,7 @@ local function arc_callback(task)
 
   local function arc_seal_cb(_, res, err, domain)
     cbdata.checked = cbdata.checked + 1
-    rspamd_logger.debugm(N, task, 'checked arc seal: %s(%s), %s processed',
+    lua_util.debugm(N, task, 'checked arc seal: %s(%s), %s processed',
         res, err, cbdata.checked)
 
     if not res then
@@ -225,9 +234,7 @@ local function arc_callback(task)
   end
 
   local function arc_signature_cb(_, res, err, domain)
-    cbdata.checked = cbdata.checked + 1
-
-    rspamd_logger.debugm(N, task, 'checked arc signature %s: %s(%s), %s processed',
+    lua_util.debugm(N, task, 'checked arc signature %s: %s(%s), %s processed',
       domain, res, err, cbdata.checked)
 
     if not res then
@@ -236,47 +243,83 @@ local function arc_callback(task)
         table.insert(cbdata.errors, string.format('sig:%s:%s', domain, err))
       end
     end
-
-    if cbdata.checked == #arc_sig_headers then
-      if cbdata.res == 'success' then
-        -- Verify seals
-        cbdata.checked = 0
-        fun.each(
-          function(sig)
-            local ret, lerr = dkim_verify(task, sig.header, arc_seal_cb, 'arc-seal')
-            if not ret then
-              cbdata.res = 'fail'
-              table.insert(cbdata.errors, string.format('sig:%s:%s', sig.d or '', lerr))
-              cbdata.checked = cbdata.checked + 1
-              rspamd_logger.debugm(N, task, 'checked arc seal %s: %s(%s), %s processed',
-                sig.d, ret, lerr, cbdata.checked)
-            end
-          end, cbdata.seals)
-      else
-        task:insert_result(arc_symbols['reject'], 1.0,
-          rspamd_logger.slog('signature check failed: %s, %s', cbdata.res,
-            cbdata.errors))
-      end
+    if cbdata.res == 'success' then
+      -- Verify seals
+      cbdata.checked = 0
+      fun.each(
+        function(sig)
+          local ret, lerr = dkim_verify(task, sig.header, arc_seal_cb, 'arc-seal')
+          if not ret then
+            cbdata.res = 'fail'
+            table.insert(cbdata.errors, string.format('sig:%s:%s', sig.d or '', lerr))
+            cbdata.checked = cbdata.checked + 1
+            lua_util.debugm(N, task, 'checked arc seal %s: %s(%s), %s processed',
+              sig.d, ret, lerr, cbdata.checked)
+          end
+        end, cbdata.seals)
+    else
+      task:insert_result(arc_symbols['reject'], 1.0,
+        rspamd_logger.slog('signature check failed: %s, %s', cbdata.res,
+          cbdata.errors))
     end
   end
 
-  -- Now we can verify all signatures
+  --[[
+  1.  Collect all ARC Sets currently attached to the message.  If there
+       are none, the Chain Validation Status is "none" and the algorithm
+       stops here.  The maximum number of ARC Sets that can be attached
+       to a message is 50.  If more than the maximum number exist the
+       Chain Validation Status is "fail" and the algorithm stops here.
+       In the following algorithm, the maximum ARC instance value is
+       referred to as "N".
+
+   2.  If the Chain Validation Status of the highest instance value ARC
+       Set is "fail", then the Chain Validation status is "fail" and the
+       algorithm stops here.
+
+   3.  Validate the structure of the Authenticated Received Chain.  A
+       valid ARC has the following conditions:
+
+       1.  Each ARC Set MUST contain exactly one each of the three ARC
+           header fields (AAR, AMS, and AS).
+
+       2.  The instance values of the ARC Sets MUST form a continuous
+           sequence from 1..N with no gaps or repetition.
+
+       3.  The "cv" value for all ARC-Seal header fields must be non-
+           failing.  For instance values > 1, the value must be "pass".
+           For instance value = 1, the value must be "none".
+
+       *  If any of these conditions are not met, the Chain Validation
+          Status is "fail" and the algorithm stops here.
+
+   4.  Validate the AMS with the greatest instance value (most recent).
+       If validation fails, then the Chain Validation Status is "fail"
+       and the algorithm stops here.
+
+   5 - 7. Optional, not implemented
+   8.  Validate each AS beginning with the greatest instance value and
+       proceeding in decreasing order to the AS with the instance value
+       of 1.  If any AS fails to validate, the Chain Validation Status
+       is "fail" and the algorithm stops here.
+   9.  If the algorithm reaches this step, then the Chain Validation
+       Status is "pass", and the algorithm is complete.
+  ]]--
+
   local processed = 0
-  fun.each(
-    function(sig)
-      local ret,err = dkim_verify(task, sig.header, arc_signature_cb, 'arc-sign')
+  local sig = cbdata.sigs[#cbdata.sigs] -- last AMS
+  local ret,err = dkim_verify(task, sig.header, arc_signature_cb, 'arc-sign')
 
-      if not ret then
-        cbdata.res = 'fail'
-        table.insert(cbdata.errors, string.format('sig:%s:%s', sig.d or '', err))
-      else
-        processed = processed + 1
-        rspamd_logger.debugm(N, task, 'processed arc signature %s: %s(%s), %s processed',
-          sig.d, ret, err, cbdata.checked)
-      end
-    end, cbdata.sigs)
+  if not ret then
+    cbdata.res = 'fail'
+    table.insert(cbdata.errors, string.format('sig:%s:%s', sig.d or '', err))
+  else
+    processed = processed + 1
+    lua_util.debugm(N, task, 'processed arc signature %s[%s]: %s(%s), %s processed',
+      sig.d, sig.i, ret, err, cbdata.checked)
+  end
 
-  if processed ~= #arc_sig_headers then
+  if processed == 0 then
     task:insert_result(arc_symbols['reject'], 1.0,
       rspamd_logger.slog('cannot verify %s of %s signatures: %s',
         #arc_sig_headers - processed, #arc_sig_headers, cbdata.errors))
@@ -309,62 +352,76 @@ rspamd_config:register_symbol({
   parent = id,
   type = 'virtual',
   score = -1.0,
-  group = 'arc',
+  group = 'policies',
 })
 rspamd_config:register_symbol({
   name = arc_symbols['reject'],
   parent = id,
   type = 'virtual',
   score = 2.0,
-  group = 'arc',
+  group = 'policies',
 })
 rspamd_config:register_symbol({
   name = arc_symbols['invalid'],
   parent = id,
   type = 'virtual',
   score = 1.0,
-  group = 'arc',
+  group = 'policies',
 })
 rspamd_config:register_symbol({
   name = arc_symbols['dnsfail'],
   parent = id,
   type = 'virtual',
   score = 0.0,
-  group = 'arc',
+  group = 'policies',
 })
 rspamd_config:register_symbol({
   name = arc_symbols['na'],
   parent = id,
   type = 'virtual',
   score = 0.0,
-  group = 'arc',
+  group = 'policies',
 })
 
-rspamd_config:register_dependency(id, symbols['spf_allow_symbol'])
-rspamd_config:register_dependency(id, symbols['dkim_allow_symbol'])
+rspamd_config:register_dependency('ARC_CALLBACK', symbols['spf_allow_symbol'])
+rspamd_config:register_dependency('ARC_CALLBACK', symbols['dkim_allow_symbol'])
 
 local function arc_sign_seal(task, params, header)
-  local fold_type = "crlf"
   local arc_sigs = task:cache_get('arc-sigs')
   local arc_seals = task:cache_get('arc-seals')
   local arc_auth_results = task:get_header_full('ARC-Authentication-Results') or {}
-  local cur_auth_results = auth_results.gen_auth_results(task) or ''
+  local cur_auth_results
   local privkey
 
-  if task:has_flag("milter") then
-    fold_type = "lf"
-  end
-
   if params.rawkey then
-    privkey = rspamd_rsa_privkey.load_pem(params.rawkey)
+    -- Distinguish between pem and base64
+    if string.match(params.rawkey, '^-----BEGIN') then
+      privkey = rspamd_rsa_privkey.load_pem(params.rawkey)
+    else
+      privkey = rspamd_rsa_privkey.load_base64(params.rawkey)
+    end
   elseif params.key then
     privkey = rspamd_rsa_privkey.load_file(params.key)
   end
 
   if not privkey then
     rspamd_logger.errx(task, 'cannot load private key for signing')
+    return
   end
 
+  if settings.reuse_auth_results then
+    local ar_header = task:get_header('Authentication-Results')
+
+    if ar_header then
+      rspamd_logger.debugm(N, task, 'reuse authentication results header for ARC')
+      cur_auth_results = ar_header
+    else
+      rspamd_logger.debugm(N, task, 'cannot reuse authentication results, header is missing')
+      cur_auth_results = auth_results.gen_auth_results(task) or ''
+    end
+  else
+    cur_auth_results = auth_results.gen_auth_results(task) or ''
+  end
 
   local sha_ctx = hash.create_specific('sha256')
 
@@ -377,44 +434,47 @@ local function arc_sign_seal(task, params, header)
         local s = dkim_canonicalize('ARC-Authentication-Results',
           arc_auth_results[i].value)
         sha_ctx:update(s)
-        rspamd_logger.debugm(N, task, 'update signature with header: %s', s)
+        lua_util.debugm(N, task, 'update signature with header: %s', s)
       end
       if arc_sigs[i] then
         local s = dkim_canonicalize('ARC-Message-Signature',
           arc_sigs[i].raw_header)
         sha_ctx:update(s)
-        rspamd_logger.debugm(N, task, 'update signature with header: %s', s)
+        lua_util.debugm(N, task, 'update signature with header: %s', s)
       end
       if arc_seals[i] then
         local s = dkim_canonicalize('ARC-Seal', arc_seals[i].raw_header)
         sha_ctx:update(s)
-        rspamd_logger.debugm(N, task, 'update signature with header: %s', s)
+        lua_util.debugm(N, task, 'update signature with header: %s', s)
       end
     end
   end
 
-  header = rspamd_util.fold_header(
+  header = lua_util.fold_header(task,
     'ARC-Message-Signature',
-    header, fold_type)
-
-  cur_auth_results = rspamd_util.fold_header(
-    'ARC-Authentication-Results',
-    cur_auth_results, fold_type)
+    header)
 
   cur_auth_results = string.format('i=%d; %s', cur_idx, cur_auth_results)
+  cur_auth_results = lua_util.fold_header(task,
+      'ARC-Authentication-Results',
+      cur_auth_results, ';')
+
   local s = dkim_canonicalize('ARC-Authentication-Results',
     cur_auth_results)
   sha_ctx:update(s)
-  rspamd_logger.debugm(N, task, 'update signature with header: %s', s)
+  lua_util.debugm(N, task, 'update signature with header: %s', s)
   s = dkim_canonicalize('ARC-Message-Signature', header)
   sha_ctx:update(s)
-  rspamd_logger.debugm(N, task, 'update signature with header: %s', s)
+  lua_util.debugm(N, task, 'update signature with header: %s', s)
 
   local cur_arc_seal = string.format('i=%d; s=%s; d=%s; t=%d; a=rsa-sha256; cv=%s; b=',
-      cur_idx, params.selector, params.domain, math.floor(rspamd_util.get_time()), params.arc_cv)
+      cur_idx,
+      params.selector,
+      params.domain,
+      math.floor(rspamd_util.get_time()), params.arc_cv)
   s = string.format('%s:%s', 'arc-seal', cur_arc_seal)
   sha_ctx:update(s)
-  rspamd_logger.debugm(N, task, 'initial update signature with header: %s', s)
+  lua_util.debugm(N, task, 'initial update signature with header: %s', s)
 
   local sig = rspamd_rsa.sign_memory(privkey, sha_ctx:bin())
   cur_arc_seal = string.format('%s%s', cur_arc_seal,
@@ -422,11 +482,10 @@ local function arc_sign_seal(task, params, header)
 
   task:set_milter_reply({
     add_headers = {
-      ['ARC-Authentication-Results'] = cur_auth_results,
-      ['ARC-Message-Signature'] = header,
-      ['ARC-Seal'] = rspamd_util.fold_header(
-        'ARC-Seal',
-        cur_arc_seal, fold_type),
+      ['ARC-Authentication-Results'] = {order = 1, value = cur_auth_results},
+      ['ARC-Message-Signature'] = {order = 1, value = header},
+      ['ARC-Seal'] = {order = 1, value = lua_util.fold_header(task,
+        'ARC-Seal', cur_arc_seal) }
     }
   })
   task:insert_result(settings.sign_symbol, 1.0, string.format('i=%d', cur_idx))
@@ -513,8 +572,24 @@ local function arc_signing_cb(task)
       try_redis_key(p.selector)
     end
   else
-    if (p.key and p.selector) then
-      p.key = lua_util.template(p.key, {domain = p.domain, selector = p.selector})
+    if ((p.key or p.rawkey) and p.selector) then
+      if p.key then
+        p.key = lua_util.template(p.key, {
+          domain = p.domain,
+          selector = p.selector
+        })
+
+        local exists,err = rspamd_util.file_exists(p.key)
+        if not exists then
+          if err and err == 'No such file or directory' then
+            lua_util.debugm(N, task, 'cannot read key from %s: %s', p.key, err)
+          else
+            rspamd_logger.warnx(N, task, 'cannot read key from %s: %s', p.key, err)
+          end
+          return false
+        end
+      end
+
       local dret, hdr = dkim_sign(task, p)
       if dret then
         return arc_sign_seal(task, p, hdr)
@@ -537,8 +612,12 @@ for k,v in pairs(opts) do
     settings[k] = v
   end
 end
-if not (settings.use_redis or settings.path or
-    settings.domain or settings.path_map or settings.selector_map) then
+if not (settings.use_redis or
+    settings.path or
+    settings.domain or
+    settings.path_map or
+    settings.selector_map or
+    settings.use_http_headers) then
   rspamd_logger.infox(rspamd_config, 'mandatory parameters missing, disable arc signing')
   return
 end
@@ -547,15 +626,17 @@ if settings.use_redis then
   redis_params = rspamd_parse_redis_server('arc')
 
   if not redis_params then
-    rspamd_logger.errx(rspamd_config, 'no servers are specified, but module is configured to load keys from redis, disable dkim signing')
+    rspamd_logger.errx(rspamd_config, 'no servers are specified, but module is configured to load keys from redis, disable arc signing')
     return
   end
 end
 
-id = rspamd_config:register_symbol({
+rspamd_config:register_symbol({
   name = settings['sign_symbol'],
-  callback = arc_signing_cb
+  callback = arc_signing_cb,
+  groups = {"policies", "arc"},
+  score = 0.0,
 })
 
 -- Do not sign unless valid
-rspamd_config:register_dependency(id, 'ARC_CALLBACK')
+rspamd_config:register_dependency(settings['sign_symbol'], 'ARC_CALLBACK')
