@@ -24,6 +24,7 @@ local rspamd_tcp = require "rspamd_tcp"
 local rspamd_url = require "rspamd_url"
 local rspamd_util = require "rspamd_util"
 local rspamd_redis = require "lua_redis"
+local lua_util = require "lua_util"
 local check_local = false
 local check_authed = false
 
@@ -37,7 +38,7 @@ local no_reporting_domains
 local statefile = string.format('%s/%s', rspamd_paths['DBDIR'], 'dmarc_reports_last_sent')
 local VAR_NAME = 'dmarc_reports_last_sent'
 local INTERVAL = 86400
-local pool = mempool.create()
+local pool
 
 local report_settings = {
   helo = 'rspamd',
@@ -45,8 +46,9 @@ local report_settings = {
   smtp = '127.0.0.1',
   smtp_port = 25,
   retries = 2,
+  from_name = 'Rspamd',
 }
-local report_template = [[From: "Rspamd" <%s>
+local report_template = [[From: "%s" <%s>
 To: %s
 Subject: Report Domain: %s
 	Submitter: %s
@@ -54,7 +56,7 @@ Subject: Report Domain: %s
 Date: %s
 MIME-Version: 1.0
 Message-ID: <%s>
-Content-Type: multipart/alternative;
+Content-Type: multipart/mixed;
 	boundary="----=_NextPart_000_024E_01CC9B0A.AFE54C00"
 
 This is a multipart message in MIME format.
@@ -66,10 +68,10 @@ Content-Transfer-Encoding: 7bit
 This is an aggregate report from %s.
 
 ------=_NextPart_000_024E_01CC9B0A.AFE54C00
-Content-Type: text/xml
+Content-Type: application/gzip
 Content-Transfer-Encoding: base64
 Content-Disposition: attachment;
-	filename="%s!%s!%s!%s.xml"
+	filename="%s!%s!%s!%s.xml.gz"
 
 ]]
 local report_footer = [[
@@ -134,7 +136,7 @@ local dmarc_actions = {}
 
 local E = {}
 
-local take_report_sha
+local take_report_id
 local take_report_script = [[
 local index_key = KEYS[1]
 local report_key = KEYS[2]
@@ -157,29 +159,6 @@ end
 
 local tz_offset = get_timezone_offset(os.time())
 
-local function load_scripts(cfg, ev_base)
-  local function redis_report_script_cb(err, data)
-    if err then
-      rspamd_logger.errx(cfg, 'DMARC report script loading failed: ' .. err)
-    else
-      take_report_sha = tostring(data)
-      rspamd_logger.infox(cfg, 'Loaded DMARC report script with SHA %s', take_report_sha)
-    end
-  end
-  local ret = rspamd_redis.redis_make_request_taskless(ev_base,
-    rspamd_config,
-    redis_params,
-    nil,
-    true, -- is write
-    redis_report_script_cb, --callback
-    'SCRIPT', -- command
-    {'LOAD', take_report_script}
-  )
-  if not ret then
-    rspamd_logger.errx(cfg, 'Unable to load DMARC report script')
-  end
-end
-
 local function gen_dmarc_grammar()
   local lpeg = require "lpeg"
   lpeg.locale(lpeg)
@@ -221,6 +200,8 @@ local function dmarc_callback(task)
   local function maybe_force_action(disposition)
     local force_action = dmarc_actions[disposition]
     if force_action then
+      -- Don't do anything if pre-result has been already set
+      if task:has_pre_result() then return end
       task:set_pre_result(force_action, 'Action set by DMARC')
     end
   end
@@ -259,9 +240,6 @@ local function dmarc_callback(task)
       rspamd_logger.infox(task, '<%1> dmarc report saved for %2',
         task:get_message_id(), hfromdom)
     else
-      if string.match(err, 'NOSCRIPT') then
-        load_scripts(rspamd_config, task:get_ev_base())
-      end
       rspamd_logger.errx(task, '<%1> dmarc report is not saved for %2: %3',
         task:get_message_id(), hfromdom, err)
     end
@@ -269,7 +247,6 @@ local function dmarc_callback(task)
 
   local function dmarc_dns_cb(_, to_resolve, results, err)
 
-    task:inc_dns_req()
     local lookup_domain = string.sub(to_resolve, 8)
     if err and (err ~= 'requested record is not found' and err ~= 'no records with this name') then
       task:insert_result(dmarc_symbols['dnsfail'], 1.0, lookup_domain .. ' : ' .. err)
@@ -306,7 +283,7 @@ local function dmarc_callback(task)
 
     for _,r in ipairs(results) do
       if failed_policy then break end
-      (function()
+      local function try()
         local elts = dmarc_grammar:match(r)
         if not elts then
           return
@@ -381,7 +358,8 @@ local function dmarc_callback(task)
             rua = elts['rua']
           end
         end
-      end)()
+      end
+      try()
     end
 
     if not found_policy then
@@ -566,17 +544,8 @@ local function dmarc_callback(task)
       local idx_key = table.concat({redis_keys.index_prefix, period}, redis_keys.join_char)
 
       if report_data then
-        local ret = rspamd_redis.redis_make_request(task,
-          redis_params, -- connect params
-          hfromdom, -- hash key
-          true, -- is write
-          dmarc_report_cb, --callback
-          'EVALSHA', -- command
-          {take_report_sha, 2, idx_key, dmarc_domain_key, hfromdom, report_data} -- arguments
-        )
-        if not ret then
-          rspamd_logger.errx(task, 'Unable to schedule redis request')
-        end
+        rspamd_redis.exec_redis_script(take_report_id, {task = task, is_write = true}, dmarc_report_cb,
+          {idx_key, dmarc_domain_key}, {hfromdom, report_data})
       end
     end
 
@@ -594,7 +563,7 @@ local function dmarc_callback(task)
 end
 
 local opts = rspamd_config:get_all_opt('options')
-if opts and type(opts) ~= 'table' then
+if type(opts) == 'table' then
   if type(opts['check_local']) == 'boolean' then
     check_local = opts['check_local']
   end
@@ -624,9 +593,7 @@ if opts['reporting'] == true then
     rspamd_logger.errx(rspamd_config, 'cannot parse servers parameter')
   elseif not opts['send_reports'] then
     dmarc_reporting = true
-    rspamd_config:add_on_load(function(cfg, ev_base, worker)
-      load_scripts(cfg, ev_base)
-    end)
+    take_report_id = rspamd_redis.add_redis_script(take_report_script, redis_params)
   else
     dmarc_reporting = true
     if type(opts['report_settings']) == 'table' then
@@ -640,10 +607,11 @@ if opts['reporting'] == true then
         return
       end
     end
+    take_report_id = rspamd_redis.add_redis_script(take_report_script, redis_params)
     rspamd_config:add_on_load(function(cfg, ev_base, worker)
-      load_scripts(cfg, ev_base)
-      if not (worker:get_name() == 'controller' and worker:get_index() == 0) then return end
+      if not worker:is_primary_controller() then return end
       local rresolver = rspamd_resolver.init(ev_base, rspamd_config)
+      pool = mempool.create()
       rspamd_config:register_finish_script(function ()
         local stamp = pool:get_variable(VAR_NAME, 'double')
         if not stamp then
@@ -657,6 +625,7 @@ if opts['reporting'] == true then
         end
         assert(f:write(pool:get_variable(VAR_NAME, 'double')))
         assert(f:close())
+        pool:destroy()
       end)
       local get_reporting_domain, reporting_domain, report_start, report_end, report_id, want_period, report_key
       local reporting_addr = {}
@@ -697,7 +666,7 @@ if opts['reporting'] == true then
       local function dmarc_report_xml()
         local entries = {}
         report_id = string.format('%s.%d.%d', reporting_domain, report_start, report_end)
-        rspamd_logger.debugm(N, rspamd_config, 'new report: %s', report_id)
+        lua_util.debugm(N, rspamd_config, 'new report: %s', report_id)
         local actions = {
           push = function(t)
             local data = t[1]
@@ -780,7 +749,11 @@ if opts['reporting'] == true then
         for k in pairs(reporting_addr) do
           table.insert(tmp_addr, k)
         end
-        local encoded = rspamd_util.encode_base64(table.concat({xmlf('header'), xmlf('entries'), xmlf('footer')}), 78)
+        local encoded = rspamd_util.encode_base64(rspamd_util.gzip_compress(
+              table.concat(
+                {xmlf('header'),
+                 xmlf('entries'),
+                 xmlf('footer')})), 78)
         local function mail_cb(err, data, conn)
           local function no_error(merr, mdata, wantcode)
             wantcode = wantcode or '2'
@@ -841,11 +814,23 @@ if opts['reporting'] == true then
                 table.insert(atmp, k)
               end
               local addr_string = table.concat(atmp, ', ')
-              local rhead = string.format(report_template, report_settings.email, addr_string,
-                reporting_domain, report_settings.domain, report_id, rspamd_util.time_to_string(rspamd_util.get_time()),
-                rspamd_util.random_hex(12) .. '@rspamd', report_settings.domain, report_settings.domain, reporting_domain,
-                report_start, report_end)
-              conn:add_write(pre_quit_cb, {rhead, encoded, report_footer, '\r\n.\r\n'})
+              local rhead = string.format(report_template,
+		  report_settings.from_name,
+                  report_settings.email,
+                  addr_string,
+                  reporting_domain,
+                  report_settings.domain,
+                  report_id,
+                  rspamd_util.time_to_string(rspamd_util.get_time()),
+                  rspamd_util.random_hex(12) .. '@rspamd',
+                  report_settings.domain,
+                  report_settings.domain,
+                  reporting_domain,
+                  report_start, report_end)
+              conn:add_write(pre_quit_cb, {rhead,
+                                           encoded,
+                                           report_footer,
+                                           '\r\n.\r\n'})
             end
           end
           local function data_cb(merr, mdata)
@@ -1200,19 +1185,19 @@ if opts['reporting'] == true then
       end
       local time = rspamd_util.get_time()
       if not stamp then
-        rspamd_logger.debugm(N, rspamd_config, 'No state found - sending reports immediately')
+        lua_util.debugm(N, rspamd_config, 'No state found - sending reports immediately')
         schedule_regular_send()
         send_reports(time)
         return
       end
       local delta = stamp - time + INTERVAL
       if delta <= 0 then
-        rspamd_logger.debugm(N, rspamd_config, 'Last send is too old - sending reports immediately')
+        lua_util.debugm(N, rspamd_config, 'Last send is too old - sending reports immediately')
         schedule_regular_send()
         send_reports(time)
         return
       end
-      rspamd_logger.debugm(N, rspamd_config, 'Scheduling next send in %s seconds', delta)
+      lua_util.debugm(N, rspamd_config, 'Scheduling next send in %s seconds', delta)
       schedule_intermediate_send(delta)
     end)
   end
@@ -1225,7 +1210,7 @@ if type(opts['report_settings']) == 'table' then
     report_settings[k] = v
   end
 end
-if dmarc_reporting then
+if opts['send_reports'] then
   for _, e in ipairs({'email', 'domain', 'org_name'}) do
     if not report_settings[e] then
       rspamd_logger.errx(rspamd_config, 'Missing required setting: report_settings.%s', e)
@@ -1296,6 +1281,6 @@ rspamd_config:register_symbol({
   type = 'virtual'
 })
 
-rspamd_config:register_dependency(id, symbols['spf_allow_symbol'])
-rspamd_config:register_dependency(id, symbols['dkim_allow_symbol'])
+rspamd_config:register_dependency('DMARC_CALLBACK', symbols['spf_allow_symbol'])
+rspamd_config:register_dependency('DMARC_CALLBACK', symbols['dkim_allow_symbol'])
 
