@@ -22,7 +22,7 @@ end
 -- A plugin that pushes metadata (or whole messages) to external services
 
 local redis_params
-local lutil = require "lua_util"
+local lua_util = require "lua_util"
 local rspamd_http = require "rspamd_http"
 local rspamd_tcp = require "rspamd_tcp"
 local rspamd_util = require "rspamd_util"
@@ -41,7 +41,7 @@ local settings = {
   mail_to = 'postmaster@localhost',
   helo = 'rspamd',
   email_template = [[From: "Rspamd" <$mail_from>
-To: <$mail_to>
+To: $mail_to
 Subject: Spam alert
 Date: $date
 MIME-Version: 1.0
@@ -66,7 +66,12 @@ Symbols: $symbols]],
 
 local function get_general_metadata(task, flatten, no_content)
   local r = {}
-  r.ip = tostring(task:get_from_ip()) or 'unknown'
+  local ip = task:get_from_ip()
+  if ip and ip:is_valid() then
+    r.ip = tostring(ip)
+  else
+    r.ip = 'unknown'
+  end
   r.user = task:get_user() or 'unknown'
   r.qid = task:get_queue_id() or 'unknown'
   r.action = task:get_metric_action('default')
@@ -141,13 +146,48 @@ local formatters = {
   default = function(task)
     return task:get_content()
   end,
-  email_alert = function(task, rule)
+  email_alert = function(task, rule, extra)
     local meta = get_general_metadata(task, true)
+    local display_emails = {}
     meta.mail_from = rule.mail_from or settings.mail_from
-    meta.mail_to = rule.mail_to or settings.mail_to
+    local mail_targets = rule.mail_to or settings.mail_to
+    if type(mail_targets) ~= 'table' then
+      table.insert(display_emails, string.format('<%s>', mail_targets))
+      mail_targets = {[mail_targets] = true}
+    else
+      for _, e in ipairs(mail_targets) do
+        table.insert(display_emails, string.format('<%s>', e))
+      end
+    end
+    if rule.email_alert_sender then
+      local x = task:get_from('smtp')
+      if x and string.len(x[1].addr) > 0 then
+        mail_targets[x] = true
+        table.insert(display_emails, string.format('<%s>', x[1].addr))
+      end
+    end
+    if rule.email_alert_user then
+      local x = task:get_user()
+      if x then
+        mail_targets[x] = true
+        table.insert(display_emails, string.format('<%s>', x))
+      end
+    end
+    if rule.email_alert_recipients then
+      local x = task:get_recipients('smtp')
+      if x then
+        for _, e in ipairs(x) do
+          if string.len(e.addr) > 0 then
+            mail_targets[e.addr] = true
+            table.insert(display_emails, string.format('<%s>', e.addr))
+          end
+        end
+      end
+    end
+    meta.mail_to = table.concat(display_emails, ', ')
     meta.our_message_id = rspamd_util.random_hex(12) .. '@rspamd'
     meta.date = rspamd_util.time_to_string(rspamd_util.get_time())
-    return lutil.template(rule.email_template or settings.email_template, meta)
+    return lua_util.template(rule.email_template or settings.email_template, meta), { mail_targets = mail_targets}
   end,
   json = function(task)
     return ucl.to_format(get_general_metadata(task), 'json-compact')
@@ -189,7 +229,7 @@ local selectors = {
 local function maybe_defer(task, rule)
   if rule.defer then
     rspamd_logger.warnx(task, 'deferring message')
-    task:set_metric_action('default', 'soft reject')
+    task:set_pre_result('soft reject', 'deferred')
   end
 end
 
@@ -250,7 +290,7 @@ local pushers = {
       headers=hdrs,
     })
   end,
-  send_mail = function(task, formatted, rule)
+  send_mail = function(task, formatted, rule, extra)
     local function mail_cb(err, data, conn)
       local function no_error(merr, mdata, wantcode)
         wantcode = wantcode or '2'
@@ -313,9 +353,15 @@ local pushers = {
           conn:add_read(data_done_cb, '\r\n')
         end
       end
+      local from_done_cb
       local function rcpt_done_cb(merr, mdata)
         if no_error(merr, mdata) then
-          conn:add_write(data_cb, 'DATA\r\n')
+          local k = next(extra.mail_targets)
+          if not k then
+            conn:add_write(data_cb, 'DATA\r\n')
+          else
+            from_done_cb('2', '2')
+          end
         end
       end
       local function rcpt_cb(merr, mdata)
@@ -323,10 +369,24 @@ local pushers = {
           conn:add_read(rcpt_done_cb, '\r\n')
         end
       end
-      local function from_done_cb(merr, mdata)
-        if no_error(merr, mdata) then
-          conn:add_write(rcpt_cb, {'RCPT TO: <', rule.mail_to, '>\r\n'})
+      from_done_cb = function(merr, mdata)
+        local k
+        if extra then
+          k = next(extra.mail_targets)
+        else
+          extra = {mail_targets = {}}
+          if type(rule.mail_to) == 'string' then
+            extra = {mail_targets = {}}
+            k = rule.mail_to
+          elseif type(rule.mail_to) == 'table' then
+            for _, r in ipairs(rule.mail_to) do
+              extra.mail_targets[r] = true
+            end
+            k = next(extra.mail_targets)
+          end
         end
+        extra.mail_targets[k] = nil
+        conn:add_write(rcpt_cb, {'RCPT TO: <', k, '>\r\n'})
       end
       local function from_cb(merr, mdata)
         if no_error(merr, '2') then
@@ -533,7 +593,7 @@ if type(settings.rules) ~= 'table' then
     return
   end
 elseif not next(settings.rules) then
-  rspamd_logger.debugm(N, rspamd_config, 'No rules enabled')
+  lua_util.debugm(N, rspamd_config, 'No rules enabled')
   return
 end
 if not settings.rules or not next(settings.rules) then
@@ -631,28 +691,30 @@ local function gen_exporter(rule)
     local selector = rule.selector or 'default'
     local selected = selectors[selector](task)
     if selected then
-      rspamd_logger.debugm(N, task, 'Message selected for processing')
+      lua_util.debugm(N, task, 'Message selected for processing')
       local formatter = rule.formatter or 'default'
-      local formatted = formatters[formatter](task, rule)
+      local formatted, extra = formatters[formatter](task, rule)
       if formatted then
-        pushers[rule.backend](task, formatted, rule)
+        pushers[rule.backend](task, formatted, rule, extra)
       else
-        rspamd_logger.debugm(N, task, 'Formatter [%s] returned non-truthy value [%s]', formatter, formatted)
+        lua_util.debugm(N, task, 'Formatter [%s] returned non-truthy value [%s]', formatter, formatted)
       end
     else
-      rspamd_logger.debugm(N, task, 'Selector [%s] returned non-truthy value [%s]', selector, selected)
+      lua_util.debugm(N, task, 'Selector [%s] returned non-truthy value [%s]', selector, selected)
     end
   end
 end
 
 if not next(settings.rules) then
   rspamd_logger.errx(rspamd_config, 'No rules enabled')
+  lua_util.disable_module(N, "config")
 end
 for k, r in pairs(settings.rules) do
   rspamd_config:register_symbol({
     name = 'EXPORT_METADATA_' .. k,
-    type = 'postfilter',
+    type = 'postfilter,idempotent',
     callback = gen_exporter(r),
-    priority = 10
+    priority = 10,
+    flags = 'empty',
   })
 end
