@@ -18,6 +18,7 @@
 #include "rspamd.h"
 #include "ottery.h"
 #include "lua/lua_common.h"
+#include "lua/lua_thread_pool.h"
 #include "lua_ucl.h"
 #include "unix-std.h"
 
@@ -31,9 +32,15 @@ static gboolean show_help = FALSE;
 static gboolean show_version = FALSE;
 GHashTable *ucl_vars = NULL;
 struct rspamd_main *rspamd_main = NULL;
+struct rspamd_async_session *rspamadm_session = NULL;
+lua_State *L = NULL;
 
-static void rspamadm_help (gint argc, gchar **argv);
-static const char* rspamadm_help_help (gboolean full_help);
+/* Defined in modules.c */
+extern module_t *modules[];
+extern worker_t *workers[];
+
+static void rspamadm_help (gint argc, gchar **argv, const struct rspamadm_command *);
+static const char* rspamadm_help_help (gboolean full_help, const struct rspamadm_command *);
 
 struct rspamadm_command help_command = {
 	.name = "help",
@@ -56,7 +63,7 @@ static GOptionEntry entries[] = {
 			"Redefine UCL variable", NULL},
 	{"help", 'h', 0, G_OPTION_ARG_NONE, &show_help,
 			"Show help", NULL},
-	{"version", 'v', 0, G_OPTION_ARG_NONE, &show_version,
+	{"version", 'V', 0, G_OPTION_ARG_NONE, &show_version,
 			"Show version", NULL},
 	{NULL, 0, 0, G_OPTION_ARG_NONE, NULL, NULL, NULL}
 };
@@ -70,7 +77,7 @@ rspamadm_error (void)
 static void
 rspamadm_version (void)
 {
-	printf ("Rspamadm %s\n", RVERSION);
+	rspamd_printf ("Rspamadm %s\n", RVERSION);
 }
 
 static void
@@ -79,30 +86,33 @@ rspamadm_usage (GOptionContext *context)
 	gchar *help_str;
 
 	help_str = g_option_context_get_help (context, TRUE, NULL);
-	printf ("%s", help_str);
+	rspamd_printf ("%s", help_str);
 }
 
 static void
-rspamadm_commands (void)
+rspamadm_commands (GPtrArray *all_commands)
 {
-	const struct rspamadm_command **cmd;
+	const struct rspamadm_command *cmd;
+	guint i;
 
-	printf ("Rspamadm %s\n", RVERSION);
-	printf ("Usage: rspamadm [global_options] command [command_options]\n");
-	printf ("\nAvailable commands:\n");
+	rspamd_printf ("Rspamadm %s\n", RVERSION);
+	rspamd_printf ("Usage: rspamadm [global_options] command [command_options]\n");
+	rspamd_printf ("\nAvailable commands:\n");
 
-	cmd = commands;
-
-	while (*cmd) {
-		if (!((*cmd)->flags & RSPAMADM_FLAG_NOHELP)) {
-			printf ("  %-18s %-60s\n", (*cmd)->name, (*cmd)->help (FALSE));
+	PTR_ARRAY_FOREACH (all_commands, i, cmd) {
+		if (!(cmd->flags & RSPAMADM_FLAG_NOHELP)) {
+			if (cmd->flags & RSPAMADM_FLAG_LUA) {
+				(void)cmd->help (FALSE, cmd);
+			}
+			else {
+				printf ("  %-18s %-60s\n", cmd->name, cmd->help (FALSE, cmd));
+			}
 		}
-		cmd ++;
 	}
 }
 
 static const char *
-rspamadm_help_help (gboolean full_help)
+rspamadm_help_help (gboolean full_help, const struct rspamadm_command *cmd)
 {
 	const char *help_str;
 
@@ -118,23 +128,24 @@ rspamadm_help_help (gboolean full_help)
 }
 
 static void
-rspamadm_help (gint argc, gchar **argv)
+rspamadm_help (gint argc, gchar **argv, const struct rspamadm_command *command)
 {
 	const gchar *cmd_name;
-	const struct rspamadm_command *cmd, **cmd_list;
+	const struct rspamadm_command *cmd;
+	GPtrArray *all_commands = (GPtrArray *)command->command_data;
 
-	printf ("Rspamadm %s\n", RVERSION);
-	printf ("Usage: rspamadm [global_options] command [command_options]\n\n");
+	rspamd_printf ("Rspamadm %s\n", RVERSION);
+	rspamd_printf ("Usage: rspamadm [global_options] command [command_options]\n\n");
 
 	if (argc <= 1) {
 		cmd_name = "help";
 	}
 	else {
 		cmd_name = argv[1];
-		printf ("Showing help for %s command\n\n", cmd_name);
+		rspamd_printf ("Showing help for %s command\n\n", cmd_name);
 	}
 
-	cmd = rspamadm_search_command (cmd_name);
+	cmd = rspamadm_search_command (cmd_name, all_commands);
 
 	if (cmd == NULL) {
 		fprintf (stderr, "Invalid command name: %s\n", cmd_name);
@@ -142,20 +153,30 @@ rspamadm_help (gint argc, gchar **argv)
 	}
 
 	if (strcmp (cmd_name, "help") == 0) {
-		printf ("Available commands:\n");
+		guint i;
+		rspamd_printf ("Available commands:\n");
 
-		cmd_list = commands;
-
-		while (*cmd_list) {
-			if (!((*cmd_list)->flags & RSPAMADM_FLAG_NOHELP)) {
-				printf ("  %-18s %-60s\n", (*cmd_list)->name,
-						(*cmd_list)->help (FALSE));
+		PTR_ARRAY_FOREACH (all_commands, i, cmd) {
+			if (!(cmd->flags & RSPAMADM_FLAG_NOHELP)) {
+				if (!(cmd->flags & RSPAMADM_FLAG_LUA)) {
+					printf ("  %-18s %-60s\n", cmd->name,
+							cmd->help (FALSE, cmd));
+				}
+				else {
+					/* Just call lua subr */
+					(void)cmd->help (FALSE, cmd);
+				}
 			}
-			cmd_list++;
 		}
 	}
 	else {
-		printf ("%s\n", cmd->help (TRUE));
+		if (!(cmd->flags & RSPAMADM_FLAG_LUA)) {
+			rspamd_printf ("%s\n", cmd->help (TRUE, cmd));
+		}
+		else {
+			/* Just call lua subr */
+			(void)cmd->help (TRUE, cmd);
+		}
 	}
 }
 
@@ -185,36 +206,43 @@ rspamadm_parse_ucl_var (const gchar *option_name,
 	return TRUE;
 }
 
-gboolean
-rspamadm_execute_lua_ucl_subr (gpointer pL, gint argc, gchar **argv,
-		const ucl_object_t *res, const gchar *script_name)
+static void
+lua_thread_str_error_cb (struct thread_entry *thread, int ret, const char *msg)
 {
-	lua_State *L = pL;
-	gint err_idx, i, ret;
-	GString *tb;
+	struct lua_call_data *cd = thread->cd;
+
+	msg_err ("call to rspamadm lua script failed (%d): %s", ret, msg);
+
+	cd->ret = ret;
+}
+
+gboolean
+rspamadm_execute_lua_ucl_subr (gint argc, gchar **argv,
+							   const ucl_object_t *res,
+							   const gchar *script_name,
+							   gboolean rspamadm_subcommand)
+{
+	struct thread_entry *thread = lua_thread_pool_get_for_config (rspamd_main->cfg);
+
+	lua_State *L = thread->lua_state;
+
+	gint i;
 	gchar str[PATH_MAX];
-	const struct rspamadm_command **cmd;
 
 	g_assert (script_name != NULL);
 	g_assert (res != NULL);
 	g_assert (L != NULL);
 
 	/* Init internal rspamadm routines */
-	lua_newtable (L);
-	cmd = commands;
 
-	while (*cmd) {
-		if ((*cmd)->lua_subrs != NULL) {
-			(*cmd)->lua_subrs (L);
-		}
-
-		cmd ++;
+	if (rspamadm_subcommand) {
+		rspamd_snprintf (str, sizeof (str), "return require \"%s.%s\"", "rspamadm",
+				script_name);
+	}
+	else {
+		rspamd_snprintf (str, sizeof (str), "return require \"%s\"",
+				script_name);
 	}
-
-	lua_setglobal (L, "rspamadm");
-
-	rspamd_snprintf (str, sizeof (str), "return require \"%s.%s\"", "rspamadm",
-			script_name);
 
 	if (luaL_dostring (L, str) != 0) {
 		msg_err ("cannot execute lua script %s: %s",
@@ -222,6 +250,11 @@ rspamadm_execute_lua_ucl_subr (gpointer pL, gint argc, gchar **argv,
 		return FALSE;
 	}
 	else {
+		if (lua_type (L, -1) == LUA_TTABLE) {
+			lua_pushstring (L, "handler");
+			lua_gettable (L, -2);
+		}
+
 		if (lua_type (L, -1) != LUA_TFUNCTION) {
 			msg_err ("lua script must return "
 					"function and not %s",
@@ -231,32 +264,21 @@ rspamadm_execute_lua_ucl_subr (gpointer pL, gint argc, gchar **argv,
 		}
 	}
 
-	lua_pushcfunction (L, &rspamd_lua_traceback);
-	err_idx = lua_gettop (L);
-
 	/* Push function */
-	lua_pushvalue (L, -2);
+	lua_pushvalue (L, -1);
 
 	/* Push argv */
 	lua_newtable (L);
 
 	for (i = 1; i < argc; i ++) {
 		lua_pushstring (L, argv[i]);
-		lua_rawseti (L, -2, i);
+		lua_rawseti (L, -1, i);
 	}
 
 	/* Push results */
 	ucl_object_push_lua (L, res, TRUE);
 
-	if ((ret = lua_pcall (L, 2, 0, err_idx)) != 0) {
-		tb = lua_touserdata (L, -1);
-		msg_err ("call to rspamadm lua script failed (%d): %v", ret, tb);
-
-		if (tb) {
-			g_string_free (tb, TRUE);
-		}
-
-		lua_settop (L, 0);
+	if (lua_repl_thread_call (thread, 2, NULL, lua_thread_str_error_cb) != 0) {
 
 		return FALSE;
 	}
@@ -267,6 +289,59 @@ rspamadm_execute_lua_ucl_subr (gpointer pL, gint argc, gchar **argv,
 	return TRUE;
 }
 
+static gint
+rspamdadm_commands_sort_func (gconstpointer a, gconstpointer b)
+{
+	const struct rspamadm_command *cmda = *((struct rspamadm_command const **)a),
+			*cmdb = *((struct rspamadm_command const **)b);
+
+	return strcmp (cmda->name, cmdb->name);
+}
+
+static gboolean
+rspamadm_command_maybe_match_name (const gchar *cmd, const gchar *input)
+{
+	gsize clen, inplen;
+
+	clen = strlen (cmd);
+	inplen = strlen (input);
+
+	if (rspamd_strings_levenshtein_distance (cmd, clen,
+			input, inplen, 1) == 1) {
+		return TRUE;
+	}
+	else if ((clen > inplen &&
+			  rspamd_substring_search (cmd, clen, input, inplen) != -1) ||
+			 (inplen > clen &&
+			  rspamd_substring_search (input, inplen, cmd, clen) != -1)) {
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
+
+
+static void
+rspamadm_add_lua_globals (void)
+{
+	struct rspamd_async_session  **psession;
+	struct event_base **pev_base;
+
+	rspamadm_session = rspamd_session_create (rspamd_main->cfg->cfg_pool, NULL,
+			NULL, (event_finalizer_t )NULL, NULL);
+
+	psession = lua_newuserdata (L, sizeof (struct rspamd_async_session*));
+	rspamd_lua_setclass (L, "rspamd{session}", -1);
+	*psession = rspamadm_session;
+	lua_setglobal (L, "rspamadm_session");
+
+	pev_base = lua_newuserdata (L, sizeof (struct event_base *));
+	rspamd_lua_setclass (L, "rspamd{ev_base}", -1);
+	*pev_base = rspamd_main->ev_base;
+	lua_setglobal (L, "rspamadm_ev_base");
+}
+
 gint
 main (gint argc, gchar **argv, gchar **env)
 {
@@ -278,12 +353,14 @@ main (gint argc, gchar **argv, gchar **env)
 	gchar **nargv, **targv;
 	const gchar *cmd_name;
 	const struct rspamadm_command *cmd;
+	GPtrArray *all_commands = g_ptr_array_new (); /* Discovered during check */
 	gint i, nargc, targc;
+	worker_t **pworker;
 
 	ucl_vars = g_hash_table_new_full (rspamd_strcase_hash,
 		rspamd_strcase_equal, g_free, g_free);
 	process_quark = g_quark_from_static_string ("rspamadm");
-	cfg = rspamd_config_new ();
+	cfg = rspamd_config_new (RSPAMD_CONFIG_INIT_DEFAULT);
 	cfg->libs_ctx = rspamd_init_libs ();
 	rspamd_main = g_malloc0 (sizeof (*rspamd_main));
 	rspamd_main->cfg = cfg;
@@ -291,28 +368,10 @@ main (gint argc, gchar **argv, gchar **env)
 	rspamd_main->type = process_quark;
 	rspamd_main->server_pool = rspamd_mempool_new (rspamd_mempool_suggest_size (),
 			"rspamadm");
+	rspamd_main->ev_base = event_init ();
 
-	/* Setup logger */
-	if (verbose) {
-		cfg->log_level = G_LOG_LEVEL_DEBUG;
-	}
-	else {
-		cfg->log_level = G_LOG_LEVEL_MESSAGE;
-	}
-
-	cfg->log_type = RSPAMD_LOG_CONSOLE;
-	/* Avoid timestamps printing */
-	cfg->log_flags = RSPAMD_LOG_FLAG_RSPAMADM;
-	rspamd_set_logger (cfg, process_quark, &rspamd_main->logger,
-			rspamd_main->server_pool);
-	(void) rspamd_log_open (rspamd_main->logger);
-	g_log_set_default_handler (rspamd_glib_log_function, rspamd_main->logger);
-	g_set_printerr_handler (rspamd_glib_printerr_function);
-	rspamd_config_post_load (cfg,
-			RSPAMD_CONFIG_INIT_LIBS|RSPAMD_CONFIG_INIT_URL|RSPAMD_CONFIG_INIT_NO_TLD);
-
-	gperf_profiler_init (cfg, "rspamadm");
-	setproctitle ("rspamdadm");
+	rspamadm_fill_internal_commands (all_commands);
+	help_command.command_data = all_commands;
 
 	/* Now read options and store everything till the first non-dash argument */
 	nargv = g_malloc0 (sizeof (gchar *) * (argc + 1));
@@ -336,19 +395,77 @@ main (gint argc, gchar **argv, gchar **env)
 	g_option_group_add_entries (og, entries);
 	g_option_context_set_summary (context,
 			"Summary:\n  Rspamd administration utility version "
-					RVERSION
-					"\n  Release id: "
-					RID);
+			RVERSION
+			"\n  Release id: "
+			RID);
 	g_option_context_set_main_group (context, og);
 
 	targv = nargv;
 	targc = nargc;
+
 	if (!g_option_context_parse (context, &targc, &targv, &error)) {
 		fprintf (stderr, "option parsing failed: %s\n", error->message);
 		g_error_free (error);
 		exit (1);
 	}
 
+	/* Setup logger */
+	if (verbose) {
+		cfg->log_level = G_LOG_LEVEL_DEBUG;
+	}
+	else {
+		cfg->log_level = G_LOG_LEVEL_MESSAGE;
+	}
+
+	cfg->log_type = RSPAMD_LOG_CONSOLE;
+	/* Avoid timestamps printing */
+	cfg->log_flags = RSPAMD_LOG_FLAG_RSPAMADM;
+	rspamd_set_logger (cfg, process_quark, &rspamd_main->logger,
+			rspamd_main->server_pool);
+	(void) rspamd_log_open (rspamd_main->logger);
+	g_log_set_default_handler (rspamd_glib_log_function, rspamd_main->logger);
+	g_set_printerr_handler (rspamd_glib_printerr_function);
+	rspamd_config_post_load (cfg,
+			RSPAMD_CONFIG_INIT_LIBS|RSPAMD_CONFIG_INIT_URL|RSPAMD_CONFIG_INIT_NO_TLD);
+
+	pworker = &workers[0];
+	while (*pworker) {
+		/* Init string quarks */
+		(void) g_quark_from_static_string ((*pworker)->name);
+		pworker++;
+	}
+
+	cfg->compiled_modules = modules;
+	cfg->compiled_workers = workers;
+
+	gperf_profiler_init (cfg, "rspamadm");
+	setproctitle ("rspamdadm");
+
+	L = cfg->lua_state;
+	rspamd_lua_set_path (L, NULL, ucl_vars);
+	rspamd_lua_set_globals (cfg, L, ucl_vars);
+	rspamadm_add_lua_globals ();
+
+#ifdef WITH_HIREDIS
+	rspamd_redis_pool_config (cfg->redis_pool, cfg, rspamd_main->ev_base);
+#endif
+
+	/* Init rspamadm global */
+	lua_newtable (L);
+
+	PTR_ARRAY_FOREACH (all_commands, i, cmd) {
+		if (cmd->lua_subrs != NULL) {
+			cmd->lua_subrs (L);
+		}
+
+		cmd ++;
+	}
+
+	lua_setglobal (L, "rspamadm");
+
+	rspamadm_fill_lua_commands (L, all_commands);
+	g_ptr_array_sort (all_commands, rspamdadm_commands_sort_func);
+
 	g_strfreev (nargv);
 
 	if (show_version) {
@@ -360,7 +477,7 @@ main (gint argc, gchar **argv, gchar **env)
 		exit (EXIT_SUCCESS);
 	}
 	if (list_commands) {
-		rspamadm_commands ();
+		rspamadm_commands (all_commands);
 		exit (EXIT_SUCCESS);
 	}
 
@@ -370,10 +487,29 @@ main (gint argc, gchar **argv, gchar **env)
 		cmd_name = "help";
 	}
 
-	cmd = rspamadm_search_command (cmd_name);
+	cmd = rspamadm_search_command (cmd_name, all_commands);
 
 	if (cmd == NULL) {
-		fprintf (stderr, "Invalid command name: %s\n", cmd_name);
+		rspamd_fprintf (stderr, "Invalid command name: %s\n", cmd_name);
+
+		/* Try fuzz search */
+		rspamd_fprintf (stderr, "Suggested commands:\n");
+		PTR_ARRAY_FOREACH (all_commands, i, cmd) {
+			guint j;
+			const gchar *alias;
+
+			if (rspamadm_command_maybe_match_name (cmd->name, cmd_name)) {
+				rspamd_fprintf (stderr, "%s\n", cmd->name);
+			}
+			else {
+				PTR_ARRAY_FOREACH (cmd->aliases, j, alias) {
+					if (rspamadm_command_maybe_match_name (alias, cmd_name)) {
+						rspamd_fprintf (stderr, "%s\n", alias);
+					}
+				}
+			}
+		}
+
 		exit (EXIT_FAILURE);
 	}
 
@@ -387,16 +523,19 @@ main (gint argc, gchar **argv, gchar **env)
 
 		targc = argc - nargc;
 		targv = nargv;
-		cmd->run (targc, targv);
+		cmd->run (targc, targv, cmd);
 		g_strfreev (nargv);
 	}
 	else {
-		cmd->run (0, NULL);
+		cmd->run (0, NULL, cmd);
 	}
 
-	rspamd_log_close (rspamd_main->logger);
+	event_base_loopexit (rspamd_main->ev_base, NULL);
+
 	REF_RELEASE (rspamd_main->cfg);
+	rspamd_log_close (rspamd_main->logger, TRUE);
 	g_free (rspamd_main);
+	g_ptr_array_free (all_commands, TRUE);
 
 	return 0;
 }
